@@ -1,11 +1,8 @@
-{-# LANGUAGE CPP                  #-}
-{-# LANGUAGE ScopedTypeVariables  #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Agda.Auto.SearchControl where
 
-import Control.Applicative hiding (getConst, Const(..))
 import Control.Monad
 import Data.IORef
 import Control.Monad.State
@@ -15,7 +12,6 @@ import Agda.Syntax.Common (Hiding(..))
 import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 instance Refinable (ArgList o) (RefInfo o) where
