@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 ------------------------------------------------------------------------
 -- | A command which calls a compiler
@@ -10,14 +9,12 @@ import qualified Control.Exception as E
 import Control.Monad.Trans
 
 import Data.List ( intercalate )
-import qualified Data.List as List
 import System.Exit
 import System.IO
 import System.Process
 
 import Agda.TypeChecking.Monad
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Calls a compiler:
