@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 ------------------------------------------------------------------------
 -- Pretty-printing of Haskell modules
@@ -6,19 +5,12 @@
 
 module Agda.Compiler.MAlonzo.Pretty where
 
-#if MIN_VERSION_base(4,11,0)
-import Prelude hiding ((<>))
-#endif
-
-import Data.Generics.Geniplate
 import qualified Agda.Utils.Haskell.Syntax as HS
 import Text.PrettyPrint (empty)
 
 import Agda.Compiler.MAlonzo.Encode
 import Agda.Utils.Pretty
-import Agda.Utils.Impossible
 
-#include "undefined.h"
 
 prettyPrint :: Pretty a => a -> String
 prettyPrint = show . pretty
@@ -130,7 +122,7 @@ instance Pretty HS.Type where
     case t of
       HS.TyForall xs t ->
         mparens (pr > 0) $
-          sep [ "forall" <+> fsep (map pretty xs) <> "."
+          sep [ ("forall" <+> fsep (map pretty xs)) <> "."
               , nest 2 $ pretty t ]
       HS.TyFun a b ->
         mparens (pr > 4) $
