@@ -1,27 +1,16 @@
-{-# LANGUAGE CPP #-}
 -- | Eliminates case defaults by adding an alternative for all possible
 -- constructors. Literal cases are preserved as-is.
 module Agda.Compiler.Treeless.EliminateDefaults where
 
 import Control.Monad
 import qualified Data.List as List
-import Data.Maybe
 
-import Agda.Syntax.Abstract.Name (QName)
 import Agda.Syntax.Treeless
-import Agda.Syntax.Literal
-import qualified Agda.Syntax.Internal as I
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Substitute
 
-import Agda.Compiler.Treeless.Subst
-
-import Agda.Utils.Impossible
-
-#include "undefined.h"
+import Agda.Compiler.Treeless.Subst () --instance only
 
 eliminateCaseDefaults :: TTerm -> TCM TTerm
 eliminateCaseDefaults = tr
