@@ -1,25 +1,14 @@
 -- | Translates guard alternatives to if-then-else cascades.
 --
 -- The builtin translation must be run before this transformation.
-{-# LANGUAGE CPP #-}
 module Agda.Compiler.Treeless.GuardsToPrims ( convertGuards ) where
 
 import qualified Data.List as List
 
-
-import Agda.Syntax.Abstract.Name (QName)
 import Agda.Syntax.Treeless
-import Agda.Syntax.Literal
-
-import Agda.TypeChecking.Substitute
-
-import Agda.Compiler.Treeless.Subst
 
 import Agda.Utils.Impossible
 
-#include "undefined.h"
-
-
 
 convertGuards :: TTerm -> TTerm
 convertGuards = tr
