@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 module Agda.Compiler.Treeless.Identity
   ( detectIdentityFunctions ) where
@@ -6,19 +5,15 @@ module Agda.Compiler.Treeless.Identity
 import Prelude hiding ((!!))  -- don't use partial functions
 
 import Control.Applicative ( Alternative((<|>), empty) )
-import Data.Foldable (foldMap)
 import Data.Semigroup
 import Data.List.NonEmpty (NonEmpty(..))
 import qualified Data.List as List
 
 import Agda.Syntax.Treeless
-import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Monad
 
-import Agda.Utils.Lens
 import Agda.Utils.List
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 detectIdentityFunctions :: QName -> TTerm -> TCM TTerm
