@@ -1,8 +1,6 @@
-{-# LANGUAGE CPP #-}
 
 module Agda.ImpossibleTest where
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 impossibleTest :: a
