@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 ------------------------------------------------------------------------
 -- | Code for instructing Emacs to do things
@@ -14,10 +13,6 @@ module Agda.Interaction.EmacsCommand
   , displayRunningInfo
   ) where
 
-#if MIN_VERSION_base(4,11,0)
-import Prelude hiding ((<>))
-#endif
-
 import qualified Data.List as List
 
 import Agda.Utils.Pretty
