@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 -- | Generate an import dependency graph for a given module.
 
@@ -9,7 +8,6 @@ import Control.Monad.State
 import qualified Data.Map as M
 import Data.Map(Map)
 import Data.Maybe
-import Data.Monoid
 
 import qualified Data.Set as S
 import Data.Set (Set)
@@ -18,7 +16,6 @@ import Agda.Interaction.Options
 import Agda.Syntax.Abstract
 import Agda.TypeChecking.Monad
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Internal module identifiers for construction of dependency graph.
