@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP              #-}
 
 -- | Generates data used for precise syntax highlighting.
 
@@ -12,7 +11,7 @@ module Agda.Interaction.Highlighting.Generate
   , printUnsolvedInfo
   , printHighlightingInfo
   , highlightAsTypeChecked
-  , warningHighlighting
+  , highlightWarning, warningHighlighting
   , computeUnsolvedMetaWarnings
   , computeUnsolvedConstraints
   , storeDisambiguatedName, disambiguateRecordFields
@@ -21,33 +20,33 @@ module Agda.Interaction.Highlighting.Generate
 import Prelude hiding (null)
 
 import Control.Monad
-import Control.Monad.Trans
-import Control.Monad.Reader
 import Control.Arrow (second)
 
-import Data.Monoid
 import Data.Generics.Geniplate
 import qualified Data.Map as Map
 import Data.Maybe
-import Data.List ((\\), isPrefixOf)
+import Data.List ((\\))
 import qualified Data.List as List
 import qualified Data.Foldable as Fold (fold, foldMap, toList)
 import qualified Data.IntMap as IntMap
+import Data.HashMap.Strict (HashMap)
+import qualified Data.HashMap.Strict as HMap
 import Data.Sequence (Seq)
 import qualified Data.Set as Set
 import qualified Data.Text.Lazy as T
 import Data.Void
 
-import Agda.Interaction.Response (Response(Resp_HighlightingInfo))
-import Agda.Interaction.Highlighting.Precise
-import Agda.Interaction.Highlighting.Range
 import Agda.Interaction.Response
-  (RemoveTokenBasedHighlighting(KeepHighlighting))
+  ( Response( Resp_HighlightingInfo )
+  , RemoveTokenBasedHighlighting( KeepHighlighting )
+  )
+import Agda.Interaction.Highlighting.Precise as P
+import Agda.Interaction.Highlighting.Range (rToR, minus)  -- Range is ambiguous
 
 import qualified Agda.TypeChecking.Errors as E
 import Agda.TypeChecking.MetaVars (isBlockedTerm)
 import Agda.TypeChecking.Monad
-  hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
+  hiding (ModuleInfo, MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 import qualified Agda.TypeChecking.Monad as M
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Warnings (runPM)
@@ -55,19 +54,18 @@ import Agda.TypeChecking.Warnings (runPM)
 import Agda.Syntax.Abstract (IsProjP(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Concrete (FieldAssignment'(..))
-import Agda.Syntax.Concrete.Definitions ( DeclarationWarning(..) )
+import Agda.Syntax.Concrete.Definitions as W ( DeclarationWarning(..) )
 import qualified Agda.Syntax.Common as Common
 import qualified Agda.Syntax.Concrete.Name as C
-import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
-import qualified Agda.Syntax.Info as SI
+import Agda.Syntax.Info ( ModuleInfo(..), defMacro )
 import qualified Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Literal as L
 import qualified Agda.Syntax.Parser as Pa
 import qualified Agda.Syntax.Parser.Tokens as T
 import qualified Agda.Syntax.Position as P
-import Agda.Syntax.Position (getRange)
+import Agda.Syntax.Position (Range, HasRange, getRange, noRange)
 
 import Agda.Utils.FileName
 import Agda.Utils.Function
@@ -78,67 +76,9 @@ import Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Null
 import Agda.Utils.Pretty
-import Agda.Utils.HashMap (HashMap)
-import qualified Agda.Utils.HashMap as HMap
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
--- | @highlightAsTypeChecked rPre r m@ runs @m@ and returns its
---   result. Additionally, some code may be highlighted:
---
--- * If @r@ is non-empty and not a sub-range of @rPre@ (after
---   'P.continuousPerLine' has been applied to both): @r@ is
---   highlighted as being type-checked while @m@ is running (this
---   highlighting is removed if @m@ completes /successfully/).
---
--- * Otherwise: Highlighting is removed for @rPre - r@ before @m@
---   runs, and if @m@ completes successfully, then @rPre - r@ is
---   highlighted as being type-checked.
-
-highlightAsTypeChecked
-  :: MonadTCM tcm
-  => P.Range
-  -> P.Range
-  -> tcm a
-  -> tcm a
-highlightAsTypeChecked rPre r m
-  | r /= P.noRange && delta == rPre' = wrap r'    highlight clear
-  | otherwise                        = wrap delta clear     highlight
-  where
-  rPre'     = rToR (P.continuousPerLine rPre)
-  r'        = rToR (P.continuousPerLine r)
-  delta     = rPre' `minus` r'
-  clear     = mempty
-  highlight = parserBased { otherAspects = Set.singleton TypeChecks }
-
-  wrap rs x y = do
-    p rs x
-    v <- m
-    p rs y
-    return v
-    where
-    p rs x = printHighlightingInfo KeepHighlighting (singletonC rs x)
-
--- | Lispify and print the given highlighting information.
-
-printHighlightingInfo ::
-  MonadTCM tcm =>
-  RemoveTokenBasedHighlighting ->
-  HighlightingInfo ->
-  tcm ()
-printHighlightingInfo remove info = do
-  modToSrc <- useTC stModuleToSource
-  method   <- viewTC eHighlightingMethod
-  liftTCM $ reportSLn "highlighting" 50 $ unlines
-    [ "Printing highlighting info:"
-    , show info
-    , "  modToSrc = " ++ show modToSrc
-    ]
-  unless (null $ ranges info) $ do
-    liftTCM $ appInteractionOutputCallback $
-        Resp_HighlightingInfo info remove method modToSrc
-
 -- | Highlighting levels.
 
 data Level
@@ -149,6 +89,14 @@ data Level
     -- ^ Highlighting without disambiguation of overloaded
     --   constructors.
 
+-- | Highlight a warning.
+highlightWarning :: TCWarning -> TCM ()
+highlightWarning tcwarn = do
+  let h = compress $ warningHighlighting tcwarn
+  modifyTCLens stSyntaxInfo (h <>)
+  ifTopLevelAndHighlightingLevelIs NonInteractive $
+    printHighlightingInfo KeepHighlighting h
+
 -- | Generate syntax highlighting information for the given
 -- declaration, and (if appropriate) print it. If the boolean is
 -- 'True', then the state is additionally updated with the new
@@ -166,7 +114,7 @@ generateAndPrintSyntaxInfo
   -> Bool
      -- ^ Update the state?
   -> TCM ()
-generateAndPrintSyntaxInfo decl _ _ | null $ P.getRange decl = return ()
+generateAndPrintSyntaxInfo decl _ _ | null $ getRange decl = return ()
 generateAndPrintSyntaxInfo decl hlLevel updateState = do
   file <- getCurrentPath
 
@@ -185,8 +133,11 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
 
     let nameInfo = mconcat $ map (generate modMap file kinds) names
 
-    -- Constructors are only highlighted after type checking, since they
-    -- can be overloaded.
+    -- After the code has been type checked more information may be
+    -- available for overloaded constructors, and
+    -- generateConstructorInfo takes advantage of this information.
+    -- Note, however, that highlighting for overloaded constructors is
+    -- included also in nameInfo.
     constructorInfo <- case hlLevel of
       Full{} -> generateConstructorInfo modMap file kinds decl
       _      -> return mempty
@@ -194,10 +145,7 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     cm <- P.rangeFile <$> viewTC eRange
     reportSLn "highlighting.warning" 60 $ "current path = " ++ show cm
 
-    warnInfo <- Fold.foldMap warningHighlighting
-                 . filter ((cm ==) . tcWarningOrigin) <$> useTC stTCWarnings
-
-    let (from, to) = case P.rangeToInterval (P.getRange decl) of
+    let (from, to) = case P.rangeToInterval (getRange decl) of
           Nothing -> __IMPOSSIBLE__
           Just i  -> ( fromIntegral $ P.posPos $ P.iStart i
                      , fromIntegral $ P.posPos $ P.iEnd i)
@@ -213,7 +161,6 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     let syntaxInfo = compress (mconcat [ constructorInfo
                                        , theRest modMap file
                                        , nameInfo
-                                       , warnInfo
                                        ])
                        `mappend`
                      curTokens
@@ -234,7 +181,8 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     universeBi decl
 
   -- Bound variables, dotted patterns, record fields, module names,
-  -- the "as" and "to" symbols.
+  -- the "as" and "to" symbols and some other things.
+  theRest :: SourceToModule -> AbsolutePath -> File
   theRest modMap file = mconcat
     [ Fold.foldMap getFieldDecl   $ universeBi decl
     , Fold.foldMap getVarAndField $ universeBi decl
@@ -251,10 +199,12 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     , Fold.foldMap getNamedArgP   $ universeBi decl
     , Fold.foldMap getNamedArgB   $ universeBi decl
     , Fold.foldMap getNamedArgL   $ universeBi decl
+    , Fold.foldMap getQuantityAttr$ universeBi decl
+    , Fold.foldMap getPragma      $ universeBi decl
     ]
     where
-    bound (A.BindName n) =
-      nameToFile modMap file [] (A.nameConcrete n) P.noRange
+    bound A.BindName{ unBind = n } =
+      nameToFile modMap file [] (A.nameConcrete n) noRange
                  (\isOp -> parserBased { aspect =
                              Just $ Name (Just Bound) isOp })
                  (Just $ A.nameBindingSite n)
@@ -268,12 +218,14 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
                   parserBased { aspect = Just $ Name (Just Macro) isOp }
 
     field :: [C.Name] -> C.Name -> File
-    field m n = nameToFile modMap file m n P.noRange
+    field m n = nameToFile modMap file m n noRange
                            (\isOp -> parserBased { aspect =
                                        Just $ Name (Just Field) isOp })
                            Nothing
+
+    asName :: C.Name -> File
     asName n = nameToFile modMap file []
-                          n P.noRange
+                          n noRange
                           (\isOp -> parserBased { aspect =
                                       Just $ Name (Just Module) isOp })
                           Nothing
@@ -283,14 +235,14 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     -- which defines this module.
     mod isTopLevelModule n =
       nameToFile modMap file []
-                 (A.nameConcrete n) P.noRange
+                 (A.nameConcrete n) noRange
                  (\isOp -> parserBased { aspect =
                              Just $ Name (Just Module) isOp })
                  (Just $ applyWhen isTopLevelModule P.beginningOfFile $
                            A.nameBindingSite n)
 
     getVarAndField :: A.Expr -> File
-    getVarAndField (A.Var x)            = bound $ A.BindName x
+    getVarAndField (A.Var x)            = bound $ A.mkBindName x
     -- Andreas, 2018-06-09, issue #3120
     -- The highlighting for record field tags is now created by the type checker in
     -- function disambiguateRecordFields.
@@ -314,9 +266,12 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
 
     getNamedArg :: Common.NamedArg a -> File
     getNamedArg x = caseMaybe (Common.nameOf $ Common.unArg x) mempty $ \ s ->
-      singleton (rToR $ P.getRange s) $
+      singleton (rToR $ getRange s) $
         parserBased { aspect = Just $ Name (Just Argument) False }
 
+    getBinder :: A.Binder -> File
+    getBinder (A.Binder _ n) = bound n
+
     getLet :: A.LetBinding -> File
     getLet (A.LetBind _ _ x _ _)     = bound x
     getLet A.LetPatBind{}            = mempty
@@ -325,24 +280,44 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
     getLet (A.LetDeclaredVariable x) = bound x
 
     getLam :: A.LamBinding -> File
-    getLam (A.DomainFree x)  = bound $ Common.namedArg x
-    getLam (A.DomainFull {}) = mempty
+    getLam (A.DomainFree _ xs) = getBinder (Common.namedArg xs)
+    getLam (A.DomainFull {})   = mempty
 
     getTyped :: A.TypedBinding -> File
-    getTyped (A.TBind _ xs _) = mconcat $ map (bound . Common.namedArg) xs
-    getTyped A.TLet{}         = mempty
+    getTyped (A.TBind _ _ xs _) = Fold.foldMap (getBinder . Common.namedArg) xs
+    getTyped A.TLet{}           = mempty
 
     getPatSynArgs :: A.Declaration -> File
-    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . A.BindName . Common.unArg) xs
+    getPatSynArgs (A.PatternSynDef _ xs _) = mconcat $ map (bound . A.mkBindName . Common.unArg) xs
     getPatSynArgs _                        = mempty
 
+    -- Issue #4361, highlight BUILTINs like NAT, EQUALITY etc. in keyword color.
+    getPragma :: A.Declaration -> File
+    getPragma = \case
+      A.Pragma _ p ->
+        case p of
+          A.BuiltinPragma b _      -> keyword b
+          A.BuiltinNoDefPragma b _ -> keyword b
+          A.CompilePragma b _ _    -> keyword b
+          A.RewritePragma r _      -> keyword r
+          A.OptionsPragma{}   -> mempty
+          A.StaticPragma{}    -> mempty
+          A.EtaPragma{}       -> mempty
+          A.InjectivePragma{} -> mempty
+          A.InlinePragma{}    -> mempty
+          A.DisplayPragma{}   -> mempty
+      _ -> mempty
+
+    keyword :: HasRange a => a -> File
+    keyword x = singleton (rToR $ getRange x) $ parserBased { aspect = Just Keyword }
+
     getPattern' :: IsProjP e => A.Pattern' e -> File
     getPattern' (A.VarP x)    = bound x
     getPattern' (A.AsP _ x _) = bound x
     getPattern' (A.DotP pi e)
       | Just _ <- isProjP e = mempty
       | otherwise =
-          singleton (rToR $ P.getRange pi)
+          singleton (rToR $ getRange pi)
                 (parserBased { otherAspects = Set.singleton DottedPattern })
     getPattern' (A.PatternSynP _ q _) = patsyn q
     -- Andreas, 2018-06-09, issue #3120
@@ -387,12 +362,16 @@ generateAndPrintSyntaxInfo decl hlLevel updateState = do
                    Just (C.toTopLevelModuleName $ A.mnameToConcrete m)
           []    -> False
 
-    getModuleInfo :: SI.ModuleInfo -> File
-    getModuleInfo (SI.ModuleInfo { SI.minfoAsTo   = asTo
-                                 , SI.minfoAsName = name }) =
-      singleton (rToR asTo) (parserBased { aspect = Just Symbol })
+    getModuleInfo :: ModuleInfo -> File
+    getModuleInfo (ModuleInfo{ minfoAsTo, minfoAsName }) =
+      singleton (rToR minfoAsTo) (parserBased { aspect = Just Symbol })
         `mappend`
-      maybe mempty asName name
+      maybe mempty asName minfoAsName
+
+    -- If the Quantity attribute comes with a Range, highlight the
+    -- corresponding attribute as Symbol.
+    getQuantityAttr :: Common.Quantity -> File
+    getQuantityAttr q = singleton (rToR $ getRange q) (parserBased { aspect = Just Symbol })
 
 -- | Generate and return the syntax highlighting information for the
 -- tokens in the given file.
@@ -419,8 +398,8 @@ generateTokenInfoFromSource file input =
 -- tokens in the given string, which is assumed to correspond to the
 -- given range.
 
-generateTokenInfoFromString :: P.Range -> String -> TCM CompressedFile
-generateTokenInfoFromString r _ | r == P.noRange = return mempty
+generateTokenInfoFromString :: Range -> String -> TCM CompressedFile
+generateTokenInfoFromString r _ | r == noRange = return mempty
 generateTokenInfoFromString r s = do
   runPM $ tokenHighlighting <$> Pa.parsePosString Pa.tokensParser p s
   where
@@ -437,13 +416,14 @@ tokenHighlighting = merge . map tokenToCFile
   merge = CompressedFile . concat . map ranges
 
   tokenToCFile :: T.Token -> CompressedFile
-  tokenToCFile (T.TokSetN (i, _))               = aToF PrimitiveType (P.getRange i)
-  tokenToCFile (T.TokPropN (i, _))              = aToF PrimitiveType (P.getRange i)
-  tokenToCFile (T.TokKeyword T.KwSet  i)        = aToF PrimitiveType (P.getRange i)
-  tokenToCFile (T.TokKeyword T.KwProp i)        = aToF PrimitiveType (P.getRange i)
-  tokenToCFile (T.TokKeyword T.KwForall i)      = aToF Symbol (P.getRange i)
-  tokenToCFile (T.TokKeyword _ i)               = aToF Keyword (P.getRange i)
-  tokenToCFile (T.TokSymbol  _ i)               = aToF Symbol (P.getRange i)
+  tokenToCFile (T.TokSetN (i, _))               = aToF PrimitiveType (getRange i)
+  tokenToCFile (T.TokPropN (i, _))              = aToF PrimitiveType (getRange i)
+  tokenToCFile (T.TokKeyword T.KwSet  i)        = aToF PrimitiveType (getRange i)
+  tokenToCFile (T.TokKeyword T.KwProp i)        = aToF PrimitiveType (getRange i)
+  tokenToCFile (T.TokKeyword T.KwForall i)      = aToF Symbol (getRange i)
+  tokenToCFile (T.TokKeyword T.KwREWRITE _)     = mempty  -- #4361, REWRITE is not always a Keyword
+  tokenToCFile (T.TokKeyword _ i)               = aToF Keyword (getRange i)
+  tokenToCFile (T.TokSymbol  _ i)               = aToF Symbol (getRange i)
   tokenToCFile (T.TokLiteral (L.LitNat    r _)) = aToF Number r
   tokenToCFile (T.TokLiteral (L.LitWord64 r _)) = aToF Number r
   tokenToCFile (T.TokLiteral (L.LitFloat  r _)) = aToF Number r
@@ -451,12 +431,12 @@ tokenHighlighting = merge . map tokenToCFile
   tokenToCFile (T.TokLiteral (L.LitChar   r _)) = aToF String r
   tokenToCFile (T.TokLiteral (L.LitQName  r _)) = aToF String r
   tokenToCFile (T.TokLiteral (L.LitMeta r _ _)) = aToF String r
-  tokenToCFile (T.TokComment (i, _))            = aToF Comment (P.getRange i)
-  tokenToCFile (T.TokTeX (i, _))                = aToF Background (P.getRange i)
-  tokenToCFile (T.TokMarkup (i, _))             = aToF Markup (P.getRange i)
+  tokenToCFile (T.TokComment (i, _))            = aToF Comment (getRange i)
+  tokenToCFile (T.TokTeX (i, _))                = aToF Background (getRange i)
+  tokenToCFile (T.TokMarkup (i, _))             = aToF Markup (getRange i)
   tokenToCFile (T.TokId {})                     = mempty
   tokenToCFile (T.TokQId {})                    = mempty
-  tokenToCFile (T.TokString (i,s))              = aToF Pragma (P.getRange i)
+  tokenToCFile (T.TokString (i,s))              = aToF Pragma (getRange i)
   tokenToCFile (T.TokDummy {})                  = mempty
   tokenToCFile (T.TokEOF {})                    = mempty
 
@@ -473,17 +453,25 @@ nameKinds :: Level
           -> A.Declaration
           -> TCM NameKinds
 nameKinds hlLevel decl = do
-  imported <- fix <$> useTC stImports
+  imported <- useTC $ stImports . sigDefinitions
   local    <- case hlLevel of
-    Full{} -> fix <$> useTC stSignature
+    Full{} -> useTC $ stSignature . sigDefinitions
     _      -> return HMap.empty
+  impPatSyns <- useTC stPatternSynImports
+  locPatSyns <- case hlLevel of
+    Full{} -> useTC stPatternSyns
+    _      -> return empty
       -- Traverses the syntax tree and constructs a map from qualified
       -- names to name kinds. TODO: Handle open public.
   let syntax = foldr ($) HMap.empty $ map declToKind $ universeBi decl
-  let merged = unions [local, imported, syntax]
-  return (\n -> HMap.lookup n merged)
+  return $ \ n -> unionsMaybeWith merge
+    [ defnToKind . theDef <$> HMap.lookup n local
+    , con <$ Map.lookup n locPatSyns
+    , defnToKind . theDef <$> HMap.lookup n imported
+    , con <$ Map.lookup n impPatSyns
+    , HMap.lookup n syntax
+    ]
   where
-  fix = HMap.map (defnToKind . theDef) . (^. sigDefinitions)
 
   -- | The 'M.Axiom' constructor is used to represent various things
   -- which are not really axioms, so when maps are merged 'Postulate's
@@ -494,7 +482,6 @@ nameKinds hlLevel decl = do
   merge _     Macro = Macro  -- If the abstract syntax says macro, it's a macro.
   merge k         _ = k
 
-  unions = foldr (HMap.unionWith merge) HMap.empty
   insert = HMap.insertWith merge
 
   defnToKind :: Defn -> NameKind
@@ -512,8 +499,8 @@ nameKinds hlLevel decl = do
   declToKind :: A.Declaration ->
                 HashMap A.QName NameKind -> HashMap A.QName NameKind
   declToKind (A.Axiom _ i _ _ q _)
-    | SI.defMacro i == Common.MacroDef = insert q Macro
-    | otherwise                        = insert q Postulate
+    | defMacro i == Common.MacroDef = insert q Macro
+    | otherwise                     = insert q Postulate
   declToKind (A.Field _ q _)        = insert q Field -- Function
     -- Note that the name q can be used both as a field name and as a
     -- projection function. Highlighting of field names is taken care
@@ -526,7 +513,7 @@ nameKinds hlLevel decl = do
   declToKind (A.Pragma {})          = id
   declToKind (A.ScopedDecl {})      = id
   declToKind (A.Open {})            = id
-  declToKind (A.PatternSynDef q _ _) = insert q (Constructor Common.Inductive)
+  declToKind (A.PatternSynDef q _ _) = insert q con
   declToKind (A.Generalize _ _ _ q _) = insert q Generalizable
   declToKind (A.FunDef  _ q _ _)     = insert q Function
   declToKind (A.UnquoteDecl _ _ qs _) = foldr (\ q f -> insert q Function . f) id qs
@@ -534,14 +521,13 @@ nameKinds hlLevel decl = do
   declToKind (A.DataSig _ q _ _)      = insert q Datatype
   declToKind (A.DataDef _ q _ _ cs)   = \m ->
                                       insert q Datatype $
-                                      foldr (\d -> insert (A.axiomName d)
-                                                          (Constructor Common.Inductive))
+                                      foldr (\d -> insert (A.axiomName d) con)
                                             m cs
   declToKind (A.RecSig _ q _ _)       = insert q Record
-  declToKind (A.RecDef _ q _ _ _ c _ _ _) = insert q Record .
-                                      case c of
-                                        Nothing -> id
-                                        Just q  -> insert q (Constructor Common.Inductive)
+  declToKind (A.RecDef _ q _ _ _ c _ _ _) = insert q Record . maybe id (`insert` con) c
+
+  con :: NameKind
+  con = Constructor Common.Inductive
 
 -- | Generates syntax highlighting information for all constructors
 -- occurring in patterns and expressions in the given declaration.
@@ -561,7 +547,7 @@ generateConstructorInfo modMap file kinds decl = do
   -- Get boundaries of current declaration.
   -- @noRange@ should be impossible, but in case of @noRange@
   -- it makes sense to return the empty File.
-  ifNull (P.rangeIntervals $ P.getRange decl)
+  ifNull (P.rangeIntervals $ getRange decl)
          (return mempty) $ \is -> do
     let start = fromIntegral $ P.posPos $ P.iStart $ head is
         end   = fromIntegral $ P.posPos $ P.iEnd   $ last is
@@ -576,7 +562,7 @@ generateConstructorInfo modMap file kinds decl = do
     let files = for constrs $ \ q -> generate modMap file kinds $ I.unambiguous q
     return $ Fold.fold files
 
-printSyntaxInfo :: P.Range -> TCM ()
+printSyntaxInfo :: Range -> TCM ()
 printSyntaxInfo r = do
   syntaxInfo <- useTC stSyntaxInfo
   ifTopLevelAndHighlightingLevelIs NonInteractive $
@@ -590,13 +576,12 @@ printErrorInfo e =
     errorHighlighting e
 
 -- | Generate highlighting for error.
---   Does something special for termination errors.
 
 errorHighlighting :: TCErr -> TCM File
 errorHighlighting e = do
 
   -- Erase previous highlighting.
-  let r     = P.getRange e
+  let r     = getRange e
       erase = singleton (rToR $ P.continuousPerLine r) mempty
 
   -- Print new highlighting.
@@ -613,23 +598,29 @@ warningHighlighting :: TCWarning -> File
 warningHighlighting w = case tcWarning w of
   TerminationIssue terrs     -> terminationErrorHighlighting terrs
   NotStrictlyPositive d ocs  -> positivityErrorHighlighting d ocs
-  UnreachableClauses{}       -> deadcodeHighlighting $ P.getRange w
-  CoverageIssue{}            -> coverageErrorHighlighting $ P.getRange w
-  CoverageNoExactSplit{}     -> catchallHighlighting $ P.getRange w
+  -- #3965 highlight each unreachable clause independently: they
+  -- may be interleaved with actually reachable clauses!
+  UnreachableClauses _ rs    -> Fold.foldMap deadcodeHighlighting rs
+  CoverageIssue{}            -> coverageErrorHighlighting $ getRange w
+  CoverageNoExactSplit{}     -> catchallHighlighting $ getRange w
   UnsolvedConstraints cs     -> constraintsHighlighting cs
   UnsolvedMetaVariables rs   -> metasHighlighting rs
-  AbsurdPatternRequiresNoRHS{} -> deadcodeHighlighting $ P.getRange w
-  ModuleDoesntExport{}         -> deadcodeHighlighting $ P.getRange w
+  AbsurdPatternRequiresNoRHS{} -> deadcodeHighlighting $ getRange w
+  ModuleDoesntExport{}         -> deadcodeHighlighting $ getRange w
+  FixityInRenamingModule rs    -> Fold.foldMap deadcodeHighlighting rs
   -- expanded catch-all case to get a warning for new constructors
   CantGeneralizeOverSorts{}  -> mempty
   UnsolvedInteractionMetas{} -> mempty
   OldBuiltin{}               -> mempty
-  EmptyRewritePragma{}       -> deadcodeHighlighting $ P.getRange w
-  IllformedAsClause{}        -> deadcodeHighlighting $ P.getRange w
-  UselessPublic{}            -> mempty
+  EmptyRewritePragma{}       -> deadcodeHighlighting $ getRange w
+  IllformedAsClause{}        -> deadcodeHighlighting $ getRange w
+  UselessPublic{}            -> deadcodeHighlighting $ getRange w
   UselessInline{}            -> mempty
+  ClashesViaRenaming _ xs    -> Fold.foldMap (deadcodeHighlighting . getRange) xs
+    -- #4154, TODO: clashing renamings are not dead code, but introduce problems.
+    -- Should we have a different color?
   WrongInstanceDeclaration{} -> mempty
-  InstanceWithExplicitArg{}  -> deadcodeHighlighting $ P.getRange w
+  InstanceWithExplicitArg{}  -> deadcodeHighlighting $ getRange w
   InstanceNoOutputTypeName{} -> mempty
   InstanceArgWithExplicitArg{} -> mempty
   ParseWarning{}             -> mempty
@@ -641,6 +632,9 @@ warningHighlighting w = case tcWarning w of
   SafeFlagNonTerminating     -> mempty
   SafeFlagTerminating        -> mempty
   SafeFlagWithoutKFlagPrimEraseEquality -> mempty
+  SafeFlagEta                -> mempty
+  SafeFlagInjective          -> mempty
+  SafeFlagNoCoverageCheck    -> mempty
   WithoutKFlagPrimEraseEquality -> mempty
   SafeFlagNoPositivityCheck  -> mempty
   SafeFlagPolarity           -> mempty
@@ -650,34 +644,45 @@ warningHighlighting w = case tcWarning w of
   LibraryWarning{}           -> mempty
   InfectiveImport{}          -> mempty
   CoInfectiveImport{}        -> mempty
+  RewriteNonConfluent{}      -> confluenceErrorHighlighting $ getRange w
+  RewriteMaybeNonConfluent{} -> confluenceErrorHighlighting $ getRange w
+  PragmaCompileErased{}      -> deadcodeHighlighting $ getRange w
+  NotInScopeW{}              -> deadcodeHighlighting $ getRange w
   NicifierIssue w           -> case w of
     -- we intentionally override the binding of `w` here so that our pattern of
-    -- using `P.getRange w` still yields the most precise range information we
+    -- using `getRange w` still yields the most precise range information we
     -- can get.
-    NotAllowedInMutual{} -> deadcodeHighlighting $ P.getRange w
-    EmptyAbstract{}      -> deadcodeHighlighting $ P.getRange w
-    EmptyInstance{}      -> deadcodeHighlighting $ P.getRange w
-    EmptyMacro{}         -> deadcodeHighlighting $ P.getRange w
-    EmptyMutual{}        -> deadcodeHighlighting $ P.getRange w
-    EmptyPostulate{}     -> deadcodeHighlighting $ P.getRange w
-    EmptyPrivate{}       -> deadcodeHighlighting $ P.getRange w
-    EmptyGeneralize{}    -> deadcodeHighlighting $ P.getRange w
-    UselessAbstract{}    -> deadcodeHighlighting $ P.getRange w
-    UselessInstance{}    -> deadcodeHighlighting $ P.getRange w
-    UselessPrivate{}     -> deadcodeHighlighting $ P.getRange w
+    NotAllowedInMutual{}             -> deadcodeHighlighting $ getRange w
+    EmptyAbstract{}                  -> deadcodeHighlighting $ getRange w
+    EmptyInstance{}                  -> deadcodeHighlighting $ getRange w
+    EmptyMacro{}                     -> deadcodeHighlighting $ getRange w
+    EmptyMutual{}                    -> deadcodeHighlighting $ getRange w
+    EmptyPostulate{}                 -> deadcodeHighlighting $ getRange w
+    EmptyPrimitive{}                 -> deadcodeHighlighting $ getRange w
+    EmptyPrivate{}                   -> deadcodeHighlighting $ getRange w
+    EmptyGeneralize{}                -> deadcodeHighlighting $ getRange w
+    EmptyField{}                     -> deadcodeHighlighting $ getRange w
+    UselessAbstract{}                -> deadcodeHighlighting $ getRange w
+    UselessInstance{}                -> deadcodeHighlighting $ getRange w
+    UselessPrivate{}                 -> deadcodeHighlighting $ getRange w
+    InvalidNoPositivityCheckPragma{} -> deadcodeHighlighting $ getRange w
+    InvalidNoUniverseCheckPragma{}   -> deadcodeHighlighting $ getRange w
+    InvalidTerminationCheckPragma{}  -> deadcodeHighlighting $ getRange w
+    InvalidCoverageCheckPragma{}     -> deadcodeHighlighting $ getRange w
+    OpenPublicAbstract{}             -> deadcodeHighlighting $ getRange w
+    OpenPublicPrivate{}              -> deadcodeHighlighting $ getRange w
+    W.ShadowingInTelescope nrs       -> Fold.foldMap
+                                          (shadowingTelHighlighting . snd)
+                                          nrs
+    MissingDefinitions{}             -> missingDefinitionHighlighting $ getRange w
     -- TODO: explore highlighting opportunities here!
-    EmptyPrimitive{} -> mempty
-    InvalidCatchallPragma{} -> mempty
-    InvalidNoPositivityCheckPragma{} -> mempty
-    InvalidNoUniverseCheckPragma{} -> mempty
-    InvalidTerminationCheckPragma{} -> mempty
-    MissingDefinitions{} -> mempty
+    InvalidCatchallPragma{}           -> mempty
     PolarityPragmasButNotPostulates{} -> mempty
-    PragmaNoTerminationCheck{} -> mempty
-    PragmaCompiled{} -> mempty
-    UnknownFixityInMixfixDecl{} -> mempty
-    UnknownNamesInFixityDecl{} -> mempty
-    UnknownNamesInPolarityPragmas{} -> mempty
+    PragmaNoTerminationCheck{}        -> mempty
+    PragmaCompiled{}                  -> mempty
+    UnknownFixityInMixfixDecl{}       -> mempty
+    UnknownNamesInFixityDecl{}        -> mempty
+    UnknownNamesInPolarityPragmas{}   -> mempty
 
 
 -- | Generate syntax highlighting for termination errors.
@@ -694,27 +699,41 @@ terminationErrorHighlighting termErrs = functionDefs `mappend` callSites
 -- | Generate syntax highlighting for not-strictly-positive inductive
 -- definitions.
 
--- TODO: highlight also the problematic occurrences
 positivityErrorHighlighting :: I.QName -> Seq OccursWhere -> File
 positivityErrorHighlighting q os =
-  several (rToR <$> P.getRange q : rs) m
+  several (rToR <$> getRange q : rs) m
   where
     rs = map (\(OccursWhere r _ _) -> r) (Fold.toList os)
     m  = parserBased { otherAspects = Set.singleton PositivityProblem }
 
-deadcodeHighlighting :: P.Range -> File
-deadcodeHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+deadcodeHighlighting :: Range -> File
+deadcodeHighlighting r = singleton (rToR $ P.continuous r) m
   where m = parserBased { otherAspects = Set.singleton Deadcode }
 
-coverageErrorHighlighting :: P.Range -> File
+coverageErrorHighlighting :: Range -> File
 coverageErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
   where m = parserBased { otherAspects = Set.singleton CoverageProblem }
 
+shadowingTelHighlighting :: [Range] -> File
+shadowingTelHighlighting =
+  -- we do not want to highlight the one variable in scope so we take
+  -- the @init@ segment of the ranges in question
+  Fold.foldMap (\r -> singleton (rToR $ P.continuous r) m) . init
+  where
+  m = parserBased { otherAspects =
+                      Set.singleton P.ShadowingInTelescope }
 
-catchallHighlighting :: P.Range -> File
+catchallHighlighting :: Range -> File
 catchallHighlighting r = singleton (rToR $ P.continuousPerLine r) m
   where m = parserBased { otherAspects = Set.singleton CatchallClause }
 
+confluenceErrorHighlighting :: Range -> File
+confluenceErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+  where m = parserBased { otherAspects = Set.singleton ConfluenceProblem }
+
+missingDefinitionHighlighting :: Range -> File
+missingDefinitionHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+  where m = parserBased { otherAspects = Set.singleton MissingDefinition }
 
 -- | Generates and prints syntax highlighting information for unsolved
 -- meta-variables and certain unsolved constraints.
@@ -743,7 +762,7 @@ computeUnsolvedMetaWarnings = do
   rs <- mapM getMetaRange (ms \\ is)
   return $ metasHighlighting rs
 
-metasHighlighting :: [P.Range] -> File
+metasHighlighting :: [Range] -> File
 metasHighlighting rs = several (map (rToR . P.continuousPerLine) rs)
                      $ parserBased { otherAspects = Set.singleton UnsolvedMeta }
 
@@ -763,7 +782,6 @@ constraintsHighlighting cs =
     Closure{ clValue = IsEmpty r t           } -> Just r
     Closure{ clEnv = e, clValue = ValueCmp{} } -> Just $ getRange (envRange e)
     Closure{ clEnv = e, clValue = ElimCmp{}  } -> Just $ getRange (envRange e)
-    Closure{ clEnv = e, clValue = TypeCmp{}  } -> Just $ getRange (envRange e)
     Closure{ clEnv = e, clValue = TelCmp{}   } -> Just $ getRange (envRange e)
     Closure{ clEnv = e, clValue = SortCmp{}  } -> Just $ getRange (envRange e)
     Closure{ clEnv = e, clValue = LevelCmp{} } -> Just $ getRange (envRange e)
@@ -809,12 +827,12 @@ nameToFile :: SourceToModule
               -- ^ The name qualifier (may be empty).
            -> C.Name
               -- ^ The base name.
-           -> P.Range
+           -> Range
               -- ^ The 'Range' of the name in its fixity declaration (if any).
            -> (Bool -> Aspects)
               -- ^ Meta information to be associated with the name.
               -- The argument is 'True' iff the name is an operator.
-           -> Maybe P.Range
+           -> Maybe Range
               -- ^ The definition site of the name. The calculated
               -- meta information is extended with this information,
               -- if possible.
@@ -829,9 +847,9 @@ nameToFile modMap file xs x fr m mR =
     mempty
   where
   aspects    = m $ C.isOperator x
-  fileNames  = mapMaybe (fmap P.srcFile . P.rStart . P.getRange) (x : xs)
+  fileNames  = mapMaybe (fmap P.srcFile . P.rStart . getRange) (x : xs)
   frFile     = singleton (rToR fr) (aspects { definitionSite = notHere <$> mFilePos })
-  rs         = map P.getRange (x : xs)
+  rs         = map getRange (x : xs)
 
   -- The fixity declaration should not get a symbolic anchor.
   notHere d = d { defSiteHere = False }
@@ -853,7 +871,7 @@ nameToFile modMap file xs x fr m mR =
       { defSiteModule = mod
       , defSitePos    = fromIntegral p
         -- Is our current position the definition site?
-      , defSiteHere   = r == P.getRange x
+      , defSiteHere   = r == getRange x
         -- For bound variables etc. we do not create a symbolic anchor name.
         -- Also not for names that include anonymous modules,
         -- otherwise, we do not get unique anchors.
@@ -903,24 +921,36 @@ nameToFileA modMap file x include m =
              file
              (concreteQualifier x)
              (concreteBase x)
-             r
+             rangeOfFixityDeclaration
              m
              (if include then Just $ bindingSite x else Nothing)
     `mappend` notationFile
   where
-    -- Andreas, 2016-09-08, for issue #2140:
-    -- Range of name from fixity declaration:
-    fr = theNameRange $ A.nameFixity $ A.qnameName x
-    -- Somehow we import fixity ranges from other files, we should ignore them.
-    -- (I do not understand how we get them as they should not be serialized...)
-    r = if P.rangeFile fr == Strict.Just file then fr else P.noRange
+  -- TODO: Currently we highlight fixity and syntax declarations by
+  -- producing highlighting something like once per occurrence of the
+  -- related name(s) in the file of the declaration (and we explicitly
+  -- avoid doing this for other files). Perhaps it would be better to
+  -- only produce this highlighting once.
+
+  rangeOfFixityDeclaration =
+    if P.rangeFile r == Strict.Just file
+    then r else noRange
+    where
+    r = Common.theNameRange $ A.nameFixity $ A.qnameName x
+
+  notationFile =
+    if P.rangeFile (getRange notation) == Strict.Just file
+    then mconcat $ map genPartFile notation
+    else mempty
+    where
+    notation = Common.theNotation $ A.nameFixity $ A.qnameName x
 
-    notationFile = mconcat $ map genPartFile $ theNotation $ A.nameFixity $ A.qnameName x
     boundAspect = parserBased{ aspect = Just $ Name (Just Bound) False }
-    genPartFile (BindHole r i)   = several [rToR r, rToR $ getRange i] boundAspect
-    genPartFile (NormalHole r i) = several [rToR r, rToR $ getRange i] boundAspect
-    genPartFile WildHole{}       = mempty
-    genPartFile (IdPart x)       = singleton (rToR $ P.getRange x) (m False)
+
+    genPartFile (Common.BindHole r i)   = several [rToR r, rToR $ getRange i] boundAspect
+    genPartFile (Common.NormalHole r i) = several [rToR r, rToR $ getRange i] boundAspect
+    genPartFile Common.WildHole{}       = mempty
+    genPartFile (Common.IdPart x)       = singleton (rToR $ getRange x) (m False)
 
 concreteBase :: I.QName -> C.Name
 concreteBase = A.nameConcrete . A.qnameName
@@ -928,13 +958,13 @@ concreteBase = A.nameConcrete . A.qnameName
 concreteQualifier :: I.QName -> [C.Name]
 concreteQualifier = map A.nameConcrete . A.mnameToList . A.qnameModule
 
-bindingSite :: I.QName -> P.Range
+bindingSite :: I.QName -> Range
 bindingSite = A.nameBindingSite . A.qnameName
 
 -- | Remember a name disambiguation (during type checking).
 --   To be used later during syntax highlighting.
 storeDisambiguatedName :: A.QName -> TCM ()
-storeDisambiguatedName q = whenJust (start $ P.getRange q) $ \ i ->
+storeDisambiguatedName q = whenJust (start $ getRange q) $ \ i ->
   stDisambiguatedNames `modifyTCLens` IntMap.insert i q
   where
   start r = fromIntegral . P.posPos <$> P.rStart' r
