@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP               #-}
 
 -- | Function for generating highlighted, hyperlinked HTML from Agda
 -- sources.
@@ -15,12 +14,10 @@ module Agda.Interaction.Highlighting.HTML
   ) where
 
 import Prelude hiding ((!!), concatMap)
-import Control.Arrow ((***))
 import Control.Monad
 import Control.Monad.Trans
 
 import Data.Function
-import Data.Monoid
 import Data.Foldable (toList, concatMap)
 import Data.Maybe
 import qualified Data.IntMap as IntMap
@@ -46,26 +43,18 @@ import Paths_Agda
 
 import Agda.Interaction.Options
 import Agda.Interaction.Highlighting.Precise
-import Agda.Interaction.Highlighting.Common
-
-import Agda.Interaction.Highlighting.Generate
-  (computeUnsolvedMetaWarnings, computeUnsolvedConstraints)
 
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Common
-import Agda.Syntax.Abstract.Name (ModuleName)
 
 import Agda.TypeChecking.Monad (TCM, useTC)
 import qualified Agda.TypeChecking.Monad as TCM
 
 import Agda.Utils.FileName (filePath)
 import Agda.Utils.Function
-import Agda.Utils.Lens
 import qualified Agda.Utils.IO.UTF8 as UTF8
-import Agda.Utils.Pretty hiding ((<>))
-import Agda.Utils.Tuple
+import Agda.Utils.Pretty
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | The name of the default CSS file.
@@ -154,8 +143,8 @@ generateHTMLWithPageGen generatePage = do
         cssFile <- getDataFileName defaultCSSFile
         copyFile cssFile (dir </> defaultCSSFile)
 
-      TCM.reportSLn "html" 1 $ unlines
-        [ ""
+      TCM.reportS "html" 1
+        [ "" :: String
         , "Warning: HTML is currently generated for ALL files which can be"
         , "reached from the given module, including library files."
         ]
