@@ -1,31 +1,30 @@
-{-# LANGUAGE CPP             #-}
 {-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.Interaction.MakeCase where
 
 import Prelude hiding (mapM, mapM_, null)
 
-import Control.Applicative hiding (empty)
 import Control.Monad hiding (mapM, mapM_, forM)
-import Control.Monad.Reader (asks)
 
+import Data.Either
 import qualified Data.Map as Map
 import qualified Data.List as List
 import Data.Maybe
-import Data.Traversable
+import Data.Monoid
+import Data.Traversable (mapM, forM)
 
 import Agda.Syntax.Common
+import Agda.Syntax.Info
 import Agda.Syntax.Position
-import Agda.Syntax.Concrete (NameInScope(..), LensInScope(..))
+import Agda.Syntax.Concrete (NameInScope(..))
 import qualified Agda.Syntax.Concrete as C
+import qualified Agda.Syntax.Concrete.Pattern as C
 import qualified Agda.Syntax.Abstract as A
-import qualified Agda.Syntax.Abstract.Views as A
-import qualified Agda.Syntax.Info as A
+import qualified Agda.Syntax.Abstract.Pattern as A
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
-import Agda.Syntax.Scope.Base  ( ResolvedName(..), Binder(..), KindOfName(..), allKindsOfNames )
-import Agda.Syntax.Scope.Monad ( resolveName, resolveName' )
-import Agda.Syntax.Translation.ConcreteToAbstract
+import Agda.Syntax.Scope.Base  ( ResolvedName(..), BindingSource(..), KindOfName(..), exceptKindsOfNames )
+import Agda.Syntax.Scope.Monad ( resolveName' )
 import Agda.Syntax.Translation.InternalToAbstract
 
 import Agda.TypeChecking.Monad
@@ -33,28 +32,22 @@ import Agda.TypeChecking.Coverage
 import Agda.TypeChecking.Coverage.Match ( SplitPatVar(..) , SplitPattern , applySplitPSubst , fromSplitPatterns )
 import Agda.TypeChecking.Empty ( isEmptyTel )
 import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Irrelevance
-import Agda.TypeChecking.Rules.LHS.Implicit
-import Agda.TheTypeChecker
+import Agda.TypeChecking.Rules.Def (checkClauseLHS)
+import Agda.TypeChecking.Rules.LHS (LHSResult(..))
+import Agda.TypeChecking.Rules.Term (isModuleFreeVar)
 
 import Agda.Interaction.Options
 import Agda.Interaction.BasicOps
 
 import Agda.Utils.Function
 import Agda.Utils.Functor
-import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import qualified Agda.Utils.Pretty as P
-import Agda.Utils.Singleton
 import Agda.Utils.Size
-import qualified Agda.Utils.HashMap as HMap
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 type CaseContext = Maybe ExtLamInfo
@@ -85,8 +78,9 @@ parseVariables f tel ii rng ss = do
 
     -- We might be under some lambdas, in which case the context
     -- is bigger than the number of pattern variables.
-    let nlocals = n - size tel
-    unless (nlocals >= 0) __IMPOSSIBLE__
+    let nPatVars = size tel
+    let nlocals = n - nPatVars
+    unless (nlocals >= 0) __IMPOSSIBLE__  -- cannot be negative
 
     fv <- getDefFreeVars f
     reportSDoc "interaction.case" 20 $ do
@@ -122,13 +116,11 @@ parseVariables f tel ii rng ss = do
             "Cannot split on variable " ++ s ++
             ", because let-declarations may not be defined by pattern-matching"
 
+      let cname = C.QName $ C.Name r C.InScope $ C.stringNameParts s
+      -- Note: the range in the concrete name is only approximate.
       -- Jesper, 2018-12-19: Don't consider generalizable names since
       -- they can be shadowed by hidden variables.
-      let kinds = List.delete GeneralizeName allKindsOfNames
-          cname = C.QName $ C.Name r C.InScope $ C.stringNameParts s
-      -- Note: the range in the concrete name is only approximate.
-      resName <- resolveName' kinds Nothing cname
-      case resName of
+      resolveName' (exceptKindsOfNames [GeneralizeName]) Nothing cname >>= \case
 
         -- Fail if s is a name, but not of a variable.
         DefinedName{}       -> failNotVar
@@ -164,9 +156,14 @@ parseVariables f tel ii rng ss = do
           -- Andreas, 2018-05-28, issue #3095
           -- We want to act on an ambiguous name if it corresponds to only one local index.
           let xs'' = mapMaybe (\ (_,i) -> if i < nlocals then Nothing else Just $ i - nlocals) xs'
-          when (null xs'') $ failLocal
+          when (null xs'') $ typeError $ GenericError $
+            "Cannot make hidden lambda-bound variable " ++ s ++ " visible"
           -- Filter out variable bound by parent function or module.
-          let xs''' = mapMaybe (\ i -> if i < fv then Nothing else Just i) xs''
+          params <- moduleParamsToApply $ qnameModule f
+          let isParam i = any ((== var i) . unArg) params
+              xs'''     = filter (not . isParam) xs''
+          when (null xs''') $ typeError $ GenericError $
+            "Cannot make hidden module parameter " ++ s ++ " visible"
           case xs''' of
             []  -> failModuleBound
             [i] -> return (i , C.NotInScope)
@@ -198,10 +195,24 @@ getClauseZipperForIP f clauseNo = do
         ]
       __IMPOSSIBLE__
 
+recheckAbstractClause :: Type -> Maybe Substitution -> A.SpineClause -> TCM Clause
+recheckAbstractClause t sub cl = checkClauseLHS t sub cl $ \ lhs ->
+  return Clause{ clauseLHSRange    = getRange cl
+               , clauseFullRange   = getRange cl
+               , clauseTel         = lhsVarTele lhs
+               , namedClausePats   = lhsPatterns lhs
+               , clauseBody        = Nothing -- We don't need the body for make case
+               , clauseType        = Just (lhsBodyType lhs)
+               , clauseCatchall    = False
+               , clauseRecursive   = Nothing
+               , clauseUnreachable = Nothing
+               , clauseEllipsis    = lhsEllipsis $ A.spLhsInfo $ A.clauseLHS cl
+               }
+
 -- | Entry point for case splitting tactic.
 
 makeCase :: InteractionId -> Range -> String -> TCM (QName, CaseContext, [A.Clause])
-makeCase hole rng s = withInteractionId hole $ do
+makeCase hole rng s = withInteractionId hole $ locallyTC eMakeCase (const True) $ do
 
   -- Jesper, 2018-12-10: print unsolved metas in dot patterns as _
   localTC (\ e -> e { envPrintMetasBare = True }) $ do
@@ -209,14 +220,27 @@ makeCase hole rng s = withInteractionId hole $ do
   -- Get function clause which contains the interaction point.
   InteractionPoint { ipMeta = mm, ipClause = ipCl} <- lookupInteractionPoint hole
   let meta = fromMaybe __IMPOSSIBLE__ mm
-  (f, clauseNo, rhs) <- case ipCl of
-    IPClause f clauseNo rhs -> return (f, clauseNo, rhs)
-    IPNoClause -> typeError $ GenericError $
+  (f, clauseNo, clTy, clWithSub, absCl@A.Clause{ clauseRHS = rhs }, clClos) <- case ipCl of
+    IPClause f i t sub cl clo _ -> return (f, i, t, sub, cl, clo)
+    IPNoClause                -> typeError $ GenericError $
       "Cannot split here, as we are not in a function definition"
-  (casectxt, (prevClauses, clause, follClauses)) <- getClauseZipperForIP f clauseNo
+  (casectxt, (prevClauses0, _clause, follClauses0)) <- getClauseZipperForIP f clauseNo
+
+  -- Instead of using the actual internal clause, we retype check the abstract clause (with
+  -- eMakeCase = True). This disables the forcing translation in the unifier, which allows us to
+  -- split on forced variables.
+  clause <- enterClosure clClos $ \ _ -> locallyTC eMakeCase (const True) $
+              recheckAbstractClause clTy clWithSub absCl
+
+  let (prevClauses, follClauses) = killRange (prevClauses0, follClauses0)
+    -- Andreas, 2019-08-08, issue #3966
+    -- Kill the ranges of the existing clauses to prevent wrong error
+    -- location to be set by the coverage checker (via isCovered)
+    -- for test/interaction/Issue191
   let perm = fromMaybe __IMPOSSIBLE__ $ clausePerm clause
       tel  = clauseTel  clause
       ps   = namedClausePats clause
+      ell  = clauseEllipsis clause
   reportSDoc "interaction.case" 10 $ vcat
     [ "splitting clause:"
     , nest 2 $ vcat
@@ -225,6 +249,17 @@ makeCase hole rng s = withInteractionId hole $ do
       , "tel     =" <+> (inTopContext . prettyTCM) tel
       , "perm    =" <+> text (show perm)
       , "ps      =" <+> prettyTCMPatternList ps
+      , "ell     =" <+> text (show ell)
+      ]
+    ]
+  reportSDoc "interaction.case" 60 $ vcat
+    [ "splitting clause:"
+    , nest 2 $ vcat
+      [ "f       =" <+> (text . show) f
+      , "context =" <+> ((inTopContext . (text . show)) =<< getContextTelescope)
+      , "tel     =" <+> (text . show) tel
+      , "perm    =" <+> text (show perm)
+      , "ps      =" <+> (text . show) ps
       ]
     ]
 
@@ -236,7 +271,7 @@ makeCase hole rng s = withInteractionId hole $ do
   -- This will expand an ellipsis, if present.
 
   if concat vars == "." then do
-    cl <- makeAbstractClause f rhs $ clauseToSplitClause clause
+    cl <- makeAbstractClause f rhs NoEllipsis $ clauseToSplitClause clause
     return (f, casectxt, [cl])
 
   -- If we have no split variables, split on result.
@@ -247,7 +282,7 @@ makeCase hole rng s = withInteractionId hole $ do
     -- we need to print them postfix.
     let postProjInExtLam = applyWhen (isJust casectxt) $
           withPragmaOptions $ \ opt -> opt { optPostfixProjections = True }
-    (piTel, sc) <- fixTarget $ clauseToSplitClause clause
+    (piTel, sc) <- insertTrailingArgs $ clauseToSplitClause clause
     -- Andreas, 2015-05-05 If we introduced new function arguments
     -- do not split on result.  This might be more what the user wants.
     -- To split on result, he can then C-c C-c again.
@@ -285,39 +320,62 @@ makeCase hole rng s = withInteractionId hole $ do
           -- mapM (snd <.> fixTarget) $ splitClauses cov
           return cov
     checkClauseIsClean ipCl
-    (f, casectxt,) <$> mapM (makeAbstractClause f rhs) scs
+    (f, casectxt,) <$> mapM (makeAbstractClause f rhs ell) scs
   else do
     -- split on variables
     xs <- parseVariables f tel hole rng vars
     reportSLn "interaction.case" 30 $ "parsedVariables: " ++ show (zip xs vars)
     -- Variables that are not in scope yet are brought into scope (@toShow@)
     -- The other variables are split on (@toSplit@).
-    let (toShow, toSplit) = flip mapEither (zip xs vars) $ \ ((x,nis), s) ->
+    let (toShow, toSplit) = partitionEithers $ for (zip xs vars) $ \ ((x,nis), s) ->
           if (nis == C.NotInScope) then Left x else Right x
     let sc = makePatternVarsVisible toShow $ clauseToSplitClause clause
     scs <- split f toSplit sc
     reportSLn "interaction.case" 70 $ "makeCase: survived the splitting"
 
+    -- If any of the split variables is hidden by the ellipsis, we
+    -- should force the expansion of the ellipsis.
+    splitNames <- mapM nameOfBV toSplit
+    shouldExpandEllipsis <- return (not $ null toShow) `or2M` anyEllipsisVar f absCl splitNames
+    let ell' | shouldExpandEllipsis = NoEllipsis
+             | otherwise            = ell
+
     -- CLEAN UP OF THE GENERATED CLAUSES
-    -- 1. filter out clauses that are already covered
-    scs <- filterM (not <.> isCovered f prevClauses . fst) scs
+    -- 1. filter out the generated clauses that are already covered
+    --    we consider a generated clause already covered if it is covered by:
+    --    a. a pre-existing clause defined before the one we splitted (prevClauses)
+    --    b. a pre-existing clause defined after the one we splitted (follClauses)
+    --       under the condition that it did not cover the one we splitted but was
+    --       covered by it (i.e. it was considered unreachable).
+    -- The key idea here is:
+    --       f m    zero = ?  ---- split on m --->  f (suc m) zero = ?
+    --       f zero zero = ?                        f zero    zero = ?
+    --       f _    _    = ?                        f _       _    = ?
+    -- because [f zero zero] is already defined.
+    -- However we ignore [f _ _]: [f m zero] was already a refinement of it,
+    -- hinting that we considered it more important than the catchall.
+    let sclause = clauseToSplitClause clause
+    fcs <- filterM (\ cl -> (isCovered f [clause] (clauseToSplitClause cl)) `and2M`
+                            (not <$> isCovered f [cl] sclause))
+                   follClauses
+    scs <- filterM (not <.> isCovered f (prevClauses ++ fcs) . fst) scs
     reportSLn "interaction.case" 70 $ "makeCase: survived filtering out already covered clauses"
     -- 2. filter out trivially impossible clauses not asked for by the user
     cs <- catMaybes <$> do
      forM scs $ \ (sc, isAbsurd) -> if isAbsurd
       -- absurd clause coming from a split asked for by the user
-      then Just <$> makeAbsurdClause f sc
+      then Just <$> makeAbsurdClause f ell' sc
       -- trivially empty clause due to the refined patterns
       else
         ifM (liftTCM $ isEmptyTel (scTel sc))
           {- then -} (pure Nothing)
-          {- else -} (Just <$> makeAbstractClause f rhs sc)
+          {- else -} (Just <$> makeAbstractClause f rhs ell' sc)
     reportSLn "interaction.case" 70 $ "makeCase: survived filtering out impossible clauses"
     -- 3. If the cleanup removed everything then we know that none of the clauses where
     --    absurd but that all of them were trivially empty. In this case we rewind and
     --    insert all the clauses (garbage in, garbage out!)
     cs <- if not (null cs) then pure cs
-          else mapM (makeAbstractClause f rhs . fst) scs
+          else mapM (makeAbstractClause f rhs ell' . fst) scs
 
     reportSDoc "interaction.case" 65 $ vcat
       [ "split result:"
@@ -370,13 +428,13 @@ makePatternVarsVisible is sc@SClause{ scPats = ps } =
       -- We could introduce extra consistency checks, like
       -- if visible ai then __IMPOSSIBLE__ else
       -- or passing the parsed name along and comparing it with @x@
-      Arg (setOrigin CaseSplit ai) $ Named n $ VarP PatOSplit $ SplitPatVar x i ls
+      Arg (setOrigin CaseSplit ai) $ Named n $ VarP (PatternInfo PatOSplit []) $ SplitPatVar x i ls
   mkVis np = np
 
 -- | Make clause with no rhs (because of absurd match).
 
-makeAbsurdClause :: QName -> SplitClause -> TCM A.Clause
-makeAbsurdClause f (SClause tel sps _ _ t) = do
+makeAbsurdClause :: QName -> ExpandedEllipsis -> SplitClause -> TCM A.Clause
+makeAbsurdClause f ell (SClause tel sps _ _ t) = do
   let ps = fromSplitPatterns sps
   reportSDoc "interaction.case" 10 $ vcat
     [ "Interaction.MakeCase.makeAbsurdClause: split clause:"
@@ -384,6 +442,7 @@ makeAbsurdClause f (SClause tel sps _ _ t) = do
       [ "context =" <+> do (inTopContext . prettyTCM) =<< getContextTelescope
       , "tel     =" <+> do inTopContext $ prettyTCM tel
       , "ps      =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList ps -- P.sep <$> prettyTCMPatterns ps
+      , "ell     =" <+> text (show ell)
       ]
     ]
   withCurrentModule (qnameModule f) $ do
@@ -391,7 +450,7 @@ makeAbsurdClause f (SClause tel sps _ _ t) = do
     -- Contract implicit record patterns before printing.
     -- c <- translateRecordPatterns $ Clause noRange tel perm ps NoBody t False
     -- Jesper, 2015-09-19 Don't contract, since we do on-demand splitting
-    let c = Clause noRange noRange tel ps Nothing t False Nothing
+    let c = Clause noRange noRange tel ps Nothing (argFromDom <$> t) False Nothing Nothing ell
     -- Normalise the dot patterns
     ps <- addContext tel $ normalise $ namedClausePats c
     reportSDoc "interaction.case" 60 $ "normalized patterns: " <+> prettyTCMPatternList ps
@@ -400,13 +459,35 @@ makeAbsurdClause f (SClause tel sps _ _ t) = do
 
 -- | Make a clause with a question mark as rhs.
 
-makeAbstractClause :: QName -> A.RHS -> SplitClause -> TCM A.Clause
-makeAbstractClause f rhs cl = do
+makeAbstractClause :: QName -> A.RHS -> ExpandedEllipsis -> SplitClause -> TCM A.Clause
+makeAbstractClause f rhs ell cl = do
 
-  lhs <- A.clauseLHS <$> makeAbsurdClause f cl
+  lhs <- A.clauseLHS <$> makeAbsurdClause f ell cl
   reportSDoc "interaction.case" 60 $ "reified lhs: " <+> prettyA lhs
   return $ A.Clause lhs [] rhs A.noWhereDecls False
   -- let ii = InteractionId (-1)  -- Dummy interaction point since we never type check this.
   --                              -- Can end up in verbose output though (#1842), hence not __IMPOSSIBLE__.
   -- let info = A.emptyMetaInfo   -- metaNumber = Nothing in order to print as ?, not ?n
   -- return $ A.Clause lhs [] (A.RHS $ A.QuestionMark info ii) [] False
+
+anyEllipsisVar :: QName -> A.SpineClause -> [Name] -> TCM Bool
+anyEllipsisVar f cl xs = do
+  let lhs = A.clauseLHS cl
+      ps  = A.spLhsPats lhs
+      ell = lhsEllipsis $ A.spLhsInfo lhs
+      anyVar :: A.Pattern -> Any -> Any
+      anyVar p acc = Any $ getAny acc || case p of
+        A.VarP x -> A.unBind x `elem` xs
+        _        -> False
+  case ell of
+    NoEllipsis           -> return False
+    ExpandedEllipsis _ k -> do
+      ps' <- snd <$> reifyDisplayFormP f ps []
+      let ellipsisPats :: A.Patterns
+          ellipsisPats = fst $ C.splitEllipsis k ps'
+      reportSDoc "interaction.case.ellipsis" 40 $ vcat
+        [ "should we expand the ellipsis?"
+        , nest 2 $ "xs           =" <+> prettyList_ (map prettyA xs)
+        , nest 2 $ "ellipsisPats =" <+> prettyList_ (map prettyA ellipsisPats)
+        ]
+      return $ getAny $ A.foldrAPattern anyVar ellipsisPats
