@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 module Agda.Interaction.Options.Help
        (
@@ -8,14 +7,8 @@ module Agda.Interaction.Options.Help
        , allHelpTopics
        ) where
 
-import Data.Tuple ( swap )
-import Data.Maybe ( fromMaybe )
-
 import Agda.Interaction.Options.Warnings
 
-#include "undefined.h"
-import Agda.Utils.Impossible
-
 -- | Interface to the @help@ function
 data Help
   = GeneralHelp
@@ -46,5 +39,7 @@ helpTopicUsage tp = case tp of
 string2HelpTopic :: String -> Maybe HelpTopic
 string2HelpTopic str = lookup str allHelpTopics
 
-helpTopic2String :: HelpTopic -> String
-helpTopic2String w = fromMaybe __IMPOSSIBLE__ $ lookup w (map swap allHelpTopics)
+-- UNUSED Liang-Ting Chen 2019-07-15
+--helpTopic2String :: HelpTopic -> String
+--helpTopic2String w = fromMaybe __IMPOSSIBLE__ $ lookup w (map swap allHelpTopics)
+--
