@@ -146,12 +146,14 @@ builtinModulesWithSafePostulates =
   , "Agda" </> "Builtin" </> "Char" </> "Properties.agda"
   , "Agda" </> "Builtin" </> "Coinduction.agda"
   , "Agda" </> "Builtin" </> "Cubical" </> "Glue.agda"
+  , "Agda" </> "Builtin" </> "Cubical" </> "HCompU.agda"
   , "Agda" </> "Builtin" </> "Cubical" </> "Id.agda"
   , "Agda" </> "Builtin" </> "Cubical" </> "Path.agda"
   , "Agda" </> "Builtin" </> "Cubical" </> "Sub.agda"
   , "Agda" </> "Builtin" </> "Equality" </> "Erase.agda"
   , "Agda" </> "Builtin" </> "Equality.agda"
   , "Agda" </> "Builtin" </> "Float.agda"
+  , "Agda" </> "Builtin" </> "Float" </> "Properties.agda"
   , "Agda" </> "Builtin" </> "FromNat.agda"
   , "Agda" </> "Builtin" </> "FromNeg.agda"
   , "Agda" </> "Builtin" </> "FromString.agda"
@@ -160,6 +162,7 @@ builtinModulesWithSafePostulates =
   , "Agda" </> "Builtin" </> "List.agda"
   , "Agda" </> "Builtin" </> "Nat.agda"
   , "Agda" </> "Builtin" </> "Reflection.agda"
+  , "Agda" </> "Builtin" </> "Reflection" </> "Properties.agda"
   , "Agda" </> "Builtin" </> "Sigma.agda"
   , "Agda" </> "Builtin" </> "Size.agda"
   , "Agda" </> "Builtin" </> "Strict.agda"
