@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP                       #-}
 {-# LANGUAGE GADTs                     #-}
 {-# LANGUAGE NoMonoLocalBinds          #-}
 {-# LANGUAGE NoMonomorphismRestriction #-}
@@ -9,12 +8,8 @@ import Control.Applicative ( Const(Const), getConst )
 import Control.Arrow (first)
 import Control.Monad.Identity
 
-import Data.Foldable (foldMap)
-import Data.Monoid
-import Data.Traversable
 import Data.Void
 
-import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Abstract as A
 import Agda.Syntax.Concrete (FieldAssignment', exprFieldA)
@@ -22,7 +17,6 @@ import Agda.Syntax.Info
 import Agda.Syntax.Scope.Base (emptyScopeInfo)
 
 import Agda.Utils.Either
-import Agda.Utils.Lens
 
 data AppView' arg = Application Expr [NamedArg arg]
   deriving (Functor)
@@ -153,14 +147,12 @@ instance ExprLike Expr where
       Rec ei bs               -> Rec ei <$> recurse bs
       RecUpdate ei e bs       -> RecUpdate ei <$> recurse e <*> recurse bs
       ScopedExpr sc e         -> ScopedExpr sc <$> recurse e
-      QuoteGoal ei n e        -> QuoteGoal ei n <$> recurse e
-      QuoteContext ei         -> pure e0
       Quote{}                 -> pure e0
       QuoteTerm{}             -> pure e0
       Unquote{}               -> pure e0
       DontCare e              -> DontCare <$> recurse e
       PatternSyn{}            -> pure e0
-      Tactic ei e xs ys       -> Tactic ei <$> recurse e <*> recurse xs <*> recurse ys
+      Tactic ei e xs          -> Tactic ei <$> recurse e <*> recurse xs
       Macro{}                 -> pure e0
 
   foldExpr f e =
@@ -190,12 +182,10 @@ instance ExprLike Expr where
       Rec _ as             -> m `mappend` fold as
       RecUpdate _ e as     -> m `mappend` fold e `mappend` fold as
       ScopedExpr _ e       -> m `mappend` fold e
-      QuoteGoal _ _ e      -> m `mappend` fold e
-      QuoteContext _       -> m
       Quote{}              -> m
       QuoteTerm{}          -> m
       Unquote{}            -> m
-      Tactic _ e xs ys     -> m `mappend` fold e `mappend` fold xs `mappend` fold ys
+      Tactic _ e xs        -> m `mappend` fold e `mappend` fold xs
       DontCare e           -> m `mappend` fold e
    where
      m    = f e
@@ -227,17 +217,16 @@ instance ExprLike Expr where
       Rec ei bs               -> f =<< Rec ei <$> trav bs
       RecUpdate ei e bs       -> f =<< RecUpdate ei <$> trav e <*> trav bs
       ScopedExpr sc e         -> f =<< ScopedExpr sc <$> trav e
-      QuoteGoal ei n e        -> f =<< QuoteGoal ei n <$> trav e
-      QuoteContext{}          -> f e
       Quote{}                 -> f e
       QuoteTerm{}             -> f e
       Unquote{}               -> f e
-      Tactic ei e xs ys       -> f =<< Tactic ei <$> trav e <*> trav xs <*> trav ys
+      Tactic ei e xs          -> f =<< Tactic ei <$> trav e <*> trav xs
       DontCare e              -> f =<< DontCare <$> trav e
       PatternSyn{}            -> f e
       Macro{}                 -> f e
 
 instance ExprLike a => ExprLike (Arg a)     where
+instance ExprLike a => ExprLike (Maybe a)   where
 instance ExprLike a => ExprLike (Named x a) where
 instance ExprLike a => ExprLike [a]         where
 
@@ -263,16 +252,16 @@ instance ExprLike QName where
 instance ExprLike LamBinding where
   recurseExpr f e =
     case e of
-      DomainFree{}  -> pure e
-      DomainFull bs -> DomainFull <$> recurseExpr f bs
+      DomainFree t x -> DomainFree <$> recurseExpr f t <*> pure x
+      DomainFull bs  -> DomainFull <$> recurseExpr f bs
   foldExpr f e =
     case e of
-      DomainFree{}  -> mempty
+      DomainFree t _ -> foldExpr f t
       DomainFull bs -> foldExpr f bs
   traverseExpr f e =
     case e of
-      DomainFree{}  -> pure e
-      DomainFull bs -> DomainFull <$> traverseExpr f bs
+      DomainFree t x -> DomainFree <$> traverseExpr f t <*> pure x
+      DomainFull bs  -> DomainFull <$> traverseExpr f bs
 
 instance ExprLike GeneralizeTelescope where
   recurseExpr  f (GeneralizeTel s tel) = GeneralizeTel s <$> recurseExpr f tel
@@ -287,16 +276,16 @@ instance ExprLike DataDefParams where
 instance ExprLike TypedBinding where
   recurseExpr f e =
     case e of
-      TBind r xs e -> TBind r xs <$> recurseExpr f e
-      TLet r ds    -> TLet r <$> recurseExpr f ds
+      TBind r t xs e -> TBind r <$> recurseExpr f t <*> pure xs <*> recurseExpr f e
+      TLet r ds      -> TLet r <$> recurseExpr f ds
   foldExpr f e =
     case e of
-      TBind _ _ e  -> foldExpr f e
-      TLet _ ds    -> foldExpr f ds
+      TBind _ t _ e -> foldExpr f t `mappend` foldExpr f e
+      TLet _ ds     -> foldExpr f ds
   traverseExpr f e =
     case e of
-      TBind r xs e -> TBind r xs <$> traverseExpr f e
-      TLet r ds    -> TLet r <$> traverseExpr f ds
+      TBind r t xs e -> TBind r <$> traverseExpr f t <*> pure xs <*> traverseExpr f e
+      TLet r ds      -> TLet r <$> traverseExpr f ds
 
 instance ExprLike LetBinding where
   recurseExpr f e = do
@@ -341,6 +330,11 @@ instance ExprLike RHS where
       RewriteRHS xes spats rhs ds -> RewriteRHS <$> rec xes <*> pure spats <*> rec rhs <*> rec ds
     where rec e = recurseExpr f e
 
+instance (ExprLike qn, ExprLike p, ExprLike e) => ExprLike (RewriteEqn' qn p e) where
+  recurseExpr f = \case
+    Rewrite es    -> Rewrite <$> recurseExpr f es
+    Invert qn pes -> Invert <$> recurseExpr f qn <*> recurseExpr f pes
+
 instance ExprLike WhereDeclarations where
   recurseExpr f (WhereDecls a b) = WhereDecls a <$> recurseExpr f b
 
@@ -369,7 +363,8 @@ instance ExprLike Pragma where
 instance ExprLike LHS where
   recurseExpr f (LHS i p) = LHS i <$> recurseExpr f p
 
-instance ExprLike a => ExprLike (LHSCore' a) where
+instance ExprLike a => ExprLike (LHSCore' a)   where
+instance ExprLike a => ExprLike (WithHiding a) where
 
 instance ExprLike SpineLHS where
   recurseExpr f (SpineLHS i x ps) = SpineLHS i x <$> recurseExpr f ps
