@@ -1,24 +1,19 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE TypeFamilies #-} -- for type equality ~
 
 {-| Names in the concrete syntax are just strings (or lists of strings for
     qualified names).
 -}
 module Agda.Syntax.Concrete.Name where
 
-#if MIN_VERSION_base(4,11,0)
-import Prelude hiding ((<>))
-#endif
-
 import Control.DeepSeq
 
 import Data.ByteString.Char8 (ByteString)
 import Data.Function
+import qualified Data.Foldable as Fold
 import qualified Data.List as List
 import Data.Data (Data)
-import Data.Maybe
 
 import GHC.Generics (Generic)
 
@@ -29,12 +24,10 @@ import Agda.Syntax.Position
 
 import Agda.Utils.FileName
 import Agda.Utils.Lens
-import Agda.Utils.List
 import Agda.Utils.Pretty
 import Agda.Utils.Size
 import Agda.Utils.Suffix
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 {-| A name is a non-empty list of alternating 'Id's and 'Hole's. A normal name
@@ -299,12 +292,16 @@ qnameParts :: QName -> [Name]
 qnameParts (Qual x q) = x : qnameParts q
 qnameParts (QName x)  = [x]
 
--- | Is the name qualified?
+-- | Is the name (un)qualified?
 
 isQualified :: QName -> Bool
 isQualified Qual{}  = True
 isQualified QName{} = False
 
+isUnqualified :: QName -> Maybe Name
+isUnqualified Qual{}    = Nothing
+isUnqualified (QName n) = Just n
+
 ------------------------------------------------------------------------
 -- * Operations on 'TopLevelModuleName'
 ------------------------------------------------------------------------
@@ -365,6 +362,9 @@ noName r = NoName r (NameId 0 0)
 class IsNoName a where
   isNoName :: a -> Bool
 
+  default isNoName :: (Foldable t, IsNoName b, t b ~ a) => a -> Bool
+  isNoName = Fold.all isNoName
+
 instance IsNoName String where
   isNoName = isUnderscore
 
@@ -382,6 +382,9 @@ instance IsNoName QName where
   isNoName (QName x) = isNoName x
   isNoName Qual{}    = False        -- M.A._ does not qualify as empty name
 
+instance IsNoName a => IsNoName (Ranged a) where
+instance IsNoName a => IsNoName (WithOrigin a) where
+
 -- no instance for TopLevelModuleName
 
 ------------------------------------------------------------------------
