@@ -21,6 +21,7 @@ import Agda.Syntax.Fixity
 import Agda.Syntax.Scope.Base (ScopeInfo, emptyScopeInfo)
 
 import Agda.Utils.Function
+import Agda.Utils.Functor
 import Agda.Utils.Null
 
 {--------------------------------------------------------------------------
@@ -70,10 +71,11 @@ instance KillRange ExprInfo where
  --------------------------------------------------------------------------}
 
 -- | Information about application
-data AppInfo = AppInfo { appRange  :: Range
-                       , appOrigin :: Origin
-                       , appParens :: ParenPreference -- ^ Do we prefer a appbda argument with or without parens?
-                       }
+data AppInfo = AppInfo
+  { appRange  :: Range
+  , appOrigin :: Origin
+  , appParens :: ParenPreference -- ^ Do we prefer a appbda argument with or without parens?
+  }
   deriving (Data, Show, Eq, Ord)
 
 -- | Default is system inserted and prefer parens.
@@ -92,7 +94,7 @@ instance KillRange AppInfo where
 
 instance LensOrigin AppInfo where
   getOrigin = appOrigin
-  setOrigin o i = i { appOrigin = o }
+  mapOrigin f i = i { appOrigin = f (appOrigin i) }
 
 {--------------------------------------------------------------------------
     Module information
@@ -109,9 +111,7 @@ data ModuleInfo = ModuleInfo
   , minfoDirective :: Maybe ImportDirective
     -- ^ Retained for @abstractToConcrete@ of 'ModuleMacro'.
   }
-  deriving (Data, Eq)
-
-deriving instance Show ModuleInfo
+  deriving (Data, Eq, Show)
 
 instance HasRange ModuleInfo where
   getRange = minfoRange
@@ -139,31 +139,40 @@ instance KillRange LetInfo where
     Definition information (declarations that actually define something)
  --------------------------------------------------------------------------}
 
-data DefInfo = DefInfo
+data DefInfo' t = DefInfo
   { defFixity   :: Fixity'
   , defAccess   :: Access
   , defAbstract :: IsAbstract
   , defInstance :: IsInstance
   , defMacro    :: IsMacro
   , defInfo     :: DeclInfo
+  , defTactic   :: Maybe t
   }
   deriving (Data, Show, Eq)
 
-mkDefInfo :: Name -> Fixity' -> Access -> IsAbstract -> Range -> DefInfo
-mkDefInfo x f a ab r = DefInfo f a ab NotInstanceDef NotMacroDef (DeclInfo x r)
+mkDefInfo :: Name -> Fixity' -> Access -> IsAbstract -> Range -> DefInfo' t
+mkDefInfo x f a ab r = mkDefInfoInstance x f a ab NotInstanceDef NotMacroDef r
 
 -- | Same as @mkDefInfo@ but where we can also give the @IsInstance@
-mkDefInfoInstance :: Name -> Fixity' -> Access -> IsAbstract -> IsInstance -> IsMacro -> Range -> DefInfo
-mkDefInfoInstance x f a ab i m r = DefInfo f a ab i m (DeclInfo x r)
+mkDefInfoInstance :: Name -> Fixity' -> Access -> IsAbstract -> IsInstance -> IsMacro -> Range -> DefInfo' t
+mkDefInfoInstance x f a ab i m r = DefInfo f a ab i m (DeclInfo x r) Nothing
 
-instance HasRange DefInfo where
+instance HasRange (DefInfo' t) where
   getRange = getRange . defInfo
 
-instance SetRange DefInfo where
+instance SetRange (DefInfo' t) where
   setRange r i = i { defInfo = setRange r (defInfo i) }
 
-instance KillRange DefInfo where
-  killRange i = i { defInfo = killRange $ defInfo i }
+instance KillRange t => KillRange (DefInfo' t) where
+  killRange i = i { defInfo   = killRange $ defInfo i,
+                    defTactic = killRange $ defTactic i }
+
+instance LensIsAbstract (DefInfo' t) where
+  lensIsAbstract f i = (f $! defAbstract i) <&> \ a -> i { defAbstract = a }
+
+instance AnyIsAbstract (DefInfo' t) where
+  anyIsAbstract = defAbstract
+
 
 {--------------------------------------------------------------------------
     General declaration information
@@ -189,15 +198,16 @@ instance KillRange DeclInfo where
  --------------------------------------------------------------------------}
 
 data MutualInfo = MutualInfo
-  { mutualTermCheck       :: TerminationCheck Name
-  , mutualPositivityCheck :: PositivityCheck
-  , mutualRange           :: Range
+  { mutualTerminationCheck :: TerminationCheck Name
+  , mutualCoverageCheck    :: CoverageCheck
+  , mutualPositivityCheck  :: PositivityCheck
+  , mutualRange            :: Range
   }
   deriving (Data, Show, Eq)
 
 -- | Default value for 'MutualInfo'.
 instance Null MutualInfo where
-  empty = MutualInfo TerminationCheck True noRange
+  empty = MutualInfo TerminationCheck YesCoverageCheck YesPositivityCheck noRange
 
 instance HasRange MutualInfo where
   getRange = mutualRange
@@ -209,14 +219,20 @@ instance KillRange MutualInfo where
     Left hand side information
  --------------------------------------------------------------------------}
 
-newtype LHSInfo = LHSRange Range
-  deriving (Data, Show, Eq, Null)
+data LHSInfo = LHSInfo
+  { lhsRange    :: Range
+  , lhsEllipsis :: ExpandedEllipsis
+  } deriving (Data, Show, Eq)
 
 instance HasRange LHSInfo where
-  getRange (LHSRange r) = r
+  getRange (LHSInfo r _) = r
 
 instance KillRange LHSInfo where
-  killRange (LHSRange r) = LHSRange noRange
+  killRange (LHSInfo r ell) = LHSInfo noRange ell
+
+instance Null LHSInfo where
+  null i = null (lhsRange i) && null (lhsEllipsis i)
+  empty  = LHSInfo empty empty
 
 {--------------------------------------------------------------------------
     Pattern information
@@ -233,21 +249,16 @@ patNoRange = PatRange noRange
 
 -- | Constructor pattern info.
 data ConPatInfo = ConPatInfo
-  { patOrigin   :: ConOrigin
+  { conPatOrigin   :: ConOrigin
     -- ^ Does this pattern come form the eta-expansion of an implicit pattern?
     ---  Or from a user written constructor or record pattern?
-  , patInfo     :: PatInfo
-  , patLazy     :: ConPatLazy
+  , conPatInfo     :: PatInfo
+  , conPatLazy     :: ConPatLazy
   }
-  deriving (Data, Eq)
-
-instance Show ConPatInfo where
-  show (ConPatInfo po i l) =
-    applyWhen (l  == ConPatLazy) ("lazy " ++) $
-    applyWhen (po == ConOSystem) ("implicit " ++) $ show i
+  deriving (Data, Eq, Show)
 
 instance HasRange ConPatInfo where
-  getRange = getRange . patInfo
+  getRange = getRange . conPatInfo
 
 instance KillRange ConPatInfo where
   killRange (ConPatInfo b i l) = ConPatInfo b (killRange i) l
