@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE ViewPatterns #-}
 
 -- | Preprocessors for literate code formats.
@@ -23,13 +22,12 @@ module Agda.Syntax.Parser.Literate
   where
 
 import Prelude hiding (getLine)
-import Data.Char (isSpace, isControl)
+import Data.Char (isSpace)
 import Data.List (isPrefixOf)
 import Agda.Syntax.Common
 import Agda.Syntax.Position
 import Text.Regex.TDFA
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Role of a character in the file.
