@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable    #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
@@ -32,32 +31,24 @@ module Agda.Syntax.Parser.Monad
     )
     where
 
-#if MIN_VERSION_base(4,11,0)
-import Prelude hiding ((<>))
-#endif
-
-import Control.Exception (catch, displayException)
+import Control.Exception (displayException)
 import Data.Int
 
 import Data.Data (Data)
-import qualified Data.Text.Lazy as T
 
 import Control.Monad.State
-import Control.Monad.Except
 
 import Agda.Interaction.Options.Warnings
 
 import Agda.Syntax.Position
 
-import Agda.Utils.Except ( MonadError(catchError, throwError) )
+import Agda.Utils.Except ( MonadError(throwError) )
 import Agda.Utils.FileName
 import Agda.Utils.List ( tailWithDefault )
-import qualified Agda.Utils.IO.UTF8 as UTF8
 import qualified Agda.Utils.Maybe.Strict as Strict
 
 import Agda.Utils.Pretty
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 {--------------------------------------------------------------------------
@@ -65,7 +56,7 @@ import Agda.Utils.Impossible
  --------------------------------------------------------------------------}
 
 -- | The parse monad.
-newtype Parser a = P { runP :: StateT ParseState (Either ParseError) a }
+newtype Parser a = P { _runP :: StateT ParseState (Either ParseError) a }
   deriving (Functor, Applicative, Monad, MonadState ParseState, MonadError ParseError)
 
 -- | The parser state. Contains everything the parser and the lexer could ever
@@ -184,17 +175,17 @@ instance Show ParseError where
 
 instance Pretty ParseError where
   pretty ParseError{errPos,errSrcFile,errMsg,errPrevToken,errInput} = vcat
-      [ pretty (errPos { srcFile = errSrcFile }) <> colon <+>
+      [ (pretty (errPos { srcFile = errSrcFile }) <> colon) <+>
         text errMsg
       , text $ errPrevToken ++ "<ERROR>"
       , text $ take 30 errInput ++ "..."
       ]
   pretty OverlappingTokensError{errRange} = vcat
-      [ pretty errRange <> colon <+>
+      [ (pretty errRange <> colon) <+>
         "Multi-line comment spans one or more literate text blocks."
       ]
   pretty InvalidExtensionError{errPath,errValidExts} = vcat
-      [ pretty errPath <> colon <+>
+      [ (pretty errPath <> colon) <+>
         "Unsupported extension."
       , "Supported extensions are:" <+> prettyList_ errValidExts
       ]
@@ -218,7 +209,7 @@ instance Show ParseWarning where
 
 instance Pretty ParseWarning where
   pretty OverlappingTokensWarning{warnRange} = vcat
-      [ pretty warnRange <> colon <+>
+      [ (pretty warnRange <> colon) <+>
         "Multi-line comment spans one or more literate text blocks."
       ]
 instance HasRange ParseWarning where
@@ -297,8 +288,9 @@ getParseInterval = do
 getLexState :: Parser [LexState]
 getLexState = parseLexState <$> get
 
-setLexState :: [LexState] -> Parser ()
-setLexState ls = modify $ \ s -> s { parseLexState = ls }
+-- UNUSED Liang-Ting Chen 2019-07-16
+--setLexState :: [LexState] -> Parser ()
+--setLexState ls = modify $ \ s -> s { parseLexState = ls }
 
 modifyLexState :: ([LexState] -> [LexState]) -> Parser ()
 modifyLexState f = modify $ \ s -> s { parseLexState = f (parseLexState s) }
