@@ -6,8 +6,6 @@ module Agda.Syntax.Translation.ReflectedToAbstract where
 
 import Control.Monad.Reader
 
-import Data.Traversable as Trav hiding (mapM)
-
 import Agda.Syntax.Fixity
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
@@ -16,23 +14,32 @@ import Agda.Syntax.Common
 import Agda.Syntax.Abstract as A hiding (Apply)
 import Agda.Syntax.Abstract.Pattern
 import Agda.Syntax.Reflected as R
+import Agda.Syntax.Internal (Dom,Dom'(..))
 
 import Agda.TypeChecking.Monad as M hiding (MetaInfo)
 import Agda.Syntax.Scope.Monad (getCurrentModule)
 
-import Agda.Utils.Maybe
+import Agda.Utils.Except
 import Agda.Utils.Monad
 import Agda.Utils.List
+import Agda.Utils.Null
 import Agda.Utils.Functor
 import Agda.Utils.Size
 
 type Names = [Name]
 
-type WithNames a = ReaderT Names TCM a
--- Note: we only need the TCM for fresh names
+type MonadReflectedToAbstract m =
+  ( MonadReader Names m
+  , MonadFresh NameId m
+  , MonadError TCErr m
+  , MonadTCEnv m
+  , ReadTCState m
+  , HasOptions m
+  , HasConstInfo m
+  )
 
 -- | Adds a new unique name to the current context.
-withName :: String -> (Name -> WithNames a) -> WithNames a
+withName :: MonadReflectedToAbstract m => String -> (Name -> m a) -> m a
 withName s f = do
   name <- freshName_ s
   ctx  <- asks $ map nameConcrete
@@ -42,18 +49,34 @@ withName s f = do
     notTaken xs x = isNoName x || nameConcrete x `notElem` xs
 
 -- | Returns the name of the variable with the given de Bruijn index.
-askName :: Int -> WithNames (Maybe Name)
+askName :: MonadReflectedToAbstract m => Int -> m (Maybe Name)
 askName i = reader (!!! i)
 
 class ToAbstract r a | r -> a where
-  toAbstract :: r -> WithNames a
+  toAbstract :: MonadReflectedToAbstract m => r -> m a
 
 -- | Translate reflected syntax to abstract, using the names from the current typechecking context.
-toAbstract_ :: ToAbstract r a => r -> TCM a
+toAbstract_ ::
+  (ToAbstract r a
+  , MonadFresh NameId m
+  , MonadError TCErr m
+  , MonadTCEnv m
+  , ReadTCState m
+  , HasOptions m
+  , HasConstInfo m
+  ) => r -> m a
 toAbstract_ = withShowAllArguments . toAbstractWithoutImplicit
 
 -- | Drop implicit arguments unless --show-implicit is on.
-toAbstractWithoutImplicit :: ToAbstract r a => r -> TCM a
+toAbstractWithoutImplicit ::
+  (ToAbstract r a
+  , MonadFresh NameId m
+  , MonadError TCErr m
+  , MonadTCEnv m
+  , ReadTCState m
+  , HasOptions m
+  , HasConstInfo m
+  ) => r -> m a
 toAbstractWithoutImplicit x = runReaderT (toAbstract x) =<< getContextNames
 
 instance ToAbstract r a => ToAbstract (Named name r) (Named name a) where
@@ -66,14 +89,14 @@ instance ToAbstract [Arg Term] [NamedArg Expr] where
   toAbstract = traverse toAbstract
 
 instance ToAbstract r Expr => ToAbstract (Dom r, Name) (A.TypedBinding) where
-  toAbstract (Dom{domInfo = i,unDom = x}, name) = do
+  toAbstract (Dom{domInfo = i,unDom = x, domTactic = tac}, name) = do
     dom <- toAbstract x
-    return $ TBind noRange [unnamedArg i $ BindName name] dom
+    return $ mkTBind noRange [unnamedArg i $ mkBinder_ name] dom
 
 instance ToAbstract (Expr, Elim) Expr where
   toAbstract (f, Apply arg) = do
     arg     <- toAbstract arg
-    showImp <- lift showImplicitArguments
+    showImp <- showImplicitArguments
     return $ if showImp || visible arg
              then App (setOrigin Reflected defaultAppInfo_) f arg
              else f
@@ -94,21 +117,21 @@ instance ToAbstract Term Expr where
       mname <- askName i
       case mname of
         Nothing -> do
-          cxt   <- lift $ getContextTelescope
+          cxt   <- getContextTelescope
           names <- asks $ drop (size cxt) . reverse
-          lift $ withShowAllArguments' False $ typeError $ DeBruijnIndexOutOfScope i cxt names
+          withShowAllArguments' False $ typeError $ DeBruijnIndexOutOfScope i cxt names
         Just name -> toAbstract (A.Var name, es)
     R.Con c es -> toAbstract (A.Con (unambiguous $ killRange c), es)
     R.Def f es -> do
-      af <- lift $ mkDef (killRange f)
+      af <- mkDef (killRange f)
       toAbstract (af, es)
     R.Lam h t  -> do
       (e, name) <- toAbstract t
       let info  = setHiding h $ setOrigin Reflected defaultArgInfo
-      return $ A.Lam exprNoRange (DomainFree $ unnamedArg info $ BindName name) e
+      return $ A.Lam exprNoRange (mkDomainFree $ unnamedArg info $ mkBinder_ name) e
     R.ExtLam cs es -> do
       name <- freshName_ extendedLambdaName
-      m    <- lift $ getCurrentModule
+      m    <- getCurrentModule
       let qname   = qualify m name
           cname   = nameConcrete name
           defInfo = mkDefInfo cname noFixity' PublicAccess ConcreteDef noRange
@@ -124,7 +147,7 @@ instance ToAbstract Term Expr where
       where info = emptyMetaInfo{ metaNumber = Just x }
     R.Unknown      -> return $ Underscore emptyMetaInfo
 
-mkDef :: QName -> TCM A.Expr
+mkDef :: HasConstInfo m => QName -> m A.Expr
 mkDef f =
   ifM (isMacro . theDef <$> getConstInfo f)
       (return $ A.Macro f)
@@ -144,16 +167,16 @@ instance ToAbstract R.Pattern (Names, A.Pattern) where
       (names, args) <- toAbstractPats args
       return (names, A.ConP (ConPatInfo ConOCon patNoRange ConPatEager) (unambiguous $ killRange c) args)
     R.DotP    -> return ([], A.WildP patNoRange)
-    R.VarP s | isNoName s -> withName "z" $ \ name -> return ([name], A.VarP $ BindName name)
+    R.VarP s | isNoName s -> withName "z" $ \ name -> return ([name], A.VarP $ mkBindName name)
         -- Ulf, 2016-08-09: Also bind noNames (#2129). This to make the
         -- behaviour consistent with lambda and pi.
         -- return ([], A.WildP patNoRange)
-    R.VarP s  -> withName s $ \ name -> return ([name], A.VarP $ BindName name)
+    R.VarP s  -> withName s $ \ name -> return ([name], A.VarP $ mkBindName name)
     R.LitP l  -> return ([], A.LitP l)
     R.AbsurdP -> return ([], A.AbsurdP patNoRange)
     R.ProjP d -> return ([], A.ProjP patNoRange ProjSystem $ unambiguous $ killRange d)
 
-toAbstractPats :: [Arg R.Pattern] -> WithNames (Names, [NamedArg A.Pattern])
+toAbstractPats :: MonadReflectedToAbstract m => [Arg R.Pattern] -> m (Names, [NamedArg A.Pattern])
 toAbstractPats pats = case pats of
     []   -> return ([], [])
     p:ps -> do
@@ -165,11 +188,11 @@ instance ToAbstract (QNamed R.Clause) A.Clause where
   toAbstract (QNamed name (R.Clause pats rhs)) = do
     (names, pats) <- toAbstractPats pats
     rhs           <- local (names++) $ toAbstract rhs
-    let lhs = spineToLhs $ SpineLHS (LHSRange noRange) name pats
+    let lhs = spineToLhs $ SpineLHS empty name pats
     return $ A.Clause lhs [] (RHS rhs Nothing) noWhereDecls False
   toAbstract (QNamed name (R.AbsurdClause pats)) = do
     (_, pats) <- toAbstractPats pats
-    let lhs = spineToLhs $ SpineLHS (LHSRange noRange) name pats
+    let lhs = spineToLhs $ SpineLHS empty name pats
     return $ A.Clause lhs [] AbsurdRHS noWhereDecls False
 
 instance ToAbstract [QNamed R.Clause] [A.Clause] where
