@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 {- | Sparse matrices.
 
@@ -52,7 +51,7 @@ import Data.Array
 import Data.Function
 import qualified Data.List as List
 import Data.Maybe
-import Data.Monoid
+
 
 import Data.Foldable (Foldable)
 import qualified Data.Foldable as Fold
@@ -63,15 +62,14 @@ import qualified Text.PrettyPrint.Boxes as Boxes
 import Agda.Termination.Semiring (HasZero(..), Semiring)
 import qualified Agda.Termination.Semiring as Semiring
 
-import Agda.Utils.Functor
+
 import Agda.Utils.List
 import Agda.Utils.Maybe
-import Agda.Utils.Monad
+
 import Agda.Utils.PartialOrd
 import Agda.Utils.Pretty hiding (isEmpty)
 import Agda.Utils.Tuple
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 ------------------------------------------------------------------------
@@ -94,11 +92,12 @@ data MIx i = MIx
   }
   deriving (Eq, Ord, Show, Ix)
 
--- | Convert a 'Size' to a set of bounds suitable for use with
---   the matrices in this module.
-
-toBounds :: Num i => Size i -> (MIx i, MIx i)
-toBounds sz = (MIx { row = 1, col = 1 }, MIx { row = rows sz, col = cols sz })
+-- UNUSED Liang-Ting Chen 2019-07-15
+---- | Convert a 'Size' to a set of bounds suitable for use with
+----   the matrices in this module.
+--
+--toBounds :: Num i => Size i -> (MIx i, MIx i)
+--toBounds sz = (MIx { row = 1, col = 1 }, MIx { row = rows sz, col = cols sz })
 
 -- | Type of matrices, parameterised on the type of values.
 --
@@ -191,16 +190,17 @@ blowUpSparseVec zero n l = aux 1 l
           | i == j         = b    : aux (i + 1) l'
           | otherwise      = zero : aux (i + 1) l
 
--- Older implementation without replicate.
-blowUpSparseVec' :: (Ord i, Num i, Enum i) => b -> i -> [(i,b)] -> [b]
-blowUpSparseVec' zero n l = aux 1 l
-  where aux i [] | i > n = []
-                 | otherwise = zero : aux (i+1) []
-        aux i ((j,b):l) | i <= n && j == i = b : aux (succ i) l
-        aux i ((j,b):l) | i <= n && j >= i = zero : aux (succ i) ((j,b):l)
-        aux i l = __IMPOSSIBLE__
-          -- error $ "blowUpSparseVec (n = " ++ show n ++ ") aux i=" ++ show i ++ " j=" ++ show (fst (head l)) ++ " length l = " ++ show (length l)
-
+-- UNUSED Liang-Ting Chen 2019-07-15
+---- Older implementation without replicate.
+--blowUpSparseVec' :: (Ord i, Num i, Enum i) => b -> i -> [(i,b)] -> [b]
+--blowUpSparseVec' zero n l = aux 1 l
+--  where aux i [] | i > n = []
+--                 | otherwise = zero : aux (i+1) []
+--        aux i ((j,b):l) | i <= n && j == i = b : aux (succ i) l
+--        aux i ((j,b):l) | i <= n && j >= i = zero : aux (succ i) ((j,b):l)
+--        aux i l = __IMPOSSIBLE__
+--          -- error $ "blowUpSparseVec (n = " ++ show n ++ ") aux i=" ++ show i ++ " j=" ++ show (fst (head l)) ++ " length l = " ++ show (length l)
+--
 -- | Converts a matrix to a list of row lists.
 --   @O(size)@ where @size = rows × cols@.
 
