@@ -1,5 +1,5 @@
 module Agda.TheTypeChecker
-  ( checkDecls, checkDecl
+  ( checkDecls, checkDecl, checkDeclCached
   , inferExpr, checkExpr
   ) where
 
