@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP                #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 
 -- | Case trees.
@@ -14,9 +13,9 @@ import Prelude hiding (null)
 import qualified Data.Map as Map
 import Data.Map (Map)
 import Data.Semigroup hiding (Arg(..))
-import Data.Foldable hiding (null)
-import Data.Traversable
-import Data.Maybe
+
+
+
 
 import Data.Data (Data)
 
@@ -27,9 +26,8 @@ import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
 import Agda.Utils.Null
-import Agda.Utils.Pretty hiding ((<>))
+import Agda.Utils.Pretty
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 data WithArity c = WithArity { arity :: Int, content :: c }
