@@ -1,22 +1,16 @@
-{-# LANGUAGE CPP           #-}
 
 module Agda.TypeChecking.CompiledClause.Compile where
 
 import Prelude hiding (null)
 
 import Control.Applicative
-import Control.Arrow (first, second)
 import Control.Monad
+import Control.Monad.Trans.Identity
 
 import Data.Maybe
-import Data.Monoid
 import qualified Data.Map as Map
-import Data.List (nubBy, findIndex)
+import Data.List (nubBy)
 import Data.Function
-import qualified Data.IntSet as IntSet
-import Data.Traversable (traverse)
-
-import Debug.Trace
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -29,17 +23,15 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.RecordPatterns
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.Free
 import Agda.TypeChecking.Free.Precompute
 import Agda.TypeChecking.Reduce
 
 import Agda.Utils.Functor
 import Agda.Utils.Maybe
-import Agda.Utils.Null
 import Agda.Utils.List
 import qualified Agda.Utils.Pretty as P
+import Agda.Utils.Update
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 data RunRecordPatternTranslation = RunRecordPatternTranslation | DontRunRecordPatternTranslation
@@ -47,16 +39,12 @@ data RunRecordPatternTranslation = RunRecordPatternTranslation | DontRunRecordPa
 
 compileClauses' :: RunRecordPatternTranslation -> [Clause] -> Maybe SplitTree -> TCM CompiledClauses
 compileClauses' recpat cs mSplitTree = do
-  -- Apply forcing translation. This only shuffles the deBruijn variables
-  -- so doesn't affect the right hand side.
-  cs <- sequence [ forcingTranslation ps <&> \ qs -> c{ namedClausePats = qs }
-                 | c@Clause{ namedClausePats = ps } <- cs ]
 
   -- Throw away the unreachable clauses (#2723).
   let notUnreachable = (Just True /=) . clauseUnreachable
   cs <- map unBruijn <$> normaliseProjP (filter notUnreachable cs)
 
-  let translate | recpat == RunRecordPatternTranslation = translateCompiledClauses
+  let translate | recpat == RunRecordPatternTranslation = runIdentityT . translateCompiledClauses
                 | otherwise                             = return
 
   translate $ caseMaybe mSplitTree (compile cs) $ \splitTree ->
@@ -71,17 +59,19 @@ compileClauses' recpat cs mSplitTree = do
 --   Phases 1. and 2. are skipped if @Nothing@.
 compileClauses ::
   Maybe (QName, Type) -- ^ Translate record patterns and coverage check with given type?
-  -> [Clause] -> TCM (Maybe SplitTree, CompiledClauses)
+  -> [Clause]
+  -> TCM (Maybe SplitTree, Bool, CompiledClauses)
+     -- ^ The 'Bool' indicates whether we turned a record expression into a copattern match.
 compileClauses mt cs = do
   -- Construct clauses with pattern variables bound in left-to-right order.
   -- Discard de Bruijn indices in patterns.
   case mt of
-    Nothing -> (Nothing,) . compile . map unBruijn <$> normaliseProjP cs
+    Nothing -> (Nothing,False,) . compile . map unBruijn <$> normaliseProjP cs
     Just (q, t)  -> do
       splitTree <- coverageCheck q t cs
 
       reportSDoc "tc.cc.tree" 20 $ vcat
-        [ "split tree from coverage check "
+        [ "split tree of " <+> prettyTCM q <+> " from coverage check "
         , return $ P.pretty splitTree
         ]
 
@@ -93,21 +83,21 @@ compileClauses mt cs = do
       let cls = map unBruijn cs
 
       reportSDoc "tc.cc" 30 $ sep $ do
-        "clauses patterns  before compilation" : do
+        ("clauses patterns of " <+> prettyTCM q <+> " before compilation") : do
           map (prettyTCM . map unArg . clPats) cls
       reportSDoc "tc.cc" 50 $
-        "clauses before compilation" <?> pretty cs
+        "clauses of " <+> prettyTCM q <+> " before compilation" <?> pretty cs
       let cc = compileWithSplitTree splitTree cls
       reportSDoc "tc.cc" 20 $ sep
-        [ "compiled clauses (still containing record splits)"
+        [ "compiled clauses of " <+> prettyTCM q <+> " (still containing record splits)"
         , nest 2 $ return $ P.pretty cc
         ]
-      cc <- translateCompiledClauses cc
+      (cc, becameCopatternLHS) <- runChangeT $ translateCompiledClauses cc
       reportSDoc "tc.cc" 12 $ sep
-        [ "compiled clauses"
+        [ "compiled clauses of " <+> prettyTCM q
         , nest 2 $ return $ P.pretty cc
         ]
-      return (Just splitTree, fmap precomputeFreeVars_ cc)
+      return (Just splitTree, becameCopatternLHS, fmap precomputeFreeVars_ cc)
 
 -- | Stripped-down version of 'Agda.Syntax.Internal.Clause'
 --   used in clause compiler.
@@ -132,7 +122,7 @@ unBruijn c = Cl (applySubst sub $ (map . fmap) (fmap dbPatVarName . namedThing)
 
 compileWithSplitTree :: SplitTree -> Cls -> CompiledClauses
 compileWithSplitTree t cs = case t of
-  SplitAt i ts -> Case i $ compiles ts $ splitOn (length ts == 1) (unArg i) cs
+  SplitAt i lz ts -> Case i $ compiles lz ts $ splitOn (length ts == 1) (unArg i) cs
         -- if there is just one case, we force expansion of catch-alls
         -- this is needed to generate a sound tree on which we can
         -- collapse record pattern splits
@@ -140,18 +130,20 @@ compileWithSplitTree t cs = case t of
     -- after end of split tree, continue with left-to-right strategy
 
   where
-    compiles :: SplitTrees -> Case Cls -> Case CompiledClauses
-    compiles ts br@Branches{ projPatterns = cop
-                           , conBranches = cons
-                           , etaBranch   = Nothing
-                           , litBranches = lits
-                           , fallThrough = fT
-                           , catchAllBranch = catchAll }
+    compiles :: LazySplit -> SplitTrees -> Case Cls -> Case CompiledClauses
+    compiles lz ts br@Branches{ projPatterns = cop
+                              , conBranches = cons
+                              , etaBranch   = Nothing
+                              , litBranches = lits
+                              , fallThrough = fT
+                              , catchAllBranch = catchAll
+                              , lazyMatch = lazy }
       = br{ conBranches    = updCons cons
           , etaBranch      = Nothing
           , litBranches    = updLits lits
           , fallThrough    = fT
           , catchAllBranch = updCatchall catchAll
+          , lazyMatch      = lazy || lz == LazySplit
           }
       where
         updCons = Map.mapWithKey $ \ c cl ->
@@ -160,7 +152,7 @@ compileWithSplitTree t cs = case t of
         updLits = Map.mapWithKey $ \ l cl ->
           caseMaybe (lookup (SplitLit l) ts) compile compileWithSplitTree cl
         updCatchall = fmap $ caseMaybe (lookup SplitCatchall ts) compile compileWithSplitTree
-    compiles _ Branches{etaBranch = Just{}} = __IMPOSSIBLE__  -- we haven't inserted eta matches yet
+    compiles _ _ Branches{etaBranch = Just{}} = __IMPOSSIBLE__  -- we haven't inserted eta matches yet
 
 compile :: Cls -> CompiledClauses
 compile [] = Fail
@@ -193,7 +185,7 @@ nextSplit (Cl ps _ : cs) = findSplit nonLazy ps <|> findSplit allAgree ps
     nonLazy _ (ConP _ cpi _) = not $ conPLazy cpi
     nonLazy _ _              = True
 
-    findSplit okPat ps = headMaybe (catMaybes $
+    findSplit okPat ps = listToMaybe (catMaybes $
       zipWith (\ (Arg ai p) n -> (, Arg ai n) <$> properSplit p <* guard (okPat n p)) ps [0..])
 
     allAgree i (ConP c _ _) = all ((== Just (conName c)) . getCon . map unArg . drop i . clPats) cs
@@ -205,7 +197,7 @@ nextSplit (Cl ps _ : cs) = findSplit nonLazy ps <|> findSplit allAgree ps
 -- | Is is not a variable pattern?
 --   And if yes, is it a record pattern and/or a fallThrough one?
 properSplit :: Pattern' a -> Maybe Bool
-properSplit (ConP _ cpi _) = Just (Just PatORec == conPRecord cpi || conPFallThrough cpi)
+properSplit (ConP _ cpi _) = Just ((conPRecord cpi && patOrigin (conPInfo cpi) == PatORec) || conPFallThrough cpi)
 properSplit DefP{}    = Just False
 properSplit LitP{}    = Just False
 properSplit ProjP{}   = Just False
@@ -240,12 +232,12 @@ splitC n (Cl ps b) = caseMaybe mp fallback $ \case
                    Cl (ps0 ++ map (fmap namedThing) qs ++ ps1) b) { lazyMatch = conPLazy i }
   DefP o q qs -> (conCase q False $ WithArity (length qs) $
                    Cl (ps0 ++ map (fmap namedThing) qs ++ ps1) b) { lazyMatch = False }
-  LitP l      -> litCase l $ Cl (ps0 ++ ps1) b
+  LitP _ l    -> litCase l $ Cl (ps0 ++ ps1) b
   VarP{}      -> fallback
   DotP{}      -> fallback
   where
     (ps0, rest) = splitAt n ps
-    mp          = unArg <$> headMaybe rest
+    mp          = unArg <$> listToMaybe rest
     ps1         = drop 1 rest
     fallback    = catchAll $ Cl ps b
 
@@ -343,14 +335,14 @@ expandCatchAlls single n cs =
     -- True if nth pattern exists and is variable.
     exCatchAllNth ps = any (isVar . unArg) $ take 1 $ drop n ps
 
-    classify (LitP l)     = Left l
+    classify (LitP _ l)   = Left l
     classify (ConP c _ _) = Right (Left c)
     classify (DefP _ q _) = Right (Right q)
     classify _            = __IMPOSSIBLE__
 
     -- All non-catch-all patterns following this one (at position n).
     -- These are the cases the wildcard needs to be expanded into.
-    expansions = nubBy ((==) `on` (classify . unArg . snd))
+    expansions = nubOn (classify . unArg . snd)
                . mapMaybe (notVarNth . clPats)
                $ cs
     notVarNth
@@ -359,7 +351,7 @@ expandCatchAlls single n cs =
                , Arg Pattern)  -- @n+1@st pattern, not a variable
     notVarNth ps = do
       let (ps1, ps2) = splitAt n ps
-      p <- headMaybe ps2
+      p <- listToMaybe ps2
       guard $ not $ isVar $ unArg p
       return (ps1, p)
 
@@ -374,7 +366,7 @@ expandCatchAlls single n cs =
             -- TODO Andrea: might need these to sometimes be IApply?
             conPArgs = map (fmap ($> varP "_")) qs'
             conArgs  = zipWith (\ q' i -> q' $> var i) qs' $ downFrom m
-        LitP l -> Cl (ps0 ++ [q $> LitP l] ++ ps1) (substBody n' 0 (Lit l) b)
+        LitP i l -> Cl (ps0 ++ [q $> LitP i l] ++ ps1) (substBody n' 0 (Lit l) b)
         DefP o d qs' -> Cl (ps0 ++ [q $> DefP o d conPArgs] ++ ps1)
                             (substBody n' m (Def d (map Apply conArgs)) b)
           where
