@@ -1,7 +1,17 @@
-{-# LANGUAGE CPP           #-}
+{-| Given
+
+    1. the function clauses @cs@
+    2. the patterns @ps@ of the split clause
+
+we want to compute a variable index (in the split clause) to split on next.
+
+The matcher here checks whether the split clause is covered by one of
+the given clauses @cs@ or whether further splitting is needed (and
+when yes, where).
+-}
 
 module Agda.TypeChecking.Coverage.Match
-  ( Match(..), match
+  ( Match(..), match, matchClause
   , SplitPattern, SplitPatVar(..), fromSplitPatterns, toSplitPatterns
   , toSplitPSubst, applySplitPSubst
   , isTrivialPattern
@@ -15,15 +25,12 @@ import Control.Monad.State
 import Prelude hiding ( null )
 
 import qualified Data.List as List
-import Data.Maybe (mapMaybe, isJust, fromMaybe)
-import Data.Monoid ( Monoid, mempty, mappend, mconcat )
-import Data.Semigroup ( Semigroup, (<>), Any(..) )
-import Data.Traversable (traverse)
+import Data.Maybe (mapMaybe, fromMaybe)
+import Data.Semigroup ( Semigroup, (<>))
 
 import Agda.Syntax.Abstract (IsProjP(..))
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
-import Agda.Syntax.Internal.Pattern ()
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
@@ -35,41 +42,81 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Null
-import Agda.Utils.Permutation
 import Agda.Utils.Pretty ( Pretty(..), text, (<+>), cat , prettyList_ )
-import qualified Agda.Utils.Pretty as P
-import Agda.Utils.Size
-import Agda.Utils.List
 import Agda.Utils.Monad
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
-{-| Given
+-- | If matching is inconclusive (@Block@) we want to know which
+--   variables or projections are blocking the match.
+data Match a
+  = Yes a   -- ^ Matches unconditionally.
+  | No      -- ^ Definitely does not match.
+  | Block
+    { blockedOnResult :: BlockedOnResult
+      -- ^ @BlockedOnProj o@ if the clause has a result split.
+    , blockedOnVars :: BlockingVars
+      -- ^ @BlockingVar i cs ls o@ means variable @i@ is blocked on
+      -- constructors @cs@ and literals @ls@.
+    }
+  deriving (Functor)
 
-    1. the function clauses @cs@
-    2. the patterns @ps@
+-- | Missing elimination blocking a match.
+data BlockedOnResult
+  = BlockedOnProj      -- ^ Blocked on unsplit projection.
+     { blockedOnResultOverlap :: Bool
+       -- ^ True if there are also matching clauses without an unsplit
+       -- copattern.
+     }
+  | BlockedOnApply     -- ^ Blocked on unintroduced argument.
+     { blockedOnResultIApply :: ApplyOrIApply
+       -- ^ Is the unintroduced argument an 'IApply' pattern?
+     }
+  | NotBlockedOnResult
 
-we want to compute a variable index of the split clause to split on next.
+data ApplyOrIApply = IsApply | IsIApply
 
-First, we find the set @cs'@ of all the clauses that are
-instances (via substitutions @rhos@) of the split clause.
+-- | Variable blocking a match.
+data BlockingVar = BlockingVar
+  { blockingVarNo  :: Nat
+    -- ^ De Bruijn index of variable blocking the match.
+  , blockingVarCons :: [ConHead]
+    -- ^ Constructors in this position.
+  , blockingVarLits :: [Literal]
+    -- ^ Literals in this position.
+  , blockingVarOverlap :: Bool
+    -- ^ True if at least one clause has a variable pattern in this
+    --   position.
+  , blockingVarLazy :: Bool
+    -- ^ True if at least one clause has a lazy pattern in this position.
+  } deriving (Show)
 
-In these substitutions, we look for a column that has only constructor patterns.
-We try to split on this column first.
--}
+type BlockingVars = [BlockingVar]
 
-type SplitAssignment = (Nat,SplitPattern)
+-- | Substitution of 'SplitPattern's for de Bruijn indices in covering
+--   clause to match 'SplitClause'.
+type SplitInstantiation = [(Nat,SplitPattern)]
 
--- | Match the given patterns against a list of clauses
-match :: (MonadReduce m , HasConstInfo m , HasBuiltins m) => [Clause] -> [NamedArg SplitPattern] -> m (Match (Nat,[SplitAssignment]))
+-- | Match the given patterns against a list of clauses.
+--
+-- If successful, return the index of the covering clause.
+--
+match :: (MonadReduce m , HasConstInfo m , HasBuiltins m)
+      => [Clause]                           -- ^ Search for clause that covers the patterns.
+      -> [NamedArg SplitPattern]            -- ^ Patterns of the current 'SplitClause'.
+      -> m (Match (Nat, SplitInstantiation))
 match cs ps = foldr choice (return No) $ zipWith matchIt [0..] cs
   where
-    matchIt :: (MonadReduce m , HasConstInfo m , HasBuiltins m) => Nat -> Clause -> m (Match (Nat,[SplitAssignment]))
+    matchIt :: (MonadReduce m , HasConstInfo m , HasBuiltins m)
+            => Nat     -- ^ Clause number.
+            -> Clause
+            -> m (Match (Nat, SplitInstantiation))
     matchIt i c = fmap (i,) <$> matchClause ps c
 
 -- | For each variable in the patterns of a split clause, we remember the
 --   de Bruijn-index and the literals excluded by previous matches.
+
+--  (See issue #708.)
 data SplitPatVar = SplitPatVar
   { splitPatVarName   :: PatVarName
   , splitPatVarIndex  :: Int
@@ -78,10 +125,10 @@ data SplitPatVar = SplitPatVar
 
 instance Pretty SplitPatVar where
   prettyPrec _ x =
-    (text $ patVarNameToString (splitPatVarName x)) P.<>
-    (text $ "@" ++ show (splitPatVarIndex x)) P.<>
+    (text $ patVarNameToString (splitPatVarName x)) <>
+    (text $ "@" ++ show (splitPatVarIndex x)) <>
     (ifNull (splitExcludedLits x) empty $ \lits ->
-      "\\{" P.<> prettyList_ lits P.<> "}")
+      "\\{" <> prettyList_ lits <> "}")
 
 instance PrettyTCM SplitPatVar where
   prettyTCM = prettyTCM . var . splitPatVarIndex
@@ -104,10 +151,6 @@ toSplitPatterns = (fmap . fmap . fmap . fmap) toSplitVar
 fromSplitPatterns :: [NamedArg SplitPattern] -> [NamedArg DeBruijnPattern]
 fromSplitPatterns = (fmap . fmap . fmap . fmap) fromSplitVar
 
-instance DeBruijn SplitPattern where
-  debruijnNamedVar n i  = varP $ SplitPatVar n i []
-  deBruijnView _        = Nothing
-
 type SplitPSubstitution = Substitution' SplitPattern
 
 toSplitPSubst :: PatternSubstitution -> SplitPSubstitution
@@ -124,19 +167,19 @@ applySplitPSubst = applyPatSubst . fromSplitPSubst
 instance Subst SplitPattern SplitPattern where
   applySubst IdS p = p
   applySubst rho p = case p of
-    VarP o x     ->
-      usePatOrigin o $
+    VarP i x     ->
+      usePatternInfo i $
       useName (splitPatVarName x) $
       useExcludedLits (splitExcludedLits x) $
       lookupS rho $ splitPatVarIndex x
-    DotP o u     -> DotP o $ applySplitPSubst rho u
+    DotP i u     -> DotP i $ applySplitPSubst rho u
     ConP c ci ps -> ConP c ci $ applySubst rho ps
-    DefP o q ps -> DefP o q $ applySubst rho ps
-    LitP x       -> p
+    DefP i q ps -> DefP i q $ applySubst rho ps
+    LitP{}       -> p
     ProjP{}      -> p
-    IApplyP o l r x  ->
+    IApplyP i l r x  ->
       useEndPoints (applySplitPSubst rho l) (applySplitPSubst rho r) $
-      usePatOrigin o $
+      usePatternInfo i $
       useName (splitPatVarName x) $
       useExcludedLits (splitExcludedLits x) $
       lookupS rho $ splitPatVarIndex x
@@ -166,7 +209,7 @@ isTrivialPattern :: (HasConstInfo m) => Pattern' a -> m Bool
 isTrivialPattern p = case p of
   VarP{}      -> return True
   DotP{}      -> return True
-  ConP c i ps -> andM $ (isEtaCon $ conName c)
+  ConP c i ps -> andM $ ((conPLazy i ||) <$> isEtaCon (conName c))
                       : (map (isTrivialPattern . namedArg) ps)
   DefP{}      -> return False
   LitP{}      -> return False
@@ -175,84 +218,42 @@ isTrivialPattern p = case p of
 
 -- | If matching succeeds, we return the instantiation of the clause pattern vector
 --   to obtain the split clause pattern vector.
-type MatchResult = Match [SplitAssignment]
-
--- | If matching is inconclusive (@Block@) we want to know which
---   variables or projections are blocking the match.
-data Match a
-  = Yes a   -- ^ Matches unconditionally.
-  | No      -- ^ Definitely does not match.
-  | Block
-    { blockedOnResult :: BlockedOnResult
-      -- ^ @BlockedOnProj o@ if the clause has a result split
-    , blockedOnVars :: BlockingVars
-      -- ^ @BlockingVar i cs ls o@ means variable @i@ is blocked on
-      -- constructors @cs@ and literals @ls@.
-    }
-  deriving (Functor)
-
-data ApplyOrIApply = IsApply | IsIApply
-
-data BlockedOnResult
-  = BlockedOnProj      -- ^ Blocked on unsplit projection
-     { blockedOnResultOverlap :: Bool
-       -- ^ True if there are also matching clauses without an unsplit
-       -- copattern.
-     }
-  | BlockedOnApply     -- ^ Blocked on unintroduced argument
-     { blockedOnResultIApply :: ApplyOrIApply
-       -- ^ True if the unintroduced argument was an IApply pattern
-     }
-  | NotBlockedOnResult
-
--- | Variable blocking a match.
-data BlockingVar = BlockingVar
-  { blockingVarNo  :: Nat
-    -- ^ De Bruijn index of variable blocking the match.
-  , blockingVarCons :: [ConHead]
-    -- ^ Constructors in this position.
-  , blockingVarLits :: [Literal]
-    -- ^ Literals in this position.
-  , blockingVarOverlap :: Bool
-    -- ^ True if at least one clause has a variable pattern in this
-    --   position.
-  } deriving (Show)
+type MatchResult = Match SplitInstantiation
 
 instance Pretty BlockingVar where
-  pretty (BlockingVar i cs ls o) = cat
+  pretty (BlockingVar i cs ls o l) = cat
     [ text ("variable " ++ show i)
     , if null cs then empty else " blocked on constructors" <+> pretty cs
     , if null ls then empty else " blocked on literals" <+> pretty ls
     , if o then " (overlapping)" else empty
+    , if l then " (lazy)" else empty
     ]
 
-type BlockingVars = [BlockingVar]
-
 yes :: Monad m => a -> m (Match a)
 yes = return . Yes
 
 no :: Monad m => m (Match a)
 no = return No
 
-blockedOnConstructor :: Monad m => Nat -> ConHead -> m (Match a)
-blockedOnConstructor i c = return $ Block NotBlockedOnResult [BlockingVar i [c] [] False]
+blockedOnConstructor :: Monad m => Nat -> ConHead -> ConPatternInfo -> m (Match a)
+blockedOnConstructor i c ci = return $ Block NotBlockedOnResult [BlockingVar i [c] [] False $ conPLazy ci]
 
 blockedOnLiteral :: Monad m => Nat -> Literal -> m (Match a)
-blockedOnLiteral i l = return $ Block NotBlockedOnResult [BlockingVar i [] [l] False]
+blockedOnLiteral i l = return $ Block NotBlockedOnResult [BlockingVar i [] [l] False False]
 
 blockedOnProjection :: Monad m => m (Match a)
 blockedOnProjection = return $ Block (BlockedOnProj False) []
 
 blockedOnApplication :: Monad m => ApplyOrIApply -> m (Match a)
 blockedOnApplication b = return $ Block (BlockedOnApply b) []
-
--- | Lens for 'blockingVarCons'.
-mapBlockingVarCons :: ([ConHead] -> [ConHead]) -> BlockingVar -> BlockingVar
-mapBlockingVarCons f b = b { blockingVarCons = f (blockingVarCons b) }
-
--- | Lens for 'blockingVarLits'.
-mapBlockingVarLits :: ([Literal] -> [Literal]) -> BlockingVar -> BlockingVar
-mapBlockingVarLits f b = b { blockingVarLits = f (blockingVarLits b) }
+--UNUSED Liang-Ting Chen 2019-07-16
+---- | Lens for 'blockingVarCons'.
+--mapBlockingVarCons :: ([ConHead] -> [ConHead]) -> BlockingVar -> BlockingVar
+--mapBlockingVarCons f b = b { blockingVarCons = f (blockingVarCons b) }
+--
+---- | Lens for 'blockingVarLits'.
+--mapBlockingVarLits :: ([Literal] -> [Literal]) -> BlockingVar -> BlockingVar
+--mapBlockingVarLits f b = b { blockingVarLits = f (blockingVarLits b) }
 
 setBlockingVarOverlap :: BlockingVar -> BlockingVar
 setBlockingVarOverlap = \x -> x { blockingVarOverlap = True }
@@ -264,9 +265,9 @@ overlapping = map setBlockingVarOverlap
 zipBlockingVars :: BlockingVars -> BlockingVars -> BlockingVars
 zipBlockingVars xs ys = map upd xs
   where
-    upd (BlockingVar x cons lits o) = case List.find ((x ==) . blockingVarNo) ys of
-      Just (BlockingVar _ cons' lits' o') -> BlockingVar x (cons ++ cons') (lits ++ lits') (o || o')
-      Nothing -> BlockingVar x cons lits True
+    upd (BlockingVar x cons lits o l) = case List.find ((x ==) . blockingVarNo) ys of
+      Just (BlockingVar _ cons' lits' o' l') -> BlockingVar x (cons ++ cons') (lits ++ lits') (o || o') (l || l')
+      Nothing -> BlockingVar x cons lits True l
 
 setBlockedOnResultOverlap :: BlockedOnResult -> BlockedOnResult
 setBlockedOnResultOverlap b = case b of
@@ -284,9 +285,8 @@ anyBlockedOnResult b1 b2 = case (b1,b2) of
 choiceBlockedOnResult :: BlockedOnResult -> BlockedOnResult -> BlockedOnResult
 choiceBlockedOnResult b1 b2 = case (b1,b2) of
   (NotBlockedOnResult  , _                 ) -> NotBlockedOnResult
-  (BlockedOnProj _     , NotBlockedOnResult) -> BlockedOnProj True
   (BlockedOnProj o1    , BlockedOnProj o2  ) -> BlockedOnProj (o1 || o2)
-  (BlockedOnProj o1    , BlockedOnApply{}  ) -> BlockedOnProj True
+  (BlockedOnProj _     , _                 ) -> BlockedOnProj True
   (BlockedOnApply b    , _                 ) -> BlockedOnApply b
 
 -- | @choice m m'@ combines the match results @m@ of a function clause
@@ -386,13 +386,14 @@ combine m m' = m >>= \case
 
 -- | @matchPat p q@ checks whether a function clause pattern @p@
 --   covers a split clause pattern @q@.  There are three results:
---   @Yes rs@ means it covers, because @p@ is a variable pattern. @rs@ collects
---   the instantiations of the variables in @p@ s.t. @p[rs] = q@.
---   @No@ means it does not cover.
---   @Block [x]@ means @p@ is a proper instance of @q@ and could become
---   a cover if @q@ was split on variable @x@.
---   @BlockLit [x] means @p@ is a proper instance of @q@ and could become
---   a cover if variable @x@ is instantiated with an appropriate literal.
+--
+--   1. @Yes rs@ means it covers, because @p@ is a variable pattern. @rs@ collects
+--      the instantiations of the variables in @p@ s.t. @p[rs] = q@.
+--
+--   2. @No@ means it does not cover.
+--
+--   3. @Block [x]@ means @p@ is a proper instance of @q@ and could become
+--      a cover if @q@ was split on variable @x@.
 
 matchPat
   :: (MonadReduce m , HasConstInfo m , HasBuiltins m, DeBruijn a)
@@ -405,61 +406,68 @@ matchPat
      --   If 'Yes', also the instantiation @rs@ of the clause pattern variables
      --   to produce the split clause pattern, @p[rs] = q@.
 matchPat p q = case p of
+
   VarP _ x   -> yes [(fromMaybe __IMPOSSIBLE__ (deBruijnView x),q)]
+
   DotP{}   -> yes []
   -- Jesper, 2014-11-04: putting 'Yes [q]' here triggers issue 1333.
   -- Not checking for trivial patterns should be safe here, as dot patterns are
   -- guaranteed to match if the rest of the pattern does, so some extra splitting
   -- on them doesn't change the reduction behaviour.
-  p@(LitP l) -> case q of
+
+  p@(LitP _ l) -> case q of
     VarP _ x -> if l `elem` splitExcludedLits x
                 then no
                 else blockedOnLiteral (splitPatVarIndex x) l
     _ -> isLitP q >>= \case
       Just l' -> if l == l' then yes [] else no
       Nothing -> no
+
   ProjP _ d -> case q of
     ProjP _ d' -> do
       d <- getOriginalProjection d
       if d == d' then yes [] else no
     _          -> __IMPOSSIBLE__
+
   IApplyP _ _ _ x -> yes [(fromMaybe __IMPOSSIBLE__ (deBruijnView x),q)]
-  ConP c _ ps -> unDotP q >>= \case
-    VarP _ x -> blockedOnConstructor (splitPatVarIndex x) c
+
+                           --    Issue #4179: If the inferred pattern is a literal
+                           -- v  we need to turn it into a constructor pattern.
+  ConP c ci ps -> unDotP q >>= unLitP >>= \case
+    VarP _ x -> blockedOnConstructor (splitPatVarIndex x) c ci
     ConP c' i qs
       | c == c'   -> matchPats ps qs
       | otherwise -> no
     DotP o t  -> no
-    LitP l    -> isLitP p >>= \case
-      Just l' -> if l == l' then yes [] else no
-      Nothing -> no
     DefP{}   -> no
+    LitP{}    -> __IMPOSSIBLE__  -- excluded by typing and unLitP
     ProjP{}   -> __IMPOSSIBLE__  -- excluded by typing
-    IApplyP _ _ _ x -> blockedOnConstructor (splitPatVarIndex x) c
-  (DefP o c ps) -> unDotP q >>= \case
+    IApplyP _ _ _ x -> blockedOnConstructor (splitPatVarIndex x) c ci
+
+  DefP o c ps -> unDotP q >>= \case
     VarP _ x -> __IMPOSSIBLE__ -- blockedOnConstructor (splitPatVarIndex x) c
     ConP c' i qs -> no
     DotP o t  -> no
-    LitP _    -> no
+    LitP{}    -> no
     DefP o c' qs
       | c == c'   -> matchPats ps qs
       | otherwise -> no
     ProjP{}   -> __IMPOSSIBLE__  -- excluded by typing
-    IApplyP _ _ _ x -> __IMPOSSIBLE__ --blockedOnConstructor (splitPatVarIndex x) c
+    IApplyP _ _ _ x -> __IMPOSSIBLE__ -- blockedOnConstructor (splitPatVarIndex x) c
 
--- unfold one level of a dot pattern to a proper pattern if possible
-unDotP :: (MonadReduce m, DeBruijn (Pattern' a)) => Pattern' a -> m (Pattern' a)
+-- | Unfold one level of a dot pattern to a proper pattern if possible.
+unDotP :: (MonadReduce m, DeBruijn a) => Pattern' a -> m (Pattern' a)
 unDotP (DotP o v) = reduce v >>= \case
   Var i [] -> return $ deBruijnVar i
   Con c _ vs -> do
     let ps = map (fmap $ unnamed . DotP o) $ fromMaybe __IMPOSSIBLE__ $ allApplyElims vs
     return $ ConP c noConPatternInfo ps
-  Lit l -> return $ LitP l
+  Lit l -> return $ LitP (PatternInfo PatODot []) l
   v     -> return $ dotP v
 unDotP p = return p
 
 isLitP :: (MonadReduce m, HasBuiltins m) => Pattern' a -> m (Maybe Literal)
-isLitP (LitP l) = return $ Just l
+isLitP (LitP _ l) = return $ Just l
 isLitP (DotP _ u) = reduce u >>= \case
   Lit l -> return $ Just l
   _     -> return $ Nothing
@@ -476,3 +484,14 @@ isLitP (ConP c ci [a]) | visible a && isRelevant a = do
     inc (LitNat r n) = LitNat (fuseRange c r) $ n + 1
     inc _ = __IMPOSSIBLE__
 isLitP _ = return Nothing
+
+unLitP :: HasBuiltins m => Pattern' a -> m (Pattern' a)
+unLitP (LitP info l@(LitNat _ n)) | n >= 0 = do
+  Con c ci es <- constructorForm' (fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinZero)
+                                  (fromMaybe __IMPOSSIBLE__ <$> getBuiltin' builtinSuc)
+                                  (Lit l)
+  let toP (Apply (Arg i (Lit l))) = Arg i (LitP info l)
+      toP _ = __IMPOSSIBLE__
+      cpi   = noConPatternInfo { conPInfo = info }
+  return $ ConP c cpi $ map (fmap unnamed . toP) es
+unLitP p = return p
