@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable         #-}
 
 {-| Split tree for transforming pattern clauses into case trees.
@@ -15,18 +14,17 @@ each leaf of the split tree.
 module Agda.TypeChecking.Coverage.SplitTree where
 
 import Data.Tree
-import Data.Data (Data, toConstr)
+import Data.Data (Data)
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
-import Agda.Syntax.Internal as I
+import Agda.Syntax.Concrete.Pretty () --instance only
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
-import Agda.Utils.Monad
 import Agda.Utils.Pretty
+import Agda.Utils.Null
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 type SplitTree  = SplitTree'  SplitTag
@@ -42,10 +40,14 @@ data SplitTree' a
   | -- | A split is necessary.
     SplitAt
     { splitArg   :: Arg Int       -- ^ Arg. no to split at.
+    , splitLazy  :: LazySplit
     , splitTrees :: SplitTrees' a -- ^ Sub split trees.
     }
   deriving (Data, Show)
 
+data LazySplit = LazySplit | StrictSplit
+  deriving (Data, Show, Eq, Ord)
+
 -- | Split tree branching.  A finite map from constructor names to splittrees
 --   A list representation seems appropriate, since we are expecting not
 --   so many constructors per data type, and there is no need for
@@ -62,7 +64,7 @@ data SplitTag
   deriving (Show, Eq, Ord, Data)
 
 instance Pretty SplitTag where
-  pretty (SplitCon c)  = pretty c
+  pretty (SplitCon c) = pretty c
   pretty (SplitLit l)  = pretty l
   pretty SplitCatchall = underscore
 
@@ -71,21 +73,24 @@ instance Pretty SplitTag where
 data SplitTreeLabel a = SplitTreeLabel
   { lblConstructorName :: Maybe a   -- ^ 'Nothing' for root of split tree
   , lblSplitArg        :: Maybe (Arg Int)
+  , lblLazy            :: LazySplit
   , lblBindings        :: Maybe Int
   }
 instance Pretty a => Pretty (SplitTreeLabel a) where
   pretty = \case
-    SplitTreeLabel Nothing Nothing (Just n)  -> text $ "done, " ++ prettyShow n ++ " bindings"
-    SplitTreeLabel Nothing (Just n) Nothing  -> text $ "split at " ++ prettyShow n
-    SplitTreeLabel (Just q) Nothing (Just n) -> pretty q <+> text (" -> done, " ++ prettyShow n ++ " bindings")
-    SplitTreeLabel (Just q) (Just n) Nothing -> pretty q <+> text (" -> split at " ++ prettyShow n)
+    SplitTreeLabel Nothing Nothing   _  (Just n) -> text $ "done, " ++ prettyShow n ++ " bindings"
+    SplitTreeLabel Nothing (Just n)  lz Nothing  -> lzp lz <+> text ("split at " ++ prettyShow n)
+    SplitTreeLabel (Just q) Nothing  _  (Just n) -> pretty q <+> text ("-> done, " ++ prettyShow n ++ " bindings")
+    SplitTreeLabel (Just q) (Just n) lz Nothing  -> pretty q <+> text "->" <+> lzp lz <+> text ("split at " ++ prettyShow n)
     _ -> __IMPOSSIBLE__
+    where lzp lz | lz == LazySplit = "lazy"
+                 | otherwise       = empty
 
 -- | Convert a split tree into a 'Data.Tree' (for printing).
 toTree :: SplitTree' a -> Tree (SplitTreeLabel a)
 toTree t = case t of
-  SplittingDone n -> Node (SplitTreeLabel Nothing Nothing (Just n)) []
-  SplitAt n ts    -> Node (SplitTreeLabel Nothing (Just n) Nothing) $ toTrees ts
+  SplittingDone n -> Node (SplitTreeLabel Nothing Nothing StrictSplit (Just n)) []
+  SplitAt n lz ts    -> Node (SplitTreeLabel Nothing (Just n) lz Nothing) $ toTrees ts
 
 toTrees :: SplitTrees' a -> Forest (SplitTreeLabel a)
 toTrees = map (\ (c,t) -> setCons c $ toTree t)
@@ -105,4 +110,4 @@ instance KillRange SplitTag where
 instance KillRange a => KillRange (SplitTree' a) where
   killRange = \case
     SplittingDone n -> SplittingDone n
-    SplitAt i ts    -> killRange1 (SplitAt i) ts
+    SplitAt i lz ts -> killRange1 (SplitAt i lz) ts
