@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Datatypes where
 
@@ -17,11 +16,9 @@ import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Pretty
 
 import Agda.Utils.Either
-import Agda.Utils.Functor
 import Agda.Utils.Pretty ( prettyShow )
 import Agda.Utils.Size
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
@@ -78,9 +75,10 @@ consOfHIT c = do
 --   @Nothing@ if @t@ is not a data/record type or does not have
 --   a constructor @c@.
 getConType
-  :: ConHead  -- ^ Constructor.
+  :: (MonadReduce m, MonadAddContext m, HasConstInfo m, MonadDebug m)
+  => ConHead  -- ^ Constructor.
   -> Type     -- ^ Ending in data/record type.
-  -> TCM (Maybe ((QName, Type, Args), Type))
+  -> m (Maybe ((QName, Type, Args), Type))
        -- ^ @Nothing@ if not ends in data or record type.
        --
        --   @Just ((d, dt, pars), ct)@ otherwise, where
@@ -158,14 +156,14 @@ getFullyAppliedConType c t = do
 
 data ConstructorInfo
   = DataCon Nat                  -- ^ Arity.
-  | RecordCon HasEta [Arg QName] -- ^ List of field names.
+  | RecordCon HasEta [Dom QName] -- ^ List of field names.
 
 -- | Return the number of non-parameter arguments to a data constructor,
 --   or the field names of a record constructor.
 --
 --   For getting just the arity of constructor @c@,
 --   use @either id size <$> getConstructorArity c@.
-getConstructorInfo :: QName -> TCM ConstructorInfo
+getConstructorInfo :: HasConstInfo m => QName -> m ConstructorInfo
 getConstructorInfo c = do
   (theDef <$> getConstInfo c) >>= \case
     Constructor{ conData = d, conArity = n } -> do
@@ -238,14 +236,22 @@ getConHeads d = fromMaybe __IMPOSSIBLE__ <$>
 
 -- | 'Nothing' if not data or record type name.
 getConHeads' :: QName -> TCM (Maybe [ConHead])
-getConHeads' d = getConHeads_ . theDef <$> getConstInfo d
+getConHeads' d = theDef <$> getConstInfo d >>= \case
+  Record{recConHead = h}  -> return $ Just [h]
+  Datatype{dataCons = cs} -> Just <$> mapM makeConHead cs
+  _                       -> return $ Nothing
 
--- | 'Nothing' if not data or record definition.
-getConHeads_ :: Defn -> Maybe [ConHead]
-getConHeads_ = \case
-    Datatype{dataCons = cs} -> Just $ map (\ c -> ConHead c Inductive []) cs
-    Record{recConHead = h}  -> Just [h]
-    _                       -> Nothing
+-- | Fills in the fields.
+makeConHead :: QName -> TCM ConHead
+makeConHead c = do
+  def <- getConstInfo c
+  case theDef def of
+    Constructor{conPars = n, conProj = Just fs} -> do
+      TelV tel _ <- telView (defType def)
+      let ai = map getArgInfo $ drop n $ telToList tel
+      return $ ConHead c Inductive $ zipWith Arg ai fs
+    Constructor{conProj = Nothing} -> return $ ConHead c Inductive []
+    _ -> __IMPOSSIBLE__
 
 {- UNUSED
 data DatatypeInfo = DataInfo
