@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 -- | Compute eta long normal forms.
 module Agda.TypeChecking.EtaExpand where
@@ -12,13 +11,10 @@ import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 
-import Agda.Utils.Monad
-
-#include "undefined.h"
-import Agda.Utils.Impossible
-
 -- | Eta-expand a term if its type is a function type or an eta-record type.
-etaExpandOnce :: Type -> Term -> TCM Term
+etaExpandOnce
+  :: (MonadReduce m, MonadTCEnv m, HasOptions m, HasConstInfo m, MonadDebug m)
+  => Type -> Term -> m Term
 etaExpandOnce a v = reduce a >>= \case
   El _ (Pi a b) -> return $
     Lam ai $ mkAbs (absName b) $ raise 1 v `apply` [ Arg ai $ var 0 ]
@@ -34,9 +30,11 @@ etaExpandOnce a v = reduce a >>= \case
 -- | Eta-expand functions and expressions of eta-record
 -- type wherever possible.
 deepEtaExpand :: Term -> Type -> TCM Term
-deepEtaExpand = checkInternal' etaExpandAction
+deepEtaExpand v a = checkInternal' etaExpandAction v CmpLeq a
 
-etaExpandAction :: Action
+etaExpandAction
+  :: (MonadReduce m, MonadTCEnv m, HasOptions m, HasConstInfo m, MonadDebug m)
+  => Action m
 etaExpandAction = Action
   { preAction       = etaExpandOnce
   , postAction      = \ _ -> return
