@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE TypeFamilies #-}
 -- | Precompute free variables in a term (and store in 'ArgInfo').
 module Agda.TypeChecking.Free.Precompute
@@ -12,10 +11,8 @@ import Data.Traversable (Traversable, traverse)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
-import Agda.Utils.Functor
-import Agda.Utils.Impossible
 
-#include "undefined.h"
+
 
 type FV = Writer IntSet
 
@@ -83,17 +80,17 @@ instance PrecomputeFreeVars Sort where
       Prop a     -> Prop <$> precomputeFreeVars a
       Inf        -> pure s
       SizeUniv   -> pure s
-      PiSort s1 s2 -> uncurry PiSort <$> precomputeFreeVars (s1, s2)
+      PiSort a s -> uncurry PiSort <$> precomputeFreeVars (a, s)
+      FunSort s1 s2 -> uncurry FunSort <$> precomputeFreeVars (s1, s2)
       UnivSort s -> UnivSort <$> precomputeFreeVars s
       MetaS x es -> MetaS x <$> precomputeFreeVars es
       DefS d es  -> DefS d <$> precomputeFreeVars es
       DummyS{}   -> pure s
 
 instance PrecomputeFreeVars Level where
-  precomputeFreeVars (Max ls) = Max <$> precomputeFreeVars ls
+  precomputeFreeVars (Max n ls) = Max n <$> precomputeFreeVars ls
 
 instance PrecomputeFreeVars PlusLevel where
-  precomputeFreeVars l@ClosedLevel{} = pure l
   precomputeFreeVars (Plus n l) = Plus n <$> precomputeFreeVars l
 
 instance PrecomputeFreeVars LevelAtom where
