@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE PatternSynonyms #-}
 
 {-| Functions for inserting implicit arguments at the right places.
@@ -12,25 +11,26 @@ import Agda.Syntax.Internal as I
 
 import Agda.TypeChecking.Irrelevance
 import {-# SOURCE #-} Agda.TypeChecking.MetaVars
+import {-# SOURCE #-} Agda.TypeChecking.Rules.Term (unquoteTactic)
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Pretty
 
+import Agda.Utils.Functor
+import Agda.Utils.Maybe
 import Agda.Utils.Tuple
 
-#include "undefined.h"
-import Agda.Utils.Impossible
-
--- | @implicitArgs n expand t@ generates up to @n@ implicit arguments
+-- | @implicitArgs n expand t@ generates up to @n@ implicit argument
 --   metas (unbounded if @n<0@), as long as @t@ is a function type
 --   and @expand@ holds on the hiding info of its domain.
 
 implicitArgs
-  :: Int               -- ^ @n@, the maximum number of implicts to be inserted.
+  :: (MonadReduce m, MonadMetaSolver m, MonadDebug m, MonadTCM m)
+  => Int               -- ^ @n@, the maximum number of implicts to be inserted.
   -> (Hiding -> Bool)  -- ^ @expand@, the predicate to test whether we should keep inserting.
   -> Type              -- ^ The (function) type @t@ we are eliminating.
-  -> TCM (Args, Type)  -- ^ The eliminating arguments and the remaining type.
+  -> m (Args, Type)  -- ^ The eliminating arguments and the remaining type.
 implicitArgs n expand t = mapFst (map (fmap namedThing)) <$> do
   implicitNamedArgs n (\ h x -> expand h) t
 
@@ -39,18 +39,19 @@ implicitArgs n expand t = mapFst (map (fmap namedThing)) <$> do
 --   and @expand@ holds on the hiding and name info of its domain.
 
 implicitNamedArgs
-  :: Int                          -- ^ @n@, the maximum number of implicts to be inserted.
+  :: (MonadReduce m, MonadMetaSolver m, MonadDebug m, MonadTCM m)
+  => Int                          -- ^ @n@, the maximum number of implicts to be inserted.
   -> (Hiding -> ArgName -> Bool)  -- ^ @expand@, the predicate to test whether we should keep inserting.
   -> Type                         -- ^ The (function) type @t@ we are eliminating.
-  -> TCM (NamedArgs, Type)        -- ^ The eliminating arguments and the remaining type.
+  -> m (NamedArgs, Type)        -- ^ The eliminating arguments and the remaining type.
 implicitNamedArgs 0 expand t0 = return ([], t0)
 implicitNamedArgs n expand t0 = do
     t0' <- reduce t0
     reportSDoc "tc.term.args" 30 $ "implicitNamedArgs" <+> prettyTCM t0'
     reportSDoc "tc.term.args" 80 $ "implicitNamedArgs" <+> text (show t0')
     case unEl t0' of
-      Pi Dom{domInfo = info, domName = name, unDom = a} b
-        | let x = maybe "_" rangedThing name, expand (getHiding info) x -> do
+      Pi dom@Dom{domInfo = info, domTactic = tac, unDom = a} b
+        | let x = bareNameWithDefault "_" dom, expand (getHiding info) x -> do
           info' <- if hidden info then return info else do
             reportSDoc "tc.term.args.ifs" 15 $
               "inserting instance meta for type" <+> prettyTCM a
@@ -60,48 +61,54 @@ implicitNamedArgs n expand t0 = do
               ]
 
             return $ makeInstance info
-          (_, v) <- newMetaArg info' x a
-          let narg = Arg info (Named (Just $ unranged x) v)
+          (_, v) <- newMetaArg info' x CmpLeq a
+          whenJust tac $ \ tac -> liftTCM $ unquoteTactic tac v a
+          let narg = Arg info (Named (Just $ WithOrigin Inserted $ unranged x) v)
           mapFst (narg :) <$> implicitNamedArgs (n-1) expand (absApp b v)
       _ -> return ([], t0')
 
 -- | Create a metavariable according to the 'Hiding' info.
 
 newMetaArg
-  :: ArgInfo   -- ^ Kind/relevance of meta.
-  -> ArgName   -- ^ Name suggestion for meta.
-  -> Type      -- ^ Type of meta.
-  -> TCM (MetaId, Term)  -- ^ The created meta as id and as term.
-newMetaArg info x a = do
-  applyModalityToContext info $
+  :: MonadMetaSolver m
+  => ArgInfo    -- ^ Kind/relevance of meta.
+  -> ArgName    -- ^ Name suggestion for meta.
+  -> Comparison -- ^ Check (@CmpLeq@) or infer (@CmpEq@) the type.
+  -> Type       -- ^ Type of meta.
+  -> m (MetaId, Term)  -- ^ The created meta as id and as term.
+newMetaArg info x cmp a = do
+  prp <- isPropM a
+  let irrelevantIfProp = if prp then applyRelevanceToContext Irrelevant else id
+  applyModalityToContext info $ irrelevantIfProp $
     newMeta (getHiding info) (argNameToString x) a
   where
-    newMeta :: Hiding -> String -> Type -> TCM (MetaId, Term)
-    newMeta Instance{} = newInstanceMeta
-    newMeta Hidden     = newNamedValueMeta RunMetaOccursCheck
-    newMeta NotHidden  = newNamedValueMeta RunMetaOccursCheck
+    newMeta :: MonadMetaSolver m => Hiding -> String -> Type -> m (MetaId, Term)
+    newMeta Instance{} n = newInstanceMeta n
+    newMeta Hidden     n = newNamedValueMeta RunMetaOccursCheck n cmp
+    newMeta NotHidden  n = newNamedValueMeta RunMetaOccursCheck n cmp
 
 -- | Create a questionmark according to the 'Hiding' info.
 
 newInteractionMetaArg
-  :: ArgInfo   -- ^ Kind/relevance of meta.
-  -> ArgName   -- ^ Name suggestion for meta.
-  -> Type      -- ^ Type of meta.
+  :: ArgInfo    -- ^ Kind/relevance of meta.
+  -> ArgName    -- ^ Name suggestion for meta.
+  -> Comparison -- ^ Check (@CmpLeq@) or infer (@CmpEq@) the type.
+  -> Type       -- ^ Type of meta.
   -> TCM (MetaId, Term)  -- ^ The created meta as id and as term.
-newInteractionMetaArg info x a = do
+newInteractionMetaArg info x cmp a = do
   applyModalityToContext info $
     newMeta (getHiding info) (argNameToString x) a
   where
     newMeta :: Hiding -> String -> Type -> TCM (MetaId, Term)
-    newMeta Instance{} = newInstanceMeta
-    newMeta Hidden     = newNamedValueMeta' RunMetaOccursCheck
-    newMeta NotHidden  = newNamedValueMeta' RunMetaOccursCheck
+    newMeta Instance{} n = newInstanceMeta n
+    newMeta Hidden     n = newNamedValueMeta' RunMetaOccursCheck n cmp
+    newMeta NotHidden  n = newNamedValueMeta' RunMetaOccursCheck n cmp
 
 ---------------------------------------------------------------------------
 
 -- | Possible results of 'insertImplicit'.
 data ImplicitInsertion
-      = ImpInsert [Hiding] -- ^ Success: this many implicits have to be inserted (list can be empty).
+      = ImpInsert [Dom ()] -- ^ Success: this many implicits have to be inserted (list can be empty).
       | BadImplicits       -- ^ Error: hidden argument where there should have been a non-hidden argument.
       | NoSuchName ArgName -- ^ Error: bad named argument.
   deriving (Show)
@@ -118,10 +125,9 @@ insertImplicit
   :: NamedArg e  -- ^ Next given argument @a@.
   -> [Dom a]     -- ^ Expected arguments @ts@.
   -> ImplicitInsertion
-insertImplicit a doms = insertImplicit' a $ map name doms
-  where
-    name dom = x <$ argFromDom dom
-      where x = maybe "_" rangedThing $ domName dom
+insertImplicit a doms = insertImplicit' a $
+  for doms $ \ dom ->
+    dom $> bareNameWithDefault "_" dom
 
 -- | If the next given argument is @a@ and the expected arguments are @ts@
 --   @insertImplicit' a ts@ returns the prefix of @ts@ that precedes @a@.
@@ -130,18 +136,18 @@ insertImplicit a doms = insertImplicit' a $ map name doms
 --
 insertImplicit'
   :: NamedArg e     -- ^ Next given argument @a@.
-  -> [Arg ArgName]  -- ^ Expected arguments @ts@.
+  -> [Dom ArgName]  -- ^ Expected arguments @ts@.
   -> ImplicitInsertion
 insertImplicit' _ [] = BadImplicits
 insertImplicit' a ts
 
   -- If @a@ is visible, then take the non-visible prefix of @ts@.
-  | visible a = ImpInsert $ takeWhile notVisible $ map getHiding ts
+  | visible a = ImpInsert $ takeWhile notVisible $ map void ts
 
   -- If @a@ is named, take prefix of @ts@ until the name of @a@ (with correct hiding).
   -- If the name is not found, throw exception 'NoSuchName'.
-  | Just x <- rangedThing <$> nameOf (unArg a) = maybe (NoSuchName x) ImpInsert $
-      takeHiddenUntil (\ t -> x == unArg t && sameHiding a t) ts
+  | Just x <- bareNameOf a = maybe (NoSuchName x) ImpInsert $
+      takeHiddenUntil (\ t -> x == unDom t && sameHiding a t) ts
 
   -- If @a@ is neither visible nor named, take prefix of @ts@ with different hiding than @a@.
   | otherwise = maybe BadImplicits ImpInsert $
@@ -153,10 +159,10 @@ insertImplicit' a ts
     --   If @p@ never holds, 'Nothing' is returned.
     --
     --   Precondition: @p@ should imply @not . visible@.
-    takeHiddenUntil :: (Arg ArgName -> Bool) -> [Arg ArgName] -> Maybe [Hiding]
+    takeHiddenUntil :: (Dom ArgName -> Bool) -> [Dom ArgName] -> Maybe [Dom ()]
     takeHiddenUntil p ts =
       case ts2 of
         []      -> Nothing  -- Predicate was never true
-        (t : _) -> if visible t then Nothing else Just $ map getHiding ts1
+        (t : _) -> if visible t then Nothing else Just $ map void ts1
       where
       (ts1, ts2) = break (\ t -> p t || visible t) ts
