@@ -1,5 +1,4 @@
 {-# LANGUAGE BangPatterns #-}
-{-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Monad.Caching
   ( -- * Log reading/writing operations
@@ -12,13 +11,13 @@ module Agda.TypeChecking.Monad.Caching
   , activateLoadedFileCache
   , cachingStarts
   , areWeCaching
-  , withoutCache
+  , localCache, withoutCache
 
     -- * Restoring the 'PostScopeState'
   , restorePostScopeState
   ) where
 
-import Control.Monad.State
+
 import qualified Data.Map as Map
 
 import Agda.Syntax.Common
@@ -27,33 +26,35 @@ import Agda.Interaction.Options
 
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Debug
-import {-# SOURCE #-} Agda.TypeChecking.Monad.Options
 
 import Agda.Utils.Lens
 import Agda.Utils.Monad
+import Agda.Utils.Null (empty)
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | To be called before any write or restore calls.
-cachingStarts :: TCM ()
+{-# SPECIALIZE cachingStarts :: TCM () #-}
+cachingStarts :: (MonadTCState m, ReadTCState m) => m ()
 cachingStarts = do
     NameId _ m <- useTC stFreshNameId
     stFreshNameId `setTCLens` NameId 1 m
     stAreWeCaching `setTCLens` True
 
-areWeCaching :: TCM Bool
-areWeCaching = useTC stAreWeCaching
+areWeCaching :: (ReadTCState m) => m Bool
+areWeCaching = useR stAreWeCaching
 
 -- | Writes a 'TypeCheckAction' to the current log, using the current
 -- 'PostScopeState'
-writeToCurrentLog :: TypeCheckAction -> TCM ()
+{-# SPECIALIZE writeToCurrentLog :: TypeCheckAction -> TCM () #-}
+writeToCurrentLog :: (MonadDebug m, MonadTCState m, ReadTCState m) => TypeCheckAction -> m ()
 writeToCurrentLog !d = do
   reportSLn "cache" 10 $ "cachePostScopeState"
   !l <- getsTC stPostScopeState
   modifyCache $ fmap $ \lfc -> lfc{ lfcCurrent = (d, l) : lfcCurrent lfc}
 
-restorePostScopeState :: PostScopeState -> TCM ()
+{-# SPECIALIZE restorePostScopeState :: PostScopeState -> TCM () #-}
+restorePostScopeState :: (MonadDebug m, MonadTCState m) => PostScopeState -> m ()
 restorePostScopeState pss = do
   reportSLn "cache" 10 $ "restorePostScopeState"
   modifyTC $ \s ->
@@ -71,36 +72,40 @@ restorePostScopeState pss = do
     mergeWarnings loading current = [ w | w <- current, not $ tcWarningCached w ]
                                  ++ [ w | w <- loading,       tcWarningCached w ]
 
+{-# SPECIALIZE modifyCache :: (Maybe LoadedFileCache -> Maybe LoadedFileCache) -> TCM () #-}
 modifyCache
-  :: (Maybe LoadedFileCache -> Maybe LoadedFileCache)
-  -> TCM ()
-modifyCache f = do
-  modifyTC $ \s -> let !p = stPersistentState s in s
-    { stPersistentState =
-                          p { stLoadedFileCache = f (stLoadedFileCache p)}
-    }
+  :: MonadTCState m
+  => (Maybe LoadedFileCache -> Maybe LoadedFileCache)
+  -> m ()
+modifyCache = modifyTCLens stLoadedFileCache
 
-getCache :: TCM (Maybe LoadedFileCache)
-getCache = do
-  getsTC (stLoadedFileCache . stPersistentState)
+{-# SPECIALIZE getCache :: TCM (Maybe LoadedFileCache) #-}
+getCache :: ReadTCState m => m (Maybe LoadedFileCache)
+getCache = useTC stLoadedFileCache
 
-putCache :: Maybe LoadedFileCache -> TCM ()
-putCache cs = modifyCache $ const cs
+{-# SPECIALIZE putCache :: Maybe LoadedFileCache -> TCM () #-}
+putCache :: MonadTCState m => Maybe LoadedFileCache -> m ()
+putCache = setTCLens stLoadedFileCache
+
+-- | Runs the action and restores the current cache at the end of it.
+{-# SPECIALIZE localCache :: TCM a -> TCM a #-}
+localCache :: (MonadTCState m, ReadTCState m) => m a -> m a
+localCache = bracket_ getCache putCache
 
 -- | Runs the action without cache and restores the current cache at
 -- the end of it.
-withoutCache :: TCM a -> TCM a
-withoutCache m =
-  bracket_ getCache putCache $ do
-    modifyCache (const Nothing)
+{-# SPECIALIZE withoutCache :: TCM a -> TCM a #-}
+withoutCache :: (MonadTCState m, ReadTCState m) => m a -> m a
+withoutCache m = localCache $ do
+    putCache empty
     m
 
 -- | Reads the next entry in the cached type check log, if present.
-readFromCachedLog :: TCM (Maybe (TypeCheckAction, PostScopeState))
+{-# SPECIALIZE readFromCachedLog :: TCM (Maybe (TypeCheckAction, PostScopeState)) #-}
+readFromCachedLog :: (MonadDebug m, MonadTCState m, ReadTCState m) => m (Maybe (TypeCheckAction, PostScopeState))
 readFromCachedLog = do
   reportSLn "cache" 10 $ "getCachedTypeCheckAction"
-  mbCache <- getCache
-  case mbCache of
+  getCache >>= \case
     Just lfc | (entry : entries) <- lfcCached lfc -> do
       putCache $ Just lfc{lfcCached = entries}
       return (Just entry)
@@ -108,7 +113,8 @@ readFromCachedLog = do
       return Nothing
 
 -- | Empties the "to read" CachedState. To be used when it gets invalid.
-cleanCachedLog :: TCM ()
+{-# SPECIALIZE cleanCachedLog :: TCM () #-}
+cleanCachedLog :: (MonadDebug m, MonadTCState m) => m ()
 cleanCachedLog = do
   reportSLn "cache" 10 $ "cleanCachedLog"
   modifyCache $ fmap $ \lfc -> lfc{lfcCached = []}
@@ -117,7 +123,8 @@ cleanCachedLog = do
 -- current log. Crashes is 'stLoadedFileCache' is already active with a
 -- dirty log.  Should be called when we start typechecking the current
 -- file.
-activateLoadedFileCache :: TCM ()
+{-# SPECIALIZE activateLoadedFileCache :: TCM () #-}
+activateLoadedFileCache :: (HasOptions m, MonadDebug m, MonadTCState m) => m ()
 activateLoadedFileCache = do
   reportSLn "cache" 10 $ "activateLoadedFileCache"
 
@@ -130,7 +137,8 @@ activateLoadedFileCache = do
 
 -- | Caches the current type check log.  Discardes the old cache.  Does
 -- nothing if caching is inactive.
-cacheCurrentLog :: TCM ()
+{-# SPECIALIZE cacheCurrentLog :: TCM () #-}
+cacheCurrentLog :: (MonadDebug m, MonadTCState m) => m ()
 cacheCurrentLog = do
   reportSLn "cache" 10 $ "cacheCurrentTypeCheckLog"
   modifyCache $ fmap $ \lfc ->
