@@ -1,11 +1,10 @@
-{-# LANGUAGE CPP #-}
 module Agda.TypeChecking.Monad.Env where
 
-import Control.Monad.Reader
+
 import qualified Data.List as List
-import qualified Data.Map as Map
+
 import Data.Maybe (fromMaybe)
-import Data.Monoid
+
 
 import Agda.Syntax.Common
 import Agda.Syntax.Abstract.Name
@@ -13,8 +12,10 @@ import Agda.Syntax.Abstract.Name
 import Agda.TypeChecking.Monad.Base
 
 import Agda.Utils.FileName
+import Agda.Utils.Function
+import Agda.Utils.SmallSet
+import qualified Agda.Utils.SmallSet as SmallSet
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Get the name of the current module, if any.
@@ -24,7 +25,7 @@ currentModule :: MonadTCEnv m => m ModuleName
 currentModule = asksTC envCurrentModule
 
 -- | Set the name of the current module.
-withCurrentModule :: ModuleName -> TCM a -> TCM a
+withCurrentModule :: (MonadTCEnv m) => ModuleName -> m a -> m a
 withCurrentModule m =
     localTC $ \ e -> e { envCurrentModule = m }
 
@@ -46,7 +47,7 @@ withAnonymousModule m n =
   localTC $ \ e -> e { envAnonymousModules = (m, n) : envAnonymousModules e }
 
 -- | Set the current environment to the given
-withEnv :: TCEnv -> TCM a -> TCM a
+withEnv :: MonadTCEnv m => TCEnv -> m a -> m a
 withEnv env = localTC $ \ env0 -> env
   -- Keep persistent settings
   { envPrintMetasBare         = envPrintMetasBare env0
@@ -72,11 +73,17 @@ withoutOptionsChecking = localTC $ \ e -> e { envCheckOptionConsistency = False
 
 -- | Restore setting for 'ExpandLast' to default.
 doExpandLast :: TCM a -> TCM a
-doExpandLast = localTC $ \ e -> e { envExpandLast = ExpandLast }
+doExpandLast = localTC $ \ e -> e { envExpandLast = setExpand (envExpandLast e) }
+  where
+    setExpand ReallyDontExpandLast = ReallyDontExpandLast
+    setExpand _                    = ExpandLast
 
 dontExpandLast :: TCM a -> TCM a
 dontExpandLast = localTC $ \ e -> e { envExpandLast = DontExpandLast }
 
+reallyDontExpandLast :: TCM a -> TCM a
+reallyDontExpandLast = localTC $ \ e -> e { envExpandLast = ReallyDontExpandLast }
+
 -- | If the reduced did a proper match (constructor or literal pattern),
 --   then record this as simplification step.
 {-# SPECIALIZE performedSimplification :: TCM a -> TCM a #-}
@@ -96,23 +103,36 @@ getSimplification = asksTC envSimplification
 updateAllowedReductions :: (AllowedReductions -> AllowedReductions) -> TCEnv -> TCEnv
 updateAllowedReductions f e = e { envAllowedReductions = f (envAllowedReductions e) }
 
-modifyAllowedReductions :: (AllowedReductions -> AllowedReductions) -> TCM a -> TCM a
+modifyAllowedReductions :: MonadTCEnv m => (AllowedReductions -> AllowedReductions) -> m a -> m a
 modifyAllowedReductions = localTC . updateAllowedReductions
 
-putAllowedReductions :: AllowedReductions -> TCM a -> TCM a
+putAllowedReductions :: MonadTCEnv m => AllowedReductions -> m a -> m a
 putAllowedReductions = modifyAllowedReductions . const
 
 -- | Reduce @Def f vs@ only if @f@ is a projection.
-onlyReduceProjections :: TCM a -> TCM a
-onlyReduceProjections = putAllowedReductions [ProjectionReductions]
+onlyReduceProjections :: MonadTCEnv m => m a -> m a
+onlyReduceProjections = putAllowedReductions $ SmallSet.singleton ProjectionReductions
 
 -- | Allow all reductions except for non-terminating functions (default).
-allowAllReductions :: TCM a -> TCM a
+allowAllReductions :: MonadTCEnv m => m a -> m a
 allowAllReductions = putAllowedReductions allReductions
 
 -- | Allow all reductions including non-terminating functions.
-allowNonTerminatingReductions :: TCM a -> TCM a
-allowNonTerminatingReductions = putAllowedReductions $ [NonTerminatingReductions] ++ allReductions
+allowNonTerminatingReductions :: MonadTCEnv m => m a -> m a
+allowNonTerminatingReductions = putAllowedReductions reallyAllReductions
+
+-- | Allow all reductions when reducing types
+onlyReduceTypes :: MonadTCEnv m => m a -> m a
+onlyReduceTypes = putAllowedReductions $ SmallSet.singleton TypeLevelReductions
+
+-- | Update allowed reductions when working on types
+typeLevelReductions :: MonadTCEnv m => m a -> m a
+typeLevelReductions = modifyAllowedReductions $ \reds -> if
+  | TypeLevelReductions `SmallSet.member` reds ->
+      if NonTerminatingReductions `SmallSet.member` reds
+       then reallyAllReductions
+       else allReductions
+  | otherwise -> reds
 
 -- * Concerning 'envInsideDotPattern'
 
