@@ -1,30 +1,25 @@
 {-# LANGUAGE BangPatterns      #-}
-{-# LANGUAGE CPP                      #-}
 {-# LANGUAGE NondecreasingIndentation #-}
+{-# LANGUAGE TypeFamilies #-} -- for type equality ~
 
 module Agda.TypeChecking.Monad.Signature where
 
 import Prelude hiding (null)
 
-import Control.Arrow (first, second, (***))
-import Control.Applicative hiding (empty)
-
-#if __GLASGOW_HASKELL__ >= 800
 import qualified Control.Monad.Fail as Fail
-#endif
 
 import Control.Monad.State
 import Control.Monad.Reader
 import Control.Monad.Writer
 import Control.Monad.Trans.Maybe
+import Control.Monad.Trans.Identity
 
 import qualified Data.List as List
 import Data.Set (Set)
 import qualified Data.Set as Set
-import Data.Map (Map)
 import qualified Data.Map as Map
+import qualified Data.HashMap.Strict as HMap
 import Data.Maybe
-import Data.Monoid
 
 import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Abstract (Ren, ScopeCopyInfo(..))
@@ -38,7 +33,6 @@ import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Context
-import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
@@ -48,34 +42,35 @@ import Agda.TypeChecking.DropArgs
 import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
-import {-# SOURCE #-} Agda.TypeChecking.Telescope
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Coverage.SplitTree
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
+import {-# SOURCE #-} Agda.TypeChecking.Pretty
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
-import Agda.TypeChecking.Monad.Builtin
 
+import {-# SOURCE #-} Agda.Compiler.Treeless.Erase
+
+import {-# SOURCE #-} Agda.Main
+
+import Agda.Utils.Either
 import Agda.Utils.Except ( ExceptT )
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
-import Agda.Utils.Map as Map
+import Agda.Utils.ListT
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Permutation
-import Agda.Utils.Pretty
 import Agda.Utils.Size
-import qualified Agda.Utils.HashMap as HMap
+import Agda.Utils.Update
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Add a constant to the signature. Lifts the definition to top level.
 addConstant :: QName -> Definition -> TCM ()
 addConstant q d = do
-  reportSLn "tc.signature" 20 $ "adding constant " ++ prettyShow q ++ " to signature"
+  reportSDoc "tc.signature" 20 $ "adding constant " <+> pretty q <+> " to signature"
   tel <- getContextTelescope
   let tel' = replaceEmptyName "r" $ killRange $ case theDef d of
               Constructor{} -> fmap hideOrKeepInstance tel
@@ -88,7 +83,7 @@ addConstant q d = do
                     Just (doms, dom) -> telFromList $ fmap hideOrKeepInstance doms ++ [dom]
               _ -> tel
   let d' = abstract tel' $ d { defName = q }
-  reportSDoc "tc.signature" 60 $ return $ "lambda-lifted definition =" <?> pretty d'
+  reportSDoc "tc.signature" 60 $ "lambda-lifted definition =" <?> pretty d'
   modifySignature $ updateDefinitions $ HMap.insertWith (+++) q d'
   i <- currentOrFreshMutualBlock
   setMutualBlock i q
@@ -126,16 +121,34 @@ modifyFunClauses q f =
 addClauses :: QName -> [Clause] -> TCM ()
 addClauses q cls = do
   tel <- getContextTelescope
-  modifySignature $ updateDefinition q $ updateTheDef
-    $ updateFunClauses      (++ abstract tel cls)
-    . updateFunCopatternLHS (|| isCopatternLHS cls)
+  modifySignature $ updateDefinition q $
+    updateTheDef (updateFunClauses (++ abstract tel cls))
+    . updateDefCopatternLHS (|| isCopatternLHS cls)
 
 mkPragma :: String -> TCM CompilerPragma
 mkPragma s = CompilerPragma <$> getCurrentRange <*> pure s
 
 -- | Add a compiler pragma `{-\# COMPILE <backend> <name> <text> \#-}`
 addPragma :: BackendName -> QName -> String -> TCM ()
-addPragma b q s = modifySignature . updateDefinition q . addCompilerPragma b =<< mkPragma s
+addPragma b q s = ifM erased
+  {- then -} (warning $ PragmaCompileErased b q)
+  {- else -} $ do
+    pragma <- mkPragma s
+    modifySignature $ updateDefinition q $ addCompilerPragma b pragma
+
+  where
+
+  erased :: TCM Bool
+  erased = do
+    def <- theDef <$> getConstInfo q
+    case def of
+      -- If we have a defined symbol, we check whether it is erasable
+      Function{} ->
+        locallyTC      eActiveBackendName (const $ Just b) $
+        locallyTCState stBackends         (const $ builtinBackends) $
+        isErasable q
+     -- Otherwise (Axiom, Datatype, Record type, etc.) we keep it
+      _ -> pure False
 
 getUniqueCompilerPragma :: BackendName -> QName -> TCM (Maybe CompilerPragma)
 getUniqueCompilerPragma backend q = do
@@ -145,7 +158,7 @@ getUniqueCompilerPragma backend q = do
     [p] -> return $ Just p
     (_:p1:_) ->
       setCurrentRange p1 $
-            genericDocError $
+            genericDocError =<< do
                   hang (text ("Conflicting " ++ backend ++ " pragmas for") <+> pretty q <+> "at") 2 $
                        vcat [ "-" <+> pretty (getRange p) | p <- ps ]
 
@@ -183,12 +196,12 @@ addSection m = do
     if (sec == sec') then do
       -- Andreas, 2015-12-02: test/Succeed/Issue1701II.agda
       -- reports a "redundantly adding existing section".
-      reportSLn "tc.section" 10 $ "warning: redundantly adding existing section " ++ prettyShow m
-      reportSLn "tc.section" 60 $ "with content " ++ prettyShow sec
+      reportSDoc "tc.section" 10 $ "warning: redundantly adding existing section" <+> pretty m
+      reportSDoc "tc.section" 60 $ "with content" <+> pretty sec
     else do
-      reportSLn "impossible" 10 $ "overwriting existing section " ++ prettyShow m
-      reportSLn "impossible" 60 $ "of content   " ++ prettyShow sec'
-      reportSLn "impossible" 60 $ "with content " ++ prettyShow sec
+      reportSDoc "impossible" 10 $ "overwriting existing section" <+> pretty m
+      reportSDoc "impossible" 60 $ "of content  " <+> pretty sec'
+      reportSDoc "impossible" 60 $ "with content" <+> pretty sec
       __IMPOSSIBLE__
   -- Add the new section.
   setModuleCheckpoint m
@@ -254,27 +267,27 @@ addDisplayForms x = do
           case unSpine <$> body of
             Just (Def y es) -> do
               let df = Display m es $ DTerm $ Def top $ map Apply args
-              reportSLn "tc.display.section" 20 $ unlines
-                [ "adding display form " ++ prettyShow y ++ " --> " ++ prettyShow top
-                , show df
+              reportSDoc "tc.display.section" 20 $ vcat
+                [ "adding display form " <+> pretty y <+> " --> " <+> pretty top
+                , text (show df)
                 ]
               addDisplayForm y df
               add args top y es
-            Just v          -> noDispForm x $ "not a def body, but " ++ show v
+            Just v          -> noDispForm x $ "not a def body, but " <+> pretty v
             Nothing         -> noDispForm x $ "bad body"
         [] | Constructor{ conSrcCon = h } <- theDef def -> do
               let y  = conName h
                   df = Display 0 [] $ DTerm $ Con (h {conName = top }) ConOSystem []
-              reportSLn "tc.display.section" 20 $ unlines
-                [ "adding display form " ++ prettyShow y ++ " --> " ++ prettyShow top
-                , show df
+              reportSDoc "tc.display.section" 20 $ vcat
+                [ "adding display form" <+> pretty y <+> "-->" <+> pretty top
+                , text (show df)
                 ]
               addDisplayForm y df
         [] -> noDispForm x "no clauses"
         (_:_:_) -> noDispForm x "many clauses"
 
-    noDispForm x reason = reportSLn "tc.display.section" 30 $
-      "no display form from " ++ prettyShow x ++ " because " ++ reason
+    noDispForm x reason = reportSDoc "tc.display.section" 30 $
+      "no display form from" <+> pretty x <+> "because" <+> reason
 
     isVar VarP{} = True
     isVar _      = False
@@ -295,31 +308,29 @@ applySection new ptel old ts ScopeCopyInfo{ renModules = rm, renNames = rd } = d
     -- and if a constructor is copied its datatype needs to be.
     closeConstructors :: Ren QName -> TCM (Ren QName)
     closeConstructors rd = do
-        ds <- List.nub . concat <$> mapM (constructorData . fst) rd
-        cs <- List.nub . concat <$> mapM (dataConstructors . fst) rd
+        ds <- nubOn id . catMaybes <$> mapM (constructorData  . fst) rd
+        cs <- nubOn id . concat    <$> mapM (dataConstructors . fst) rd
         new <- concat <$> mapM rename (ds ++ cs)
-        reportSLn "tc.mod.apply.complete" 30 $
-          "also copying: " ++ prettyShow new
+        reportSDoc "tc.mod.apply.complete" 30 $
+          "also copying: " <+> pretty new
         return $ new ++ rd
       where
         rename :: QName -> TCM (Ren QName)
         rename x =
           case lookup x rd of
-            Nothing -> do y <- freshName_ (prettyShow x)
+            Nothing -> do y <- freshName_ (show $ qnameName x)
                           return [(x, qnameFromList [y])]
             Just{}  -> return []
 
-        constructorData :: QName -> TCM [QName]
+        constructorData :: QName -> TCM (Maybe QName)
         constructorData x = do
-          def <- theDef <$> getConstInfo x
-          return $ case def of
-            Constructor{ conData = d } -> [d]
-            _                          -> []
+          (theDef <$> getConstInfo x) <&> \case
+            Constructor{ conData = d } -> Just d
+            _                          -> Nothing
 
         dataConstructors :: QName -> TCM [QName]
         dataConstructors x = do
-          def <- theDef <$> getConstInfo x
-          return $ case def of
+          (theDef <$> getConstInfo x) <&> \case
             Datatype{ dataCons = cs } -> cs
             Record{ recConHead = h }  -> [conName h]
             _                         -> []
@@ -331,7 +342,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
     forM_ (map fst rd) $ \ q -> do
       when (q `elem` noCopyList) $ typeError (TriedToCopyConstrainedPrim q)
 
-  reportSLn "tc.mod.apply" 10 $ render $ vcat
+  reportSDoc "tc.mod.apply" 10 $ vcat
     [ "applySection"
     , "new  =" <+> pretty new
     , "ptel =" <+> pretty ptel
@@ -354,7 +365,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
 
     argsToUse x = do
       let m = commonParentModule old x
-      reportSLn "tc.mod.apply" 80 $ "Common prefix: " ++ prettyShow m
+      reportSDoc "tc.mod.apply" 80 $ "Common prefix: " <+> pretty m
       size <$> lookupSection m
 
     copyDef :: Args -> (QName, QName) -> TCM ()
@@ -367,18 +378,18 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
       hidings <- map getHiding . telToList <$> lookupSection (qnameModule x)
       let ts' = zipWith setHiding hidings ts
       commonTel <- lookupSection (commonParentModule old $ qnameModule x)
-      reportSLn "tc.mod.apply" 80 $ init $ unlines
-        [ "copyDef " ++ prettyShow x ++ " -> " ++ prettyShow y
-        , "ts' = " ++ prettyShow ts' ]
+      reportSDoc "tc.mod.apply" 80 $ vcat
+        [ "copyDef" <+> pretty x <+> "->" <+> pretty y
+        , "ts' = " <+> pretty ts' ]
       copyDef' ts' np def
       where
         copyDef' ts np d = do
-          reportSLn "tc.mod.apply" 60 $ "making new def for " ++ prettyShow y ++ " from " ++ prettyShow x ++ " with " ++ show np ++ " args " ++ show (defAbstract d)
-          reportSLn "tc.mod.apply" 80 $ init $ unlines
-            [ "args = " ++ show ts'
-            , "old type = " ++ prettyShow (defType d) ]
-          reportSLn "tc.mod.apply" 80 $
-            "new type = " ++ prettyShow t
+          reportSDoc "tc.mod.apply" 60 $ "making new def for" <+> pretty y <+> "from" <+> pretty x <+> "with" <+> text (show np) <+> "args" <+> text (show $ defAbstract d)
+          reportSDoc "tc.mod.apply" 80 $ vcat
+            [ "args = " <+> text (show ts')
+            , "old type = " <+> pretty (defType d) ]
+          reportSDoc "tc.mod.apply" 80 $
+            "new type = " <+> pretty t
           addConstant y =<< nd y
           makeProjection y
           -- Issue1238: the copied def should be an 'instance' if the original
@@ -418,9 +429,11 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                     , defCompiledRep    = noCompiledRep
                     , defInstance       = inst
                     , defCopy           = True
-                    , defMatchable      = False
+                    , defMatchable      = Set.empty
                     , defNoCompilation  = defNoCompilation d
                     , defInjective      = False
+                    , defCopatternLHS   = isCopatternLHS [cl]
+                    , defBlocked        = defBlocked d
                     , theDef            = df }
             oldDef = theDef d
             isCon  = case oldDef of { Constructor{} -> True ; _ -> False }
@@ -458,7 +471,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                          }
                 GeneralizableVar -> return GeneralizableVar
                 _ -> do
-                  (mst, cc) <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
+                  (mst, _, cc) <- compileClauses Nothing [cl] -- Andreas, 2012-10-07 non need for record pattern translation
                   let newDef =
                         set funMacro  (oldDef ^. funMacro) $
                         set funStatic (oldDef ^. funStatic) $
@@ -472,9 +485,8 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         , funTerminates     = Just True
                         , funExtLam         = extlam
                         , funWith           = with
-                        , funCopatternLHS   = isCopatternLHS [cl]
                         }
-                  reportSDoc "tc.mod.apply" 80 $ return $ ("new def for" <+> pretty x) <?> pretty newDef
+                  reportSDoc "tc.mod.apply" 80 $ ("new def for" <+> pretty x) <?> pretty newDef
                   return newDef
 
             cl = Clause { clauseLHSRange  = getRange $ defClauses d
@@ -482,16 +494,19 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         , clauseTel       = EmptyTel
                         , namedClausePats = []
                         , clauseBody      = Just $ dropArgs pars $ case oldDef of
-                            Function{funProjection = Just p} -> projDropParsApply p ProjSystem ts'
+                            Function{funProjection = Just p} -> projDropParsApply p ProjSystem rel ts'
                             _ -> Def x $ map Apply ts'
                         , clauseType      = Just $ defaultArg t
                         , clauseCatchall  = False
+                        , clauseRecursive   = Just False -- definitely not recursive
                         , clauseUnreachable = Just False -- definitely not unreachable
+                        , clauseEllipsis  = NoEllipsis
                         }
               where
                 -- The number of remaining parameters. We need to drop the
                 -- lambdas corresponding to these from the clause body above.
                 pars = max 0 $ maybe 0 (pred . projIndex) proj
+                rel  = getRelevance $ defArgInfo d
 
     {- Example
 
@@ -528,11 +543,11 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
       totalArgs <- argsToUse x
       tel       <- lookupSection x
       let sectionTel =  apply tel $ take totalArgs ts
-      reportSLn "tc.mod.apply" 80 $ "Copying section " ++ prettyShow x ++ " to " ++ prettyShow y
-      reportSLn "tc.mod.apply" 80 $ "  ts           = " ++ List.intercalate "; " (map prettyShow ts)
-      reportSLn "tc.mod.apply" 80 $ "  totalArgs    = " ++ show totalArgs
-      reportSLn "tc.mod.apply" 80 $ "  tel          = " ++ List.intercalate " " (map (fst . unDom) $ telToList tel)  -- only names
-      reportSLn "tc.mod.apply" 80 $ "  sectionTel   = " ++ List.intercalate " " (map (fst . unDom) $ telToList ptel) -- only names
+      reportSDoc "tc.mod.apply" 80 $ "Copying section" <+> pretty x <+> "to" <+> pretty y
+      reportSDoc "tc.mod.apply" 80 $ "  ts           = " <+> mconcat (List.intersperse "; " (map pretty ts))
+      reportSDoc "tc.mod.apply" 80 $ "  totalArgs    = " <+> text (show totalArgs)
+      reportSDoc "tc.mod.apply" 80 $ "  tel          = " <+> text (List.intercalate " " (map (fst . unDom) $ telToList tel))  -- only names
+      reportSDoc "tc.mod.apply" 80 $ "  sectionTel   = " <+> text (List.intercalate " " (map (fst . unDom) $ telToList ptel)) -- only names
       addContext sectionTel $ addSection y
 
 -- | Add a display form to a definition (could be in this or imported signature).
@@ -544,16 +559,16 @@ addDisplayForm x df = do
     {-then-} (modifySignature add)
     {-else-} (stImportsDisplayForms `modifyTCLens` HMap.insertWith (++) x [d])
   whenM (hasLoopingDisplayForm x) $
-    typeError . GenericDocError $ "Cannot add recursive display form for" <+> pretty x
+    typeError . GenericDocError =<< do "Cannot add recursive display form for" <+> pretty x
 
-isLocal :: QName -> TCM Bool
-isLocal x = HMap.member x <$> useTC (stSignature . sigDefinitions)
+isLocal :: ReadTCState m => QName -> m Bool
+isLocal x = HMap.member x <$> useR (stSignature . sigDefinitions)
 
-getDisplayForms :: QName -> TCM [LocalDisplayForm]
+getDisplayForms :: (HasConstInfo m, ReadTCState m) => QName -> m [LocalDisplayForm]
 getDisplayForms q = do
   ds  <- either (const []) defDisplay <$> getConstInfo' q
-  ds1 <- HMap.lookupDefault [] q <$> useTC stImportsDisplayForms
-  ds2 <- HMap.lookupDefault [] q <$> useTC stImportedDisplayForms
+  ds1 <- HMap.lookupDefault [] q <$> useR stImportsDisplayForms
+  ds2 <- HMap.lookupDefault [] q <$> useR stImportedDisplayForms
   ifM (isLocal q) (return $ ds ++ ds1 ++ ds2)
                   (return $ ds1 ++ ds ++ ds2)
 
@@ -572,7 +587,7 @@ chaseDisplayForms q = go Set.empty [q]
 hasLoopingDisplayForm :: QName -> TCM Bool
 hasLoopingDisplayForm q = Set.member q <$> chaseDisplayForms q
 
-canonicalName :: QName -> TCM QName
+canonicalName :: HasConstInfo m => QName -> m QName
 canonicalName x = do
   def <- theDef <$> getConstInfo x
   case def of
@@ -585,7 +600,7 @@ canonicalName x = do
     extract (Def x _)  = x
     extract _          = __IMPOSSIBLE__
 
-sameDef :: QName -> QName -> TCM (Maybe QName)
+sameDef :: HasConstInfo m => QName -> QName -> m (Maybe QName)
 sameDef d1 d2 = do
   c1 <- canonicalName d1
   c2 <- canonicalName d2
@@ -599,7 +614,7 @@ whatInduction c = liftTCM $ do
   mz <- getBuiltinName' builtinIZero
   mo <- getBuiltinName' builtinIOne
   case def of
-    Datatype{ dataInduction = i } -> return i
+    Datatype{}                    -> return Inductive
     Record{} | not (recRecursive def) -> return Inductive
     Record{ recInduction = i    } -> return $ fromMaybe Inductive i
     Constructor{ conInd = i }     -> return i
@@ -636,11 +651,7 @@ sigError f a = \case
 
 class ( Functor m
       , Applicative m
-#if __GLASGOW_HASKELL__ < 800
-      , Monad m
-#else
       , Fail.MonadFail m
-#endif
       , HasOptions m
       , MonadDebug m
       , MonadTCEnv m
@@ -652,15 +663,27 @@ class ( Functor m
       Right d -> return d
       Left (SigUnknown err) -> __IMPOSSIBLE_VERBOSE__ err
       Left SigAbstract      -> __IMPOSSIBLE_VERBOSE__ $
-        "Abstract, thus, not in scope: " ++ prettyShow q
+        "Abstract, thus, not in scope: " ++ show q
 
   -- | Version that reports exceptions:
   getConstInfo' :: QName -> m (Either SigError Definition)
-  getConstInfo' q = Right <$> getConstInfo q
+  -- getConstInfo' q = Right <$> getConstInfo q -- conflicts with default signature
 
   -- | Lookup the rewrite rules with the given head symbol.
   getRewriteRulesFor :: QName -> m RewriteRules
 
+  -- Lifting HasConstInfo through monad transformers:
+
+  default getConstInfo'
+    :: (HasConstInfo n, MonadTrans t, m ~ t n)
+    => QName -> m (Either SigError Definition)
+  getConstInfo' = lift . getConstInfo'
+
+  default getRewriteRulesFor
+    :: (HasConstInfo n, MonadTrans t, m ~ t n)
+    => QName -> m RewriteRules
+  getRewriteRulesFor = lift . getRewriteRulesFor
+
 {-# SPECIALIZE getConstInfo :: QName -> TCM Definition #-}
 
 defaultGetRewriteRulesFor :: (Monad m) => m TCState -> QName -> m RewriteRules
@@ -685,7 +708,7 @@ instance HasConstInfo (TCMT IO) where
   getConstInfo q = getConstInfo' q >>= \case
       Right d -> return d
       Left (SigUnknown err) -> fail err
-      Left SigAbstract      -> notInScope $ qnameToConcrete q
+      Left SigAbstract      -> notInScopeError $ qnameToConcrete q
 
 defaultGetConstInfo
   :: (HasOptions m, MonadDebug m, MonadTCEnv m)
@@ -694,9 +717,9 @@ defaultGetConstInfo st env q = do
     let defs  = st^.(stSignature . sigDefinitions)
         idefs = st^.(stImports . sigDefinitions)
     case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
-        []  -> return $ Left $ SigUnknown $ "Unbound name: " ++ prettyShow q ++ " " ++ showQNameId q
+        []  -> return $ Left $ SigUnknown $ "Unbound name: " ++ show q ++ showQNameId q
         [d] -> mkAbs env d
-        ds  -> __IMPOSSIBLE_VERBOSE__ $ "Ambiguous name: " ++ prettyShow q
+        ds  -> __IMPOSSIBLE_VERBOSE__ $ "Ambiguous name: " ++ show q
     where
       mkAbs env d
         | treatAbstractly' q' env =
@@ -716,49 +739,41 @@ defaultGetConstInfo st env q = do
           dropLastModule q@QName{ qnameModule = m } =
             q{ qnameModule = mnameFromList $ ifNull (mnameToList m) __IMPOSSIBLE__ init }
 
-instance HasConstInfo m => HasConstInfo (MaybeT m) where
-  getConstInfo' = lift . getConstInfo'
-  getRewriteRulesFor = lift . getRewriteRulesFor
+-- HasConstInfo lifts through monad transformers
+-- (see default signatures in HasConstInfo class).
 
-instance HasConstInfo m => HasConstInfo (ExceptT err m) where
-  getConstInfo' = lift . getConstInfo'
-  getRewriteRulesFor = lift . getRewriteRulesFor
-
-instance HasConstInfo m => HasConstInfo (ReaderT r m) where
-  getConstInfo' = lift . getConstInfo'
-  getRewriteRulesFor = lift . getRewriteRulesFor
-
-instance (Monoid w, HasConstInfo m) => HasConstInfo (WriterT w m) where
-  getConstInfo' = lift . getConstInfo'
-  getRewriteRulesFor = lift . getRewriteRulesFor
-
-instance HasConstInfo m => HasConstInfo (StateT s m) where
-  getConstInfo' = lift . getConstInfo'
-  getRewriteRulesFor = lift . getRewriteRulesFor
+instance HasConstInfo m => HasConstInfo (ChangeT m)
+instance HasConstInfo m => HasConstInfo (ExceptT err m)
+instance HasConstInfo m => HasConstInfo (IdentityT m)
+instance HasConstInfo m => HasConstInfo (ListT m)
+instance HasConstInfo m => HasConstInfo (MaybeT m)
+instance HasConstInfo m => HasConstInfo (ReaderT r m)
+instance HasConstInfo m => HasConstInfo (StateT s m)
+instance (Monoid w, HasConstInfo m) => HasConstInfo (WriterT w m)
 
 {-# INLINE getConInfo #-}
 getConInfo :: HasConstInfo m => ConHead -> m Definition
 getConInfo = getConstInfo . conName
 
 -- | Look up the polarity of a definition.
-getPolarity :: QName -> TCM [Polarity]
+getPolarity :: HasConstInfo m => QName -> m [Polarity]
 getPolarity q = defPolarity <$> getConstInfo q
 
 -- | Look up polarity of a definition and compose with polarity
 --   represented by 'Comparison'.
-getPolarity' :: Comparison -> QName -> TCM [Polarity]
+getPolarity' :: HasConstInfo m => Comparison -> QName -> m [Polarity]
 getPolarity' CmpEq  q = map (composePol Invariant) <$> getPolarity q -- return []
 getPolarity' CmpLeq q = getPolarity q -- composition with Covariant is identity
 
 -- | Set the polarity of a definition.
 setPolarity :: QName -> [Polarity] -> TCM ()
 setPolarity q pol = do
-  reportSLn "tc.polarity.set" 20 $
-    "Setting polarity of " ++ prettyShow q ++ " to " ++ prettyShow pol ++ "."
+  reportSDoc "tc.polarity.set" 20 $
+    "Setting polarity of" <+> pretty q <+> "to" <+> pretty pol <> "."
   modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
 
 -- | Look up the forced arguments of a definition.
-getForcedArgs :: QName -> TCM [IsForced]
+getForcedArgs :: HasConstInfo m => QName -> m [IsForced]
 getForcedArgs q = defForced <$> getConstInfo q
 
 -- | Get argument occurrence info for argument @i@ of definition @d@ (never fails).
@@ -797,16 +812,21 @@ getCompiled q = do
     Function{ funTreeless = t } -> t
     _                           -> Nothing
 
+-- | Returns a list of length 'conArity'.
+--   If no erasure analysis has been performed yet, this will be a list of 'False's.
 getErasedConArgs :: QName -> TCM [Bool]
 getErasedConArgs q = do
   def <- getConstInfo q
   case theDef def of
-    Constructor{ conData = d, conPars = np, conErased = es } -> return es
+    Constructor{ conArity, conErased } -> return $
+      fromMaybe (replicate conArity False) conErased
     _ -> __IMPOSSIBLE__
 
 setErasedConArgs :: QName -> [Bool] -> TCM ()
 setErasedConArgs q args = modifyGlobalDefinition q $ updateTheDef $ \case
-    def@Constructor{} -> def{ conErased = args }
+    def@Constructor{ conArity }
+      | length args == conArity -> def{ conErased = Just args }
+      | otherwise               -> __IMPOSSIBLE__
     def -> def   -- no-op for non-constructors
 
 getTreeless :: QName -> TCM (Maybe TTerm)
@@ -859,12 +879,12 @@ getCurrentModuleFreeVars = size <$> (lookupSection =<< currentModule)
 
 --   For annoying reasons the qnameModule of a pattern lambda is not correct
 --   (#2883), so make sure to grab the right module for those.
-getDefModule :: HasConstInfo m => QName -> m ModuleName
-getDefModule f = do
-  def <- getConstInfo f
-  return $ case theDef def of
-    Function{ funExtLam = Just (ExtLamInfo m _) } -> m
-    _                                             -> qnameModule f
+getDefModule :: HasConstInfo m => QName -> m (Either SigError ModuleName)
+getDefModule f = mapRight modName <$> getConstInfo' f
+  where
+    modName def = case theDef def of
+      Function{ funExtLam = Just (ExtLamInfo m _) } -> m
+      _                                             -> qnameModule f
 
 -- | Compute the number of free variables of a defined name. This is the sum of
 --   number of parameters shared with the current module and the number of
@@ -908,22 +928,30 @@ moduleParamsToApply :: (Functor m, Applicative m, HasOptions m,
                         MonadTCEnv m, ReadTCState m, MonadDebug m)
                     => ModuleName -> m Args
 moduleParamsToApply m = do
-  -- Get the correct number of free variables (correctly raised) of @m@.
 
-  reportSLn "tc.sig.param" 90 $ "computing module parameters of " ++ prettyShow m
-  n   <- getModuleFreeVars m
-  tel <- take n . telToList <$> lookupSection m
-  sub <- getModuleParameterSub m
-  verboseS "tc.sig.param" 60 $ do
+  traceSDoc "tc.sig.param" 90 ("computing module parameters of " <+> pretty m) $ do
+
+  -- Jesper, 2020-01-22: If the module parameter substitution for the
+  -- module cannot be found, that likely means we are within a call to
+  -- @inTopContext@. In that case we should provide no arguments for
+  -- the module parameters (see #4383).
+  caseMaybeM (getModuleParameterSub m) (return []) $ \sub -> do
+
+  traceSDoc "tc.sig.param" 60 (do
     cxt <- getContext
-    reportSLn "tc.sig.param" 60 $ unlines $
-      [ "  n    = " ++ show n
-      , "  cxt  = " ++ show (map (fmap fst) cxt)
-      , "  sub  = " ++ show sub
-      ]
+    nest 2 $ vcat
+      [ "cxt  = " <+> prettyTCM (PrettyContext cxt)
+      , "sub  = " <+> pretty sub
+      ]) $ do
+
+  -- Get the correct number of free variables (correctly raised) of @m@.
+  n   <- getModuleFreeVars m
+  traceSDoc "tc.sig.param" 60 (nest 2 $ "n    = " <+> text (show n)) $ do
+  tel <- take n . telToList <$> lookupSection m
+  traceSDoc "tc.sig.param" 60 (nest 2 $ "tel  = " <+> pretty tel) $ do
   unless (size tel == n) __IMPOSSIBLE__
   let args = applySubst sub $ zipWith (\ i a -> var i <$ argFromDom a) (downFrom n) tel
-  reportSLn "tc.sig.param" 60 $ "  args = " ++ show args
+  traceSDoc "tc.sig.param" 60 (nest 2 $ "args = " <+> prettyList_ (map pretty args)) $ do
 
   -- Apply the original ArgInfo, as the hiding information in the current
   -- context might be different from the hiding information expected by @m@.
@@ -950,8 +978,8 @@ moduleParamsToApply m = do
 --   sure generated definitions work properly.
 inFreshModuleIfFreeParams :: TCM a -> TCM a
 inFreshModuleIfFreeParams k = do
-  sub <- getModuleParameterSub =<< currentModule
-  if sub == IdS then k else do
+  msub <- getModuleParameterSub =<< currentModule
+  if msub == Nothing || msub == Just IdS then k else do
     m  <- currentModule
     m' <- qualifyM m . mnameFromList . (:[]) <$>
             freshName_ ("_" :: String)
@@ -970,16 +998,16 @@ instantiateDef d = do
   verboseS "tc.sig.inst" 30 $ do
     ctx <- getContext
     m   <- currentModule
-    reportSLn "tc.sig.inst" 30 $
-      "instDef in " ++ prettyShow m ++ ": " ++ prettyShow (defName d) ++ " " ++
-      unwords (map show $ zipWith (<$) (reverse $ map (fst . unDom) ctx) vs)
+    reportSDoc "tc.sig.inst" 30 $
+      "instDef in" <+> pretty m <> ":" <+> pretty (defName d) <+>
+      text (unwords $ map show $ zipWith (<$) (reverse $ map (fst . unDom) ctx) vs)
   return $ d `apply` vs
 
 instantiateRewriteRule :: (Functor m, HasConstInfo m, HasOptions m,
                            ReadTCState m, MonadTCEnv m, MonadDebug m)
                        => RewriteRule -> m RewriteRule
 instantiateRewriteRule rew = do
-  traceSLn "rewriting" 95 ("instantiating rewrite rule " ++ show (rewName rew) ++ " to the local context.") $ do
+  traceSDoc "rewriting" 95 ("instantiating rewrite rule" <+> pretty (rewName rew) <+> "to the local context.") $ do
   vs  <- freeVarsToApply $ rewName rew
   let rew' = rew `apply` vs
   traceSLn "rewriting" 95 ("instantiated rewrite rule: ") $ do
@@ -1064,7 +1092,7 @@ treatAbstractly' :: QName -> TCEnv -> Bool
 treatAbstractly' q env = case envAbstractMode env of
   ConcreteMode       -> True
   IgnoreAbstractMode -> False
-  AbstractMode       -> not $ current == m || current `isSubModuleOf` m
+  AbstractMode       -> not $ current `isLeChildModuleOf` m
   where
     current = dropAnon $ envCurrentModule env
     m       = dropAnon $ qnameModule q
@@ -1072,19 +1100,18 @@ treatAbstractly' q env = case envAbstractMode env of
 
 -- | Get type of a constant, instantiated to the current context.
 {-# SPECIALIZE typeOfConst :: QName -> TCM Type #-}
-typeOfConst
-  :: ( Functor m, HasConstInfo m, HasOptions m
-     , ReadTCState m, MonadTCEnv m, MonadDebug m )
-  => QName -> m Type
+typeOfConst :: (HasConstInfo m, ReadTCState m) => QName -> m Type
 typeOfConst q = defType <$> (instantiateDef =<< getConstInfo q)
 
 -- | Get relevance of a constant.
-relOfConst :: QName -> TCM Relevance
-relOfConst q = defRelevance <$> getConstInfo q
+{-# SPECIALIZE relOfConst :: QName -> TCM Relevance #-}
+relOfConst :: HasConstInfo m => QName -> m Relevance
+relOfConst q = getRelevance <$> getConstInfo q
 
 -- | Get modality of a constant.
-modalityOfConst :: QName -> TCM Modality
-modalityOfConst q = getModality . defArgInfo <$> getConstInfo q
+{-# SPECIALIZE modalityOfConst :: QName -> TCM Modality #-}
+modalityOfConst :: HasConstInfo m => QName -> m Modality
+modalityOfConst q = getModality <$> getConstInfo q
 
 -- | The number of dropped parameters for a definition.
 --   0 except for projection(-like) functions and constructors.
@@ -1131,12 +1158,8 @@ projectionArgs :: Defn -> Int
 projectionArgs = maybe 0 (max 0 . pred . projIndex) . isProjection_
 
 -- | Check whether a definition uses copatterns.
-usesCopatterns :: QName -> TCM Bool
-usesCopatterns q = do
-  d <- theDef <$> getConstInfo q
-  return $ case d of
-    Function{ funCopatternLHS = b } -> b
-    _ -> False
+usesCopatterns :: (HasConstInfo m) => QName -> m Bool
+usesCopatterns q = defCopatternLHS <$> getConstInfo q
 
 -- | Apply a function @f@ to its first argument, producing the proper
 --   postfix projection if @f@ is a projection.
