@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP          #-}
 {-# LANGUAGE TypeFamilies #-}
 
 -- | Tools to manipulate patterns in abstract syntax
@@ -7,7 +6,6 @@
 module Agda.TypeChecking.Patterns.Abstract where
 
 import qualified Data.List as List
-import Data.Traversable hiding (mapM, sequence)
 import Data.Void
 
 import qualified Agda.Syntax.Abstract as A
@@ -23,9 +21,6 @@ import Agda.Syntax.Position
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 
-import Agda.Utils.Functor
-
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Expand literal integer pattern into suc/zero constructor patterns.
@@ -43,7 +38,7 @@ expandLitPattern p = case asView p of
           info  = A.PatRange r
           cinfo = A.ConPatInfo ConOCon info ConPatEager
           p'    = foldr ($) zero $ List.genericReplicate n suc
-      return $ foldr (A.AsP info) p' (map A.BindName xs)
+      return $ foldr (A.AsP info) p' (map A.mkBindName xs)
   _ -> return p
 
   where
