@@ -1,5 +1,5 @@
-{-# LANGUAGE CPP               #-}
 {-# LANGUAGE NondecreasingIndentation #-}
+{-# LANGUAGE GADTs #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -46,29 +46,22 @@ module Agda.TypeChecking.Rewriting where
 
 import Prelude hiding (null)
 
-import Control.Applicative hiding (empty)
 import Control.Monad
-import Control.Monad.Reader (local, asks)
 
-import Data.Foldable ( Foldable, foldMap )
 import Data.IntSet (IntSet)
 import qualified Data.IntSet as IntSet
 import qualified Data.List as List
-import Data.Monoid
 
 import Agda.Interaction.Options
 
+import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
+import Agda.Syntax.Internal.MetaVars
 
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Env
-import Agda.TypeChecking.EtaExpand
 import Agda.TypeChecking.Free
-import Agda.TypeChecking.Free.Lazy
-import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Conversion
 import qualified Agda.TypeChecking.Positivity.Occurrence as Pos
 import Agda.TypeChecking.Pretty
@@ -76,23 +69,19 @@ import Agda.TypeChecking.Primitive ( getBuiltinName )
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
+import Agda.TypeChecking.Rewriting.Confluence
 import Agda.TypeChecking.Rewriting.NonLinMatch
-import qualified Agda.TypeChecking.Reduce.Monad as Red
+import Agda.TypeChecking.Rewriting.NonLinPattern
 import Agda.TypeChecking.Warnings
 
 import Agda.Utils.Functor
-import qualified Agda.Utils.HashMap as HMap
-import Agda.Utils.Lens
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
-import Agda.Utils.Singleton
 import Agda.Utils.Size
-import Agda.Utils.Lens
-import qualified Agda.Utils.HashMap as HMap
+import qualified Agda.Utils.SmallSet as SmallSet
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 requireOptionRewriting :: TCM ()
@@ -143,15 +132,39 @@ relView t = do
     let [a, b] = fmap snd <$> lastTwo
     return $ Just $ RelView tel delta a b core
 
--- | Add @q : Γ → rel us lhs rhs@ as rewrite rule
+-- | Check the given rewrite rules and add them to the signature.
+addRewriteRules :: [QName] -> TCM ()
+addRewriteRules qs = do
+
+  -- Check the rewrite rules
+  rews <- mapM checkRewriteRule qs
+
+  -- Add rewrite rules to the signature
+  forM_ rews $ \rew -> do
+    let f = rewHead rew
+        matchables = getMatchables rew
+    reportSDoc "rewriting" 10 $
+      "adding rule" <+> prettyTCM (rewName rew) <+>
+      "to the definition of" <+> prettyTCM f
+    reportSDoc "rewriting" 30 $ "matchable symbols: " <+> prettyTCM matchables
+    modifySignature $ addRewriteRulesFor f [rew] matchables
+
+  -- Run confluence check for the new rules
+  -- (should be done after adding all rules, see #3795)
+  whenM (optConfluenceCheck <$> pragmaOptions) $
+    checkConfluenceOfRules rews
+
+
+-- | Check the validity of @q : Γ → rel us lhs rhs@ as rewrite rule
 --   @
 --       Γ ⊢ lhs ↦ rhs : B
 --   @
---   to the signature where @B = A[us/Δ]@.
+--   where @B = A[us/Δ]@.
 --   Remember that @rel : Δ → A → A → Set i@, so
 --   @rel us : (lhs rhs : A[us/Δ]) → Set i@.
-addRewriteRule :: QName -> TCM ()
-addRewriteRule q = do
+--   Returns the checked rewrite rule to be added to the signature.
+checkRewriteRule :: QName -> TCM RewriteRule
+checkRewriteRule q = do
   requireOptionRewriting
   let failNoBuiltin = typeError $ GenericError $
         "Cannot add rewrite rule without prior BUILTIN REWRITE"
@@ -167,8 +180,7 @@ addRewriteRule q = do
       ]
   -- We know that the type of rel is that of a relation.
   relV <- relView =<< do defType <$> getConstInfo rel
-  let RelView _tel delta a _a' _core = -- line break for CPP
-        fromMaybe __IMPOSSIBLE__ relV
+  let RelView _tel delta a _a' _core = fromMaybe __IMPOSSIBLE__ relV
   reportSDoc "rewriting" 30 $ do
     "rewrite relation at type " <+> do
       inTopContext $ prettyTCM (telFromList delta) <+> " |- " <+> do
@@ -180,15 +192,20 @@ addRewriteRule q = do
     , prettyTCM gamma1
     , " |- " <+> do addContext gamma1 $ prettyTCM core
     ]
-  let failureWrongTarget = typeError . GenericDocError =<< hsep
+  let failureWrongTarget :: TCM a
+      failureWrongTarget = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " does not target rewrite relation" ]
-  let failureMetas       = typeError . GenericDocError =<< hsep
+  let failureMetas :: TCM a
+      failureMetas       = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule, since it contains unsolved meta variables" ]
-  let failureNotDefOrCon = typeError . GenericDocError =<< hsep
+  let failureNotDefOrCon :: TCM a
+      failureNotDefOrCon = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule, since the left-hand side is neither a defined symbol nor a constructor" ]
-  let failureFreeVars xs = typeError . GenericDocError =<< hsep
+  let failureFreeVars :: IntSet -> TCM a
+      failureFreeVars xs = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule, since the following variables are not bound by the left hand side: " , prettyList_ (map (prettyTCM . var) $ IntSet.toList xs) ]
-  let failureIllegalRule = typeError . GenericDocError =<< hsep
+  let failureIllegalRule :: TCM a
+      failureIllegalRule = typeError . GenericDocError =<< hsep
         [ prettyTCM q , " is not a legal rewrite rule" ]
 
   -- Check that type of q targets rel.
@@ -214,13 +231,14 @@ addRewriteRule q = do
       -- 2017-06-18, Jesper: Unfold inlined definitions on the LHS.
       -- This is necessary to replace copies created by imports by their
       -- original definition.
-      lhs <- modifyAllowedReductions (const [InlineReductions]) $ normalise lhs
+      lhs <- modifyAllowedReductions (const $ SmallSet.singleton InlineReductions) $ normalise lhs
 
       -- Find head symbol f of the lhs, its type and its arguments.
       (f , hd , t , es) <- case lhs of
         Def f es -> do
-          t <- defType <$> getConstInfo f
-          return (f , Def f , t , es)
+          def <- getConstInfo f
+          checkAxFunOrCon f def
+          return (f , Def f , defType def , es)
         Con c ci vs -> do
           let hd = Con c ci
           ~(Just ((_ , _ , pars) , t)) <- getFullyAppliedConType c $ unDom b
@@ -230,7 +248,7 @@ addRewriteRule q = do
 
       ifNotAlreadyAdded f $ do
 
-      rew <- addContext gamma1 $ do
+      addContext gamma1 $ do
         -- Normalize lhs args: we do not want to match redexes.
         es <- normalise es
 
@@ -255,24 +273,17 @@ addRewriteRule q = do
           "variables free in the rewrite rule: " <+> text (show freeVars)
         unlessNull (freeVars IntSet.\\ boundVars) failureFreeVars
 
-        return $ RewriteRule q gamma f ps rhs (unDom b)
+        let rew = RewriteRule q gamma f ps rhs (unDom b)
 
-      reportSDoc "rewriting" 10 $ vcat
-        [ "considering rewrite rule " , prettyTCM rew ]
-      reportSDoc "rewriting" 90 $ vcat
-        [ "considering rewrite rule" , text (show rew) ]
+        reportSDoc "rewriting" 10 $ vcat
+          [ "checked rewrite rule" , prettyTCM rew ]
+        reportSDoc "rewriting" 90 $ vcat
+          [ "checked rewrite rule" , text (show rew) ]
 
-      -- NO LONGER WORKS:
-      -- -- Check whether lhs can be rewritten with itself.
-      -- -- Otherwise, there are unbound variables in either gamma or rhs.
-      -- addContext gamma $
-      --   unlessM (isJust <$> runReduceM (rewriteWith (Just b) lhs rew)) $
-      --     failureFreeVars
-
-      -- Add rewrite rule gamma ⊢ lhs ↦ rhs : b for f.
-      addRewriteRules f [rew]
+        return rew
 
     _ -> failureWrongTarget
+
   where
     checkNoLhsReduction :: QName -> Elims -> TCM ()
     checkNoLhsReduction f es = do
@@ -293,14 +304,28 @@ addRewriteRule q = do
           unless ok fail
         _ -> fail
 
-    ifNotAlreadyAdded :: QName -> TCM () -> TCM ()
+    checkAxFunOrCon :: QName -> Definition -> TCM ()
+    checkAxFunOrCon f def = case theDef def of
+      Axiom{}        -> return ()
+      Function{}     -> return ()
+      Constructor{}  -> return ()
+      AbstractDefn{} -> return ()
+      Primitive{}    -> return () -- TODO: is this fine?
+      _              -> typeError . GenericDocError =<< hsep
+        [ prettyTCM q , " is not a legal rewrite rule, since the head symbol"
+        , prettyTCM f , "is not a postulate, a function, or a constructor"
+        ]
+
+    ifNotAlreadyAdded :: QName -> TCM RewriteRule -> TCM RewriteRule
     ifNotAlreadyAdded f cont = do
       rews <- getRewriteRulesFor f
       -- check if q is already an added rewrite rule
-      if any ((q ==) . rewName) rews then
-        genericWarning =<< do
-          "Rewrite rule " <+> prettyTCM q <+> " has already been added"
-      else cont
+      case List.find ((q ==) . rewName) rews of
+        Just rew -> do
+          genericWarning =<< do
+            "Rewrite rule " <+> prettyTCM q <+> " has already been added"
+          return rew
+        Nothing -> cont
 
     usedArgs :: [Pos.Occurrence] -> IntSet
     usedArgs occs = IntSet.fromList $ map snd $ usedIxs
@@ -320,20 +345,13 @@ addRewriteRule q = do
           Var i [] | i < k -> (i :) <$> loop vs
           _                -> errorNotGeneral
 
+        errorNotGeneral :: TCM a
         errorNotGeneral = typeError . GenericDocError =<< vcat
             [ prettyTCM q <+> text " is not a legal rewrite rule, since the constructor parameters are not fully general:"
             , nest 2 $ text "Constructor: " <+> prettyTCM c
             , nest 2 $ text "Parameters: " <+> prettyList (map prettyTCM vs)
             ]
 
--- | Append rewrite rules to a definition.
-addRewriteRules :: QName -> RewriteRules -> TCM ()
-addRewriteRules f rews = do
-  reportSDoc "rewriting" 10 $ "rewrite rule ok, adding it to the definition of " <+> prettyTCM f
-  let matchables = getMatchables rews
-  reportSDoc "rewriting" 30 $ "matchable symbols: " <+> prettyTCM matchables
-  modifySignature $ addRewriteRulesFor f rews matchables
-
 -- | @rewriteWith t f es rew@ where @f : t@
 --   tries to rewrite @f es@ with @rew@, returning the reduct if successful.
 rewriteWith :: Type
@@ -376,7 +394,7 @@ rewrite block v rules es = do
         -- Andreas, 2018-09-08, issue #3211:
         -- discount module parameters for constructor heads
       _ -> __IMPOSSIBLE__
-    loop block t rules =<< instantiateFull' es
+    loop block t rules =<< instantiateFull' es -- TODO: remove instantiateFull?
   else
     return $ NoReduction (block $> v `applyE` es)
   where
@@ -397,71 +415,5 @@ rewrite block v rules es = do
             Right w               -> return $ YesReduction YesSimplification $ w `applyE` es2
      | otherwise = loop (block `mappend` NotBlocked Underapplied ()) t rews es
 
-
-------------------------------------------------------------------------
--- * Auxiliary functions
-------------------------------------------------------------------------
-
-class NLPatVars a where
-  nlPatVarsUnder :: Int -> a -> IntSet
-
-  nlPatVars :: a -> IntSet
-  nlPatVars = nlPatVarsUnder 0
-
-instance (Foldable f, NLPatVars a) => NLPatVars (f a) where
-  nlPatVarsUnder k = foldMap $ nlPatVarsUnder k
-
-instance NLPatVars NLPType where
-  nlPatVarsUnder k (NLPType l a) = nlPatVarsUnder k l `IntSet.union` nlPatVarsUnder k a
-
-instance NLPatVars NLPat where
-  nlPatVarsUnder k p =
-    case p of
-      PVar i _  -> singleton $ i - k
-      PDef _ es -> nlPatVarsUnder k es
-      PWild     -> empty
-      PLam _ p' -> nlPatVarsUnder (k+1) $ unAbs p'
-      PPi a b   -> nlPatVarsUnder k a `IntSet.union` nlPatVarsUnder (k+1) (unAbs b)
-      PBoundVar _ es -> nlPatVarsUnder k es
-      PTerm{}   -> empty
-
-rewArity :: RewriteRule -> Int
-rewArity = length . rewPats
-
--- | Get all symbols that a rewrite rule matches against
-class GetMatchables a where
-  getMatchables :: a -> [QName]
-
-instance (Foldable f, GetMatchables a) => GetMatchables (f a) where
-  getMatchables = foldMap getMatchables
-
-instance GetMatchables NLPat where
-  getMatchables p =
-    case p of
-      PVar _ _       -> empty
-      PWild          -> empty
-      PDef f _       -> singleton f
-      PLam _ x       -> empty
-      PPi a b        -> empty
-      PBoundVar i es -> empty
-      PTerm _        -> empty -- should be safe (I hope)
-
-instance GetMatchables RewriteRule where
-  getMatchables = getMatchables . rewPats
-
--- Only computes free variables that are not bound (i.e. those in a PTerm)
-instance Free NLPat where
-  freeVars' p = case p of
-    PVar _ _ -> mempty
-    PWild -> mempty
-    PDef _ es -> freeVars' es
-    PLam _ u -> freeVars' u
-    PPi a b -> freeVars' (a,b)
-    PBoundVar _ es -> freeVars' es
-    PTerm t -> freeVars' t
-
-instance Free NLPType where
-  freeVars' (NLPType l a) =
-    ifM ((IgnoreNot ==) <$> asks feIgnoreSorts)
-      {- then -} (freeVars' (l, a))
-      {- else -} (freeVars' a)
+    rewArity :: RewriteRule -> Int
+    rewArity = length . rewPats
