@@ -1,15 +1,15 @@
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE MultiWayIf #-}
 {-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Rules.Data where
 
-import Control.Monad
+import Prelude hiding (null)
 
-import Data.List (genericTake)
-import Data.Maybe (fromMaybe, catMaybes, isJust)
-import Control.Monad.Trans.Maybe
+import Control.Monad
 import Control.Monad.Trans
+import Control.Monad.Trans.Maybe
+
+import Data.Foldable (traverse_)
+import Data.Maybe (fromMaybe, catMaybes, isJust)
 import Data.Set (Set)
 import qualified Data.Set as Set
 
@@ -26,7 +26,6 @@ import Agda.Syntax.Fixity
 import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin -- (primLevel)
-import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Generalize
@@ -34,26 +33,24 @@ import Agda.TypeChecking.Implicit
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Names
 import Agda.TypeChecking.Reduce
+import Agda.TypeChecking.Positivity.Occurrence (Occurrence(StrictPos))
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Primitive hiding (Nat)
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Forcing
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Telescope
-import Agda.TypeChecking.ProjectionLike
 
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Term ( isType_ )
 
-import Agda.Interaction.Options
-
 import Agda.Utils.Except
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Null
 import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
@@ -62,7 +59,7 @@ import Agda.Utils.Impossible
 
 -- | Type check a datatype definition. Assumes that the type has already been
 --   checked.
-checkDataDef :: Info.DefInfo -> QName -> UniverseCheck -> A.DataDefParams -> [A.Constructor] -> TCM ()
+checkDataDef :: A.DefInfo -> QName -> UniverseCheck -> A.DataDefParams -> [A.Constructor] -> TCM ()
 checkDataDef i name uc (A.DataDefParams gpars ps) cs =
     traceCall (CheckDataDef (getRange name) name ps cs) $ do
 
@@ -124,7 +121,10 @@ checkDataDef i name uc (A.DataDefParams gpars ps) cs =
             let s' = case s of
                   Prop l -> Type l
                   _      -> s
-            whenM withoutKOption $ checkIndexSorts s' ixTel
+            -- Andreas, 2019-07-16, issue #3916:
+            -- NoUniverseCheck should also disable the index sort check!
+            unless (uc == NoUniverseCheck) $
+              whenM withoutKOption $ checkIndexSorts s' ixTel
 
             reportSDoc "tc.data.sort" 20 $ vcat
               [ "checking datatype" <+> prettyTCM name
@@ -143,7 +143,6 @@ checkDataDef i name uc (A.DataDefParams gpars ps) cs =
             let dataDef = Datatype
                   { dataPars       = npars
                   , dataIxs        = nofIxs
-                  , dataInduction  = Inductive
                   , dataClause     = Nothing
                   , dataCons       = []     -- Constructors are added later
                   , dataSort       = s
@@ -152,7 +151,7 @@ checkDataDef i name uc (A.DataDefParams gpars ps) cs =
                   , dataPathCons   = []     -- Path constructors are added later
                   }
 
-            escapeContext npars $ do
+            escapeContext __IMPOSSIBLE__ npars $ do
               addConstant name $
                 defaultDefn defaultArgInfo name t dataDef
                 -- polarity and argOcc.s determined by the positivity checker
@@ -213,10 +212,16 @@ checkConstructor d uc tel nofIxs s con@(A.Axiom _ i ai Nothing c e) =
           Relevant   -> return ()
           Irrelevant -> typeError $ GenericError $ "Irrelevant constructors are not supported"
           NonStrict  -> typeError $ GenericError $ "Shape-irrelevant constructors are not supported"
+        case getQuantity ai of
+          Quantityω{} -> return ()
+          Quantity0{} -> typeError $ GenericError $ "Erased constructors are not supported"
+          Quantity1{} -> typeError $ GenericError $ "Quantity-restricted constructors are not supported"
         -- check that the type of the constructor is well-formed
         (t, isPathCons) <- checkConstructorType e d
-        -- compute which constructor arguments are forced
-        forcedArgs <- computeForcingAnnotations c t
+        -- compute which constructor arguments are forced (only point constructors)
+        forcedArgs <- if isPathCons == PointCons
+                      then computeForcingAnnotations c t
+                      else return []
         -- check that the sort (universe level) of the constructor type
         -- is contained in the sort of the data type
         -- (to avoid impredicative existential types)
@@ -237,28 +242,33 @@ checkConstructor d uc tel nofIxs s con@(A.Axiom _ i ai Nothing c e) =
         -- of the datatype in an empty context (c.f. getContextSize above).
         params <- getContextTelescope
 
+        -- Cannot compose indexed inductive types yet.
+        (con, comp, projNames) <- if nofIxs /= 0 || (Info.defAbstract i == AbstractDef)
+          then return (ConHead c Inductive [], emptyCompKit, Nothing)
+          else do
+            -- Name for projection of ith field of constructor c is just c-i
+            names <- forM [0 .. size fields - 1] $ \ i ->
+              freshAbstractQName'_ $ P.prettyShow (A.qnameName c) ++ "-" ++ show i
+
+            -- nofIxs == 0 means the data type can be reconstructed
+            -- by appling the QName d to the parameters.
+            let dataT = El s $ Def d $ map Apply $ teleArgs params
+
+            reportSDoc "tc.data.con.comp" 5 $ inTopContext $ vcat $
+              [ "params =" <+> pretty params
+              , "dataT  =" <+> pretty dataT
+              , "fields =" <+> pretty fields
+              , "names  =" <+> pretty names
+              ]
+
+            let con = ConHead c Inductive $ zipWith (<$) names $ map argFromDom $ telToList fields
+
+            defineProjections d con params names fields dataT
+            comp <- inTopContext $ defineCompData d con params names fields dataT boundary
+            return (con, comp, Just names)
+
         -- add parameters to constructor type and put into signature
-        let con = ConHead c Inductive [] -- data constructors have no projectable fields and are always inductive
-        escapeContext (size tel) $ do
-
-          cnames <- if nofIxs /= 0 || (Info.defAbstract i == AbstractDef) then return (emptyCompKit, Nothing) else do
-            inTopContext $ do
-              names <- forM [0 .. size fields - 1] (\ i -> freshAbstractQName'_ (P.prettyShow (A.qnameName c) ++ "-" ++ show i))
-
-              -- nofIxs == 0 means the data type can be reconstructed
-              -- by appling the QName d to the parameters.
-              dataT <- El s <$> (pure $ Def d $ map Apply $ teleArgs params)
-
-              reportSDoc "tc.data.con.comp" 5 $ vcat $
-                [ "params =" <+> pretty params
-                , "dataT  =" <+> pretty dataT
-                , "fields =" <+> pretty fields
-                , "names  =" <+> pretty names
-                ]
-
-              defineProjections d con params names fields dataT
-              comp <- defineCompData d con params names fields dataT boundary
-              return $ (comp, Just names)
+        escapeContext __IMPOSSIBLE__ (size tel) $ do
 
           addConstant c $
             defaultDefn defaultArgInfo c (telePi tel t) $ Constructor
@@ -268,18 +278,23 @@ checkConstructor d uc tel nofIxs s con@(A.Axiom _ i ai Nothing c e) =
               , conData   = d
               , conAbstr  = Info.defAbstract i
               , conInd    = Inductive
-              , conComp   = cnames
+              , conComp   = comp
+              , conProj   = projNames
               , conForced = forcedArgs
-              , conErased = []  -- computed during compilation to treeless
+              , conErased = Nothing  -- computed during compilation to treeless
               }
 
-          case snd cnames of
-            Nothing -> return ()
-            Just names -> mapM_ makeProjection names
-
         -- Add the constructor to the instance table, if needed
-        when (Info.defInstance i == InstanceDef) $ do
-          addNamedInstance c d
+        case Info.defInstance i of
+          InstanceDef _r -> setCurrentRange c $ do
+            -- Including the range of the @instance@ keyword, like
+            -- @(getRange (r,c))@, does not produce good results.
+            -- Andreas, 2020-01-28, issue #4360:
+            -- Use addTypedInstance instead of addNamedInstance
+            -- to detect unusable instances.
+            addTypedInstance c t
+            -- addNamedInstance c d
+          NotInstanceDef -> pure ()
 
         return isPathCons
 
@@ -355,8 +370,8 @@ defineCompData d con params names fsT t boundary = do
     -- Δ^I, i : I |- sub Δ : Δ
     sub tel = parallelS [ var n `apply` [Arg defaultArgInfo $ var 0] | n <- [1..size tel] ]
     withArgInfo tel = zipWith Arg (map domInfo . telToList $ tel)
-    defineTranspOrHCompD cmd d con params names fsT t boundary
-      = do
+
+    defineTranspOrHCompD cmd d con params names fsT t boundary = do
       let project = (\ t p -> apply (Def p []) [argN t])
       stuff <- defineTranspOrHCompForFields cmd
                  (guard (not $ null boundary) >> (Just $ Con con ConOSystem $ teleElims fsT boundary))
@@ -495,7 +510,7 @@ defineCompData d con params names fsT t boundary = do
 
         -- Δ.Φ ⊢ u = Con con ConOSystem $ teleElims fsT boundary : R δ
 --        u = Con con ConOSystem $ teleElims fsT boundary
-        up = ConP con (ConPatternInfo Nothing False Nothing False) $
+        up = ConP con (ConPatternInfo defaultPatternInfo False False Nothing False) $
                telePatterns (d0 `applySubst` fsT) (liftS (size fsT) d0 `applySubst` boundary)
 --        gamma' = telFromList $ take (size gamma - 1) $ telToList gamma
 
@@ -506,32 +521,25 @@ defineCompData d con params names fsT t boundary = do
 --        (tel',theta) = (abstract gamma' (d0 `applySubst` fsT), (liftS (size fsT) d0 `applySubst` u) `consS` raiseS (size fsT))
 
       let
-        clause | null boundary
-           = Clause
-            { clauseTel = gamma
-            , clauseType = Just . argN $ ty
-            , namedClausePats = teleNamedArgs gamma
-            , clauseFullRange = noRange
-            , clauseLHSRange  = noRange
-            , clauseCatchall = False
-            , clauseBody = Just $ body
-            , clauseUnreachable = Just False
-            }
-
-               | otherwise
-           = Clause
-            { clauseTel = gamma
-            , clauseType = Just . argN $ ty
-            , namedClausePats = take (size gamma - size fsT) (teleNamedArgs gamma) ++ [argN $ unnamed $ up]
-            , clauseFullRange = noRange
-            , clauseLHSRange  = noRange
-            , clauseCatchall = False
-            , clauseBody = Just $ body
-            , clauseUnreachable = Just False
-            }
+        pats | null boundary = teleNamedArgs gamma
+             | otherwise     = take (size gamma - size fsT) (teleNamedArgs gamma) ++ [argN $ unnamed $ up]
+        clause = Clause
+          { clauseTel         = gamma
+          , clauseType        = Just . argN $ ty
+          , namedClausePats   = pats
+          , clauseFullRange   = noRange
+          , clauseLHSRange    = noRange
+          , clauseCatchall    = False
+          , clauseBody        = Just $ body
+          , clauseRecursive   = Nothing
+              -- Andreas 2020-02-06 TODO
+              -- Or: Just False;  is it known to be non-recursive?
+          , clauseUnreachable = Just False
+          , clauseEllipsis    = NoEllipsis
+          }
         cs = [clause]
       addClauses theName cs
-      (mst, cc) <- inTopContext (compileClauses Nothing cs)
+      (mst, _, cc) <- inTopContext (compileClauses Nothing cs)
       whenJust mst $ setSplitTree theName
       setCompiledClauses theName cc
       setTerminates theName True
@@ -543,6 +551,10 @@ defineCompData d con params names fsT t boundary = do
       if all isJust xs then m else return Nothing
 
 -- Andrea: TODO handle Irrelevant fields somehow.
+-- | Define projections for non-indexed data types (families don't work yet).
+--   Of course, these projections are partial functions in general.
+--
+--   Precondition: we are in the context Γ of the data type parameters.
 defineProjections :: QName      -- datatype name
                   -> ConHead
                   -> Telescope  -- Γ parameters
@@ -550,49 +562,64 @@ defineProjections :: QName      -- datatype name
                   -> Telescope  -- Γ ⊢ Φ field types
                   -> Type       -- Γ ⊢ T target type
                   -> TCM ()
-defineProjections dataname con params names fsT t = do
+defineProjections dataName con params names fsT t = do
   let
     -- Γ , (d : T) ⊢ Φ[n ↦ proj n d]
     fieldTypes = ([ Def f [] `apply` [argN $ var 0] | f <- reverse names ] ++# raiseS 1) `applySubst`
                     flattenTel fsT  -- Γ , Φ ⊢ Φ
     -- ⊢ Γ , (d : T)
-    projTel = abstract params (ExtendTel (defaultDom t) (Abs "d" EmptyTel))
+    projTel    = abstract params (ExtendTel (defaultDom t) (Abs "d" EmptyTel))
+    np         = size params
+
   forM_ (zip3 (downFrom (size fieldTypes)) names fieldTypes) $ \ (i,projName,ty) -> do
     let
       projType = abstract projTel <$> ty
-
-    inTopContext $ do
-      reportSDoc "tc.data.proj" 20 $ sep [ "proj" <+> prettyTCM (i,ty) , nest 2 $ prettyTCM projType ]
-
-    let
-      cpi  = ConPatternInfo Nothing False (Just $ argN $ raise (size fsT) t) False
-      conp = defaultArg $ ConP con cpi $ teleNamedArgs fsT
-      clause = Clause
-          { clauseTel = abstract params fsT
-          , clauseType = Just . argN $ ([Con con ConOSystem (map Apply $ teleArgs fsT)] ++# raiseS (size fsT)) `applySubst` unDom ty
-          , namedClausePats = raise (size fsT) (teleNamedArgs params) ++ [Named Nothing <$> conp]
-          , clauseFullRange = noRange
-          , clauseLHSRange  = noRange
-          , clauseCatchall = False
-          , clauseBody = Just $ var i
+      cpi    = ConPatternInfo defaultPatternInfo False False (Just $ argN $ raise (size fsT) t) False
+      conp   = defaultNamedArg $ ConP con cpi $ teleNamedArgs fsT
+      sigma  = Con con ConOSystem (map Apply $ teleArgs fsT) `consS` raiseS (size fsT)
+      clause = empty
+          { clauseTel         = abstract params fsT
+          , namedClausePats   = [ conp ]
+          , clauseBody        = Just $ var i
+          , clauseType        = Just $ argN $ applySubst sigma $ unDom ty
+          , clauseRecursive   = Just False  -- non-recursive
           , clauseUnreachable = Just False
           }
 
+    reportSDoc "tc.data.proj" 20 $ inTopContext $ sep
+      [ "proj" <+> prettyTCM (i,ty)
+      , nest 2 $ sep [ prettyTCM projName, ":", prettyTCM projType ]
+      ]
+
+    -- Andreas, 2020-02-14, issue #4437
+    -- Define data projections as projection-like from the start.
     noMutualBlock $ do
-      let cs = [clause]
-      (mst , cc) <- inTopContext $ compileClauses Nothing cs
+      let cs = [ clause ]
+      (mst, _, cc) <- compileClauses Nothing cs
       let fun = emptyFunction
-                { funClauses = cs
+                { funClauses    = cs
+                , funCompiled   = Just cc
+                , funSplitTree  = mst
+                , funProjection = Just $ Projection
+                    { projProper   = Nothing
+                    , projOrig     = projName
+                    , projFromType = Arg (getArgInfo ty) dataName
+                    , projIndex    = np + 1
+                    , projLams     = ProjLams $ map (argFromDom . fmap fst) $ telToList projTel
+                    }
+                , funMutual     = Just []
                 , funTerminates = Just True
-                , funCompiled = Just cc
-                , funSplitTree = mst
-                , funMutual = Just []
                 }
-      addConstant projName $
+      inTopContext $ addConstant projName $
         (defaultDefn defaultArgInfo projName (unDom projType) fun)
-          { defNoCompilation = True }
-      inTopContext $ do
-        reportSDoc "tc.data.proj.fun" 60 $ sep [ "proj" <+> prettyTCM i, nest 2 $ pretty fun ]
+          { defNoCompilation  = True
+          , defArgOccurrences = [StrictPos]
+          }
+
+      reportSDoc "tc.data.proj.fun" 60 $ inTopContext $ vcat
+        [ "proj" <+> prettyTCM i
+        , nest 2 $ pretty fun
+        ]
 
 
 freshAbstractQName'_ :: String -> TCM QName
@@ -1029,7 +1056,7 @@ bindParameters npars [] t ret =
                     , text (absName b) <+> text ":" <+> prettyTCM (unDom a) ]
     _ -> __IMPOSSIBLE__
 
-bindParameters npars par@(A.DomainFull (A.TBind _ xs e) : bs) a ret =
+bindParameters npars par@(A.DomainFull (A.TBind _ _ xs e) : bs) a ret =
   setCurrentRange par $
   typeError . GenericDocError =<< do
     let s | length xs > 1 = "s"
@@ -1038,16 +1065,16 @@ bindParameters npars par@(A.DomainFull (A.TBind _ xs e) : bs) a ret =
 
 bindParameters _ (A.DomainFull A.TLet{} : _) _ _ = __IMPOSSIBLE__
 
-bindParameters _ (par@(A.DomainFree arg) : ps) _ _
+bindParameters _ (par@(A.DomainFree _ arg) : ps) _ _
   | getModality arg /= defaultModality = setCurrentRange par $
      typeError . GenericDocError =<< do
        text "Unexpected modality/relevance annotation in" <+> prettyA par
 
-bindParameters npars ps0@(par@(A.DomainFree arg) : ps) t ret = do
+bindParameters npars ps0@(par@(A.DomainFree _ arg) : ps) t ret = do
   let x          = namedArg arg
       TelV tel _ = telView' t
   case insertImplicit arg $ telToList tel of
-    NoInsertNeeded -> continue ps $ A.unBind x
+    NoInsertNeeded -> continue ps $ A.unBind $ A.binderName x
     ImpInsert _    -> continue ps0 =<< freshName_ (absName b)
     BadImplicits   -> setCurrentRange par $
      typeError . GenericDocError =<< do
@@ -1210,8 +1237,7 @@ isCoinductive t = do
         Axiom       {} -> return (Just False)
         DataOrRecSig{} -> return Nothing
         Function    {} -> return Nothing
-        Datatype    { dataInduction = CoInductive } -> return (Just True)
-        Datatype    { dataInduction = Inductive   } -> return (Just False)
+        Datatype    {} -> return (Just False)
         Record      {  recInduction = Just CoInductive } -> return (Just True)
         Record      {  recInduction = _                } -> return (Just False)
         GeneralizableVar{} -> __IMPOSSIBLE__
@@ -1227,4 +1253,4 @@ isCoinductive t = do
     Sort  {} -> return (Just False)
     MetaV {} -> return Nothing
     DontCare{} -> __IMPOSSIBLE__
-    Dummy s    -> __IMPOSSIBLE_VERBOSE__ s
+    Dummy s _  -> __IMPOSSIBLE_VERBOSE__ s
