@@ -1,5 +1,3 @@
-{-# LANGUAGE CPP                #-}
-{-# LANGUAGE DeriveDataTypeable #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -9,7 +7,7 @@ import Prelude hiding (mapM)
 
 import Control.Monad.Reader hiding (mapM)
 import Control.Monad.State.Strict (gets, modify)
-import Control.Exception
+
 
 import Data.Array.IArray
 import Data.Word
@@ -17,18 +15,22 @@ import qualified Data.Foldable as Fold
 import Data.Hashable
 import qualified Data.HashTable.IO as H
 import Data.Int (Int32)
-import Data.Maybe
+
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.IntSet as IntSet
 import Data.IntSet (IntSet)
+import Data.List.NonEmpty (NonEmpty(..), nonEmpty)
+import qualified Data.List.NonEmpty as NonEmpty
 import qualified Data.Set as Set
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
 import Data.Text.Lazy (Text)
 import Data.Traversable ( mapM )
 import Data.Typeable
+import Data.HashMap.Strict (HashMap)
+import qualified Data.HashMap.Strict as HMap
 
 import Data.Void
 
@@ -36,8 +38,8 @@ import Agda.Syntax.Common
 import Agda.Syntax.Concrete.Name as C
 import qualified Agda.Syntax.Concrete as C
 import qualified Agda.Syntax.Abstract as A
+import Agda.Syntax.Info
 import Agda.Syntax.Position as P
-import Agda.Syntax.Fixity
 import Agda.Syntax.Notation
 import Agda.Syntax.Literal
 import Agda.Interaction.FindFile
@@ -46,14 +48,11 @@ import Agda.TypeChecking.Serialise.Base
 
 import Agda.Utils.BiMap (BiMap)
 import qualified Agda.Utils.BiMap as BiMap
-import Agda.Utils.HashMap (HashMap)
-import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.FileName
 import Agda.Utils.Maybe
-import Agda.Utils.NonemptyList
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Trie (Trie(..))
-import qualified Agda.Utils.Trie as Trie
+
 
 import Agda.Utils.Except
 
@@ -62,7 +61,6 @@ import qualified Agda.Utils.Empty as Empty
 
 import Agda.Utils.WithDefault
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 instance {-# OVERLAPPING #-} EmbPrj String where
@@ -204,7 +202,7 @@ instance EmbPrj AbsolutePath where
     modify $ \s -> s { modFile = mf }
     case r of
       Left err -> throwError $ findErrorToTypeError m err
-      Right f  -> return f
+      Right f  -> return (srcFilePath f)
 
 instance EmbPrj a => EmbPrj (Position' a) where
   icod_ (P.Pn file pos line col) = icodeN' P.Pn file pos line col
@@ -235,9 +233,9 @@ instance {-# OVERLAPPABLE #-} EmbPrj a => EmbPrj [a] where
 --                            valu [x, xs] = valu2 (:) x xs
 --                            valu _       = malformed
 
-instance EmbPrj a => EmbPrj (NonemptyList a) where
-  icod_ = icod_ . toList
-  value = listCaseNe malformed return <=< value
+instance EmbPrj a => EmbPrj (NonEmpty a) where
+  icod_ = icod_ . NonEmpty.toList
+  value = maybe malformed return . nonEmpty <=< value
 
 instance (Ord a, Ord b, EmbPrj a, EmbPrj b) => EmbPrj (BiMap a b) where
   icod_ m = icode (BiMap.toList m)
@@ -333,7 +331,7 @@ instance (EmbPrj a, EmbPrj b) => EmbPrj (ImportedName' a b) where
     valu [2, a] = valuN ImportedName a
     valu _ = malformed
 
-instance EmbPrj Agda.Syntax.Fixity.Associativity where
+instance EmbPrj Associativity where
   icod_ LeftAssoc  = icodeN' LeftAssoc
   icod_ RightAssoc = icodeN 1 RightAssoc
   icod_ NonAssoc   = icodeN 2 NonAssoc
@@ -344,7 +342,7 @@ instance EmbPrj Agda.Syntax.Fixity.Associativity where
     valu [2] = valuN NonAssoc
     valu _   = malformed
 
-instance EmbPrj Agda.Syntax.Fixity.PrecedenceLevel where
+instance EmbPrj FixityLevel where
   icod_ Unrelated   = icodeN' Unrelated
   icod_ (Related a) = icodeN' Related a
 
@@ -353,12 +351,12 @@ instance EmbPrj Agda.Syntax.Fixity.PrecedenceLevel where
     valu [a] = valuN Related a
     valu _   = malformed
 
-instance EmbPrj Agda.Syntax.Fixity.Fixity where
+instance EmbPrj Fixity where
   icod_ (Fixity a b c) = icodeN' Fixity a b c
 
   value = valueN Fixity
 
-instance EmbPrj Agda.Syntax.Fixity.Fixity' where
+instance EmbPrj Fixity' where
   icod_ (Fixity' a b r) = icodeN' (\ a b -> Fixity' a b r) a b  -- discard theNameRange
 
   value = valueN (\ f n -> Fixity' f n noRange)
@@ -438,11 +436,6 @@ instance EmbPrj a => EmbPrj (Arg a) where
 
   value = valueN Arg
 
-instance EmbPrj a => EmbPrj (Dom a) where
-  icod_ (Dom a b c d) = icodeN' Dom a b c d
-
-  value = valueN Dom
-
 instance EmbPrj HasEta where
   icod_ YesEta = icodeN' YesEta
   icod_ NoEta  = icodeN 1 NoEta
@@ -473,21 +466,80 @@ instance EmbPrj Hiding where
   value 3 = return (Instance YesOverlap)
   value _ = malformed
 
-instance EmbPrj Quantity where
-  icod_ Quantity0 = return 0
-  icod_ Quantity1 = return 1
-  icod_ Quantityω = return 2
+instance EmbPrj Q0Origin where
+  icod_ = \case
+    Q0Inferred -> return 0
+    Q0 _       -> return 1
+    Q0Erased _ -> return 2
 
-  value 0 = return Quantity0
-  value 1 = return Quantity1
-  value 2 = return Quantityω
+  value = \case
+    0 -> return $ Q0Inferred
+    1 -> return $ Q0       noRange
+    2 -> return $ Q0Erased noRange
+    _ -> malformed
+
+instance EmbPrj Q1Origin where
+  icod_ = \case
+    Q1Inferred -> return 0
+    Q1 _       -> return 1
+    Q1Linear _ -> return 2
+
+  value = \case
+    0 -> return $ Q1Inferred
+    1 -> return $ Q1       noRange
+    2 -> return $ Q1Linear noRange
+    _ -> malformed
+
+instance EmbPrj QωOrigin where
+  icod_ = \case
+    QωInferred -> return 0
+    Qω _       -> return 1
+    QωPlenty _ -> return 2
+
+  value = \case
+    0 -> return $ QωInferred
+    1 -> return $ Qω       noRange
+    2 -> return $ QωPlenty noRange
+    _ -> malformed
+
+instance EmbPrj Quantity where
+  icod_ = \case
+    Quantity0 a -> icodeN 0 Quantity0 a
+    Quantity1 a -> icodeN 1 Quantity1 a
+    Quantityω a -> icodeN'  Quantityω a  -- default quantity, shorter code
+
+  value = vcase $ \case
+    [0, a] -> valuN Quantity0 a
+    [1, a] -> valuN Quantity1 a
+    [a]    -> valuN Quantityω a
+    _      -> malformed
+
+-- -- ALT: forget quantity origin when serializing?
+-- instance EmbPrj Quantity where
+--   icod_ Quantity0 = return 0
+--   icod_ Quantity1 = return 1
+--   icod_ Quantityω = return 2
+
+--   value 0 = return Quantity0
+--   value 1 = return Quantity1
+--   value 2 = return Quantityω
+--   value _ = malformed
+
+instance EmbPrj Cohesion where
+  icod_ Flat       = return 0
+  icod_ Continuous = return 1
+  icod_ Squash     = return 2
+
+  value 0 = return Flat
+  value 1 = return Continuous
+  value 2 = return Squash
   value _ = malformed
 
 instance EmbPrj Modality where
-  icod_ (Modality a b) = icodeN' Modality a b
+  icod_ (Modality a b c) = icodeN' Modality a b c
 
   value = vcase $ \case
-    [a, b] -> valuN Modality a b
+    [a, b, c] -> valuN Modality a b c
     _ -> malformed
 
 instance EmbPrj Relevance where
@@ -514,6 +566,11 @@ instance EmbPrj Origin where
   value 4 = return Substitution
   value _ = malformed
 
+instance EmbPrj a => EmbPrj (WithOrigin a) where
+  icod_ (WithOrigin a b) = icodeN' WithOrigin a b
+
+  value = valueN WithOrigin
+
 instance EmbPrj FreeVariables where
   icod_ UnknownFVs   = icodeN' UnknownFVs
   icod_ (KnownFVs a) = icodeN' KnownFVs a
@@ -597,3 +654,12 @@ instance EmbPrj Empty where
   icod_ a = icod_ =<< lift (Empty.toImpossible a)
 
   value = fmap throwImpossible . value
+
+instance EmbPrj ExpandedEllipsis where
+  icod_ NoEllipsis = icodeN' NoEllipsis
+  icod_ (ExpandedEllipsis a b) = icodeN 1 ExpandedEllipsis a b
+
+  value = vcase valu where
+    valu []      = valuN NoEllipsis
+    valu [1,a,b] = valuN ExpandedEllipsis a b
+    valu _       = malformed
