@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NoMonomorphismRestriction  #-}
 {-# LANGUAGE UndecidableInstances       #-}
@@ -7,13 +6,9 @@
 
 module Agda.TypeChecking.SizedTypes.Syntax where
 
-#if MIN_VERSION_base(4,11,0)
-import Prelude hiding ( (<>), null )
-#else
 import Prelude hiding ( null )
-#endif
 
-import Data.Maybe
+
 import Data.Foldable (Foldable)
 import Data.Map (Map)
 import qualified Data.Map as Map
