@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE UndecidableInstances #-}
 
 -- | A syntactic equality check that takes meta instantiations into account,
@@ -14,7 +13,6 @@ module Agda.TypeChecking.SyntacticEquality (SynEq, checkSyntacticEquality) where
 
 import Prelude hiding (mapM)
 
-import Control.Applicative hiding ((<**>))
 import Control.Arrow ((***))
 import Control.Monad.State hiding (mapM)
 
@@ -23,32 +21,33 @@ import Agda.Interaction.Options (optSyntacticEquality)
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
-import Agda.TypeChecking.Monad (ReduceM, MonadReduce(..), pragmaOptions)
+import Agda.TypeChecking.Monad (ReduceM, MonadReduce(..), pragmaOptions, isInstantiatedMeta)
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Monad (ifM)
 
-#include "undefined.h"
-import Agda.Utils.Impossible
 
 -- | Syntactic equality check for terms.
 --   @
 --      checkSyntacticEquality v v' = do
---        (v,v') <- instantiateFull (v,v')
---         return ((v,v'), v==v')
+--        (v, v') <- instantiateFull (v, v')
+--         return ((v, v'), v==v')
 --   @
---   only that @v,v'@ are only fully instantiated to the depth
+--   only that @v, v'@ are only fully instantiated to the depth
 --   where they are equal.
+--
+--   This means in particular that the returned @v,v'@ cannot be @MetaV@s
+--   that are instantiated.
 
 {-# SPECIALIZE checkSyntacticEquality :: Term -> Term -> ReduceM ((Term, Term), Bool) #-}
 {-# SPECIALIZE checkSyntacticEquality :: Type -> Type -> ReduceM ((Type, Type), Bool) #-}
-checkSyntacticEquality :: (SynEq a, MonadReduce m) => a -> a -> m ((a, a), Bool)
+checkSyntacticEquality :: (Instantiate a, SynEq a, MonadReduce m) => a -> a -> m ((a, a), Bool)
 checkSyntacticEquality v v' = liftReduce $ do
   ifM (optSyntacticEquality <$> pragmaOptions)
   {-then-} (synEq v v' `runStateT` True)
-  {-else-} (return ((v,v'), False))
+  {-else-} ((,False) <$> instantiate (v,v'))
 
 -- | Monad for checking syntactic equality
 type SynEqM = StateT Bool ReduceM
@@ -64,24 +63,24 @@ ifEqual cont a = ifM get (cont a) (return a)
 -- Since List2 is only Applicative, not a monad, I cannot
 -- define a List2T monad transformer, so we do it manually:
 
-(<$$>) :: Functor f => (a -> b) -> f (a,a) -> f (b,b)
+(<$$>) :: Functor f => (a -> b) -> f (a, a) -> f (b, b)
 f <$$> xx = (f *** f) <$> xx
 
-pure2 :: Applicative f => a -> f (a,a)
-pure2 a = pure (a,a)
+pure2 :: Applicative f => a -> f (a, a)
+pure2 a = pure (a, a)
 
-(<**>) :: Applicative f => f (a -> b, a -> b) -> f (a,a) -> f (b,b)
+(<**>) :: Applicative f => f (a -> b, a -> b) -> f (a, a) -> f (b, b)
 ff <**> xx = pure (uncurry (***)) <*> ff <*> xx
 
 -- | Instantiate full as long as things are equal
 class SynEq a where
-  synEq  :: a -> a -> SynEqM (a,a)
-  synEq' :: a -> a -> SynEqM (a,a)
+  synEq  :: a -> a -> SynEqM (a, a)
+  synEq' :: a -> a -> SynEqM (a, a)
   synEq' a a' = ifEqual (uncurry synEq) (a, a')
 
 instance SynEq Bool where
-  synEq x y | x == y = return (x,y)
-  synEq x y | otherwise = inequal (x,y)
+  synEq x y | x == y = return (x, y)
+  synEq x y | otherwise = inequal (x, y)
 
 -- | Syntactic term equality ignores 'DontCare' stuff.
 instance SynEq Term where
@@ -89,7 +88,7 @@ instance SynEq Term where
     (v, v') <- lift $ instantiate' (v, v')
     case (v, v') of
       (Var   i vs, Var   i' vs') | i == i' -> Var i   <$$> synEq vs vs'
-      (Con c ci vs,Con c' ci' vs') | c == c' -> Con c (bestConInfo ci ci') <$$> synEq vs vs'
+      (Con c i vs, Con c' i' vs') | c == c' -> Con c (bestConInfo i i') <$$> synEq vs vs'
       (Def   f vs, Def   f' vs') | f == f' -> Def f   <$$> synEq vs vs'
       (MetaV x vs, MetaV x' vs') | x == x' -> MetaV x <$$> synEq vs vs'
       (Lit   l   , Lit   l'    ) | l == l' -> pure2 $ v
@@ -97,21 +96,24 @@ instance SynEq Term where
       (Level l   , Level l'    )           -> levelTm <$$> synEq l l'
       (Sort  s   , Sort  s'    )           -> Sort    <$$> synEq s s'
       (Pi    a b , Pi    a' b' )           -> Pi      <$$> synEq a a' <**> synEq' b b'
-      (DontCare _, DontCare _  )           -> pure (v, v')
-         -- Irrelevant things are syntactically equal. ALT:
-         -- DontCare <$$> synEq v v'
+      (DontCare u, DontCare u' )           -> DontCare <$$> synEq u u'
+         -- Irrelevant things are not syntactically equal. ALT:
+         -- pure (u, u')
+         -- Jesper, 2019-10-21: considering irrelevant things to be
+         -- syntactically equal causes implicit arguments to go
+         -- unsolved, so it is better to go under the DontCare.
       (Dummy{}   , Dummy{}     )           -> pure (v, v')
       _                                    -> inequal (v, v')
 
 instance SynEq Level where
-  synEq (Max vs) (Max vs') = levelMax <$$> synEq vs vs'
+  synEq l@(Max n vs) l'@(Max n' vs')
+    | n == n'   = levelMax n <$$> synEq vs vs'
+    | otherwise = inequal (l, l')
 
 instance SynEq PlusLevel where
-  synEq l l' = do
-    case (l, l') of
-      (ClosedLevel v, ClosedLevel v') | v == v' -> pure2 l
-      (Plus n v,      Plus n' v')     | n == n' -> Plus n <$$> synEq v v'
-      _ -> inequal (l, l')
+  synEq l@(Plus n v) l'@(Plus n' v')
+    | n == n'   = Plus n <$$> synEq v v'
+    | otherwise = inequal (l, l')
 
 instance SynEq LevelAtom where
   synEq l l' = do
@@ -138,6 +140,7 @@ instance SynEq Sort where
     case (s, s') of
       (Type l  , Type l'   ) -> Type <$$> synEq l l'
       (PiSort a b, PiSort a' b') -> piSort <$$> synEq a a' <**> synEq' b b'
+      (FunSort a b, FunSort a' b') -> funSort <$$> synEq a a' <**> synEq' b b'
       (UnivSort a, UnivSort a') -> UnivSort <$$> synEq a a'
       (SizeUniv, SizeUniv  ) -> pure2 s
       (Prop l  , Prop l'   ) -> Prop <$$> synEq l l'
@@ -173,22 +176,14 @@ instance (Subst t a, SynEq a) => SynEq (Abs a) where
       (Abs   x b, NoAbs x' b') -> Abs x  <$$> synEq b (raise 1 b')  -- TODO: mkAbs?
       (NoAbs x b, Abs   x' b') -> Abs x' <$$> synEq (raise 1 b) b'
 
-{- TRIGGERS test/fail/UnequalHiding
--- | Ignores 'ArgInfo'.
-instance SynEq a => SynEq (Arg c a) where
-  synEq (Arg ai a) (Arg ai' a') = (Arg ai *** Arg ai') <$> synEq a a'
-
--- | Ignores 'ArgInfo'.
-instance SynEq a => SynEq (Dom c a) where
-  synEq (Dom ai a) (Dom ai' a') = (Dom ai *** Dom ai') <$> synEq a a'
--}
-
+-- NOTE: Do not ignore 'ArgInfo', or test/fail/UnequalHiding will pass.
 instance SynEq a => SynEq (Arg a) where
   synEq (Arg ai a) (Arg ai' a') = Arg <$$> synEq ai ai' <**> synEq a a'
 
+-- Ignore the tactic.
 instance SynEq a => SynEq (Dom a) where
-  synEq d@(Dom ai b x a) d'@(Dom ai' b' x' a')
-    | x == x'   = Dom <$$> synEq ai ai' <**> synEq b b' <**> pure2 x <**> synEq a a'
+  synEq d@(Dom ai b x t a) d'@(Dom ai' b' x' _ a')
+    | x == x'   = Dom <$$> synEq ai ai' <**> synEq b b' <**> pure2 x <**> pure2 t <**> synEq a a'
     | otherwise = inequal (d, d')
 
 instance SynEq ArgInfo where
