@@ -1,18 +1,13 @@
-{-# LANGUAGE CPP               #-}
 
 module Agda.TypeChecking.Unquote where
 
-#if MIN_VERSION_base(4,11,0)
-import Prelude hiding ((<>))
-#endif
-
 import Control.Arrow (first, second)
 import Control.Monad.State
 import Control.Monad.Reader
 import Control.Monad.Writer hiding ((<>))
-import Control.Monad.Trans (lift)
 
 import Data.Char
+import qualified Data.HashSet as HashSet
 import Data.Maybe (fromMaybe)
 import Data.Traversable (traverse)
 import Data.Word
@@ -20,6 +15,7 @@ import Data.Word
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
 import qualified Agda.Syntax.Reflected as R
+import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 import Agda.Syntax.Fixity
@@ -27,9 +23,9 @@ import Agda.Syntax.Info
 import Agda.Syntax.Translation.ReflectedToAbstract
 
 import Agda.TypeChecking.Constraints
+import Agda.TypeChecking.MetaVars.Mention
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
@@ -49,14 +45,12 @@ import Agda.Utils.Except
   , runExceptT
   )
 import Agda.Utils.Either
-import Agda.Utils.FileName
 import Agda.Utils.Lens
 import Agda.Utils.Monad
 import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.String ( Str(Str), unStr )
 import qualified Agda.Interaction.Options.Lenses as Lens
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 agdaTermType :: TCM Type
@@ -109,7 +103,7 @@ liftU2 f m1 m2 = packUnquoteM $ \ cxt s -> f (unpackUnquoteM m1 cxt s) (unpackUn
 inOriginalContext :: UnquoteM a -> UnquoteM a
 inOriginalContext m =
   packUnquoteM $ \ cxt s ->
-    modifyContext (const cxt) $ unpackUnquoteM m cxt s
+    unsafeModifyContext (const cxt) $ unpackUnquoteM m cxt s
 
 isCon :: ConHead -> TCM Term -> UnquoteM Bool
 isCon con tm = do t <- liftTCM tm
@@ -173,15 +167,15 @@ pickName a =
               isAlpha c -> [toLower c]
     _        -> "_"
 
--- TODO: reflect Quantity
+-- TODO: reflect Quantity, Cohesion
 instance Unquote Modality where
-  unquote t = (`Modality` defaultQuantity) <$> unquote t
+  unquote t = (\ r -> Modality r defaultQuantity defaultCohesion) <$> unquote t
 
 instance Unquote ArgInfo where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ es | Just [h,r] <- allApplyElims es -> do
+      Con c _ es | Just [h,r] <- allApplyElims es ->
         choice
           [(c `isCon` primArgArgInfo,
               ArgInfo <$> unquoteN h <*> unquoteN r <*> pure Reflected <*> pure unknownFreeVariables)]
@@ -193,7 +187,7 @@ instance Unquote a => Unquote (Arg a) where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ es | Just [info,x] <- allApplyElims es -> do
+      Con c _ es | Just [info,x] <- allApplyElims es ->
         choice
           [(c `isCon` primArgArg, Arg <$> unquoteN info <*> unquoteN x)]
           __IMPOSSIBLE__
@@ -256,7 +250,7 @@ unquoteString x = unStr <$> unquote x
 unquoteNString :: Arg Term -> UnquoteM String
 unquoteNString x = unStr <$> unquoteN x
 
-data ErrorPart = StrPart String | TermPart R.Term | NamePart QName
+data ErrorPart = StrPart String | TermPart A.Expr | NamePart QName
 
 instance PrettyTCM ErrorPart where
   prettyTCM (StrPart s) = text s
@@ -269,7 +263,7 @@ instance Unquote ErrorPart where
     case t of
       Con c _ es | Just [x] <- allApplyElims es ->
         choice [ (c `isCon` primAgdaErrorPartString, StrPart  <$> unquoteNString x)
-               , (c `isCon` primAgdaErrorPartTerm,   TermPart <$> unquoteN x)
+               , (c `isCon` primAgdaErrorPartTerm,   TermPart <$> ((liftTCM . toAbstractWithoutImplicit) =<< (unquoteN x :: UnquoteM R.Term)))
                , (c `isCon` primAgdaErrorPartName,   NamePart <$> unquoteN x) ]
                __IMPOSSIBLE__
       _ -> throwError $ NonCanonical "error part" t
@@ -278,11 +272,11 @@ instance Unquote a => Unquote [a] where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ es | Just [x,xs] <- allApplyElims es -> do
+      Con c _ es | Just [x,xs] <- allApplyElims es ->
         choice
           [(c `isCon` primCons, (:) <$> unquoteN x <*> unquoteN xs)]
           __IMPOSSIBLE__
-      Con c _ [] -> do
+      Con c _ [] ->
         choice
           [(c `isCon` primNil, return [])]
           __IMPOSSIBLE__
@@ -293,7 +287,7 @@ instance Unquote Hiding where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ [] -> do
+      Con c _ [] ->
         choice
           [(c `isCon` primHidden,  return Hidden)
           ,(c `isCon` primInstance, return (Instance NoOverlap))
@@ -306,7 +300,7 @@ instance Unquote Relevance where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ [] -> do
+      Con c _ [] ->
         choice
           [(c `isCon` primRelevant,   return Relevant)
           ,(c `isCon` primIrrelevant, return Irrelevant)]
@@ -325,7 +319,7 @@ instance Unquote a => Unquote (R.Abs a) where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ es | Just [x,y] <- allApplyElims es -> do
+      Con c _ es | Just [x,y] <- allApplyElims es ->
         choice
           [(c `isCon` primAbsAbs, R.Abs <$> (hint <$> unquoteNString x) <*> unquoteN y)]
           __IMPOSSIBLE__
@@ -356,11 +350,11 @@ instance Unquote R.Sort where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ [] -> do
+      Con c _ [] ->
         choice
           [(c `isCon` primAgdaSortUnsupported, return R.UnknownS)]
           __IMPOSSIBLE__
-      Con c _ es | Just [u] <- allApplyElims es -> do
+      Con c _ es | Just [u] <- allApplyElims es ->
         choice
           [(c `isCon` primAgdaSortSet, R.SetS <$> unquoteN u)
           ,(c `isCon` primAgdaSortLit, R.LitS <$> unquoteN u)]
@@ -396,7 +390,7 @@ instance Unquote R.Term where
           [ (c `isCon` primAgdaTermUnsupported, return R.Unknown) ]
           __IMPOSSIBLE__
 
-      Con c _ es | Just [x] <- allApplyElims es -> do
+      Con c _ es | Just [x] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaTermSort,      R.Sort      <$> unquoteN x)
           , (c `isCon` primAgdaTermLit,       R.Lit       <$> unquoteN x) ]
@@ -429,18 +423,18 @@ instance Unquote R.Pattern where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ [] -> do
+      Con c _ [] ->
         choice
           [ (c `isCon` primAgdaPatAbsurd, return R.AbsurdP)
           , (c `isCon` primAgdaPatDot,    return R.DotP)
           ] __IMPOSSIBLE__
-      Con c _ es | Just [x] <- allApplyElims es -> do
+      Con c _ es | Just [x] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaPatVar,  R.VarP  <$> unquoteNString x)
           , (c `isCon` primAgdaPatProj, R.ProjP <$> unquoteN x)
           , (c `isCon` primAgdaPatLit,  R.LitP  <$> unquoteN x) ]
           __IMPOSSIBLE__
-      Con c _ es | Just [x, y] <- allApplyElims es -> do
+      Con c _ es | Just [x, y] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaPatCon, R.ConP <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
@@ -451,11 +445,11 @@ instance Unquote R.Clause where
   unquote t = do
     t <- reduceQuotedTerm t
     case t of
-      Con c _ es | Just [x] <- allApplyElims es -> do
+      Con c _ es | Just [x] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaClauseAbsurd, R.AbsurdClause <$> unquoteN x) ]
           __IMPOSSIBLE__
-      Con c _ es | Just [x, y] <- allApplyElims es -> do
+      Con c _ es | Just [x, y] <- allApplyElims es ->
         choice
           [ (c `isCon` primAgdaClauseClause, R.Clause <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
@@ -481,17 +475,19 @@ evalTCM v = do
 
   case v of
     I.Def f [] ->
-      choice [ (f `isDef` primAgdaTCMGetContext, tcGetContext)
-             , (f `isDef` primAgdaTCMCommit,     tcCommit) ]
+      choice [ (f `isDef` primAgdaTCMGetContext,       tcGetContext)
+             , (f `isDef` primAgdaTCMCommit,           tcCommit)
+             ]
              failEval
     I.Def f [u] ->
-      choice [ (f `isDef` primAgdaTCMInferType,          tcFun1 tcInferType          u)
-             , (f `isDef` primAgdaTCMNormalise,          tcFun1 tcNormalise          u)
-             , (f `isDef` primAgdaTCMReduce,             tcFun1 tcReduce             u)
-             , (f `isDef` primAgdaTCMGetType,            tcFun1 tcGetType            u)
-             , (f `isDef` primAgdaTCMGetDefinition,      tcFun1 tcGetDefinition      u)
-             , (f `isDef` primAgdaTCMIsMacro,            tcFun1 tcIsMacro            u)
-             , (f `isDef` primAgdaTCMFreshName,          tcFun1 tcFreshName          u) ]
+      choice [ (f `isDef` primAgdaTCMInferType,                  tcFun1 tcInferType                  u)
+             , (f `isDef` primAgdaTCMNormalise,                  tcFun1 tcNormalise                  u)
+             , (f `isDef` primAgdaTCMReduce,                     tcFun1 tcReduce                     u)
+             , (f `isDef` primAgdaTCMGetType,                    tcFun1 tcGetType                    u)
+             , (f `isDef` primAgdaTCMGetDefinition,              tcFun1 tcGetDefinition              u)
+             , (f `isDef` primAgdaTCMIsMacro,                    tcFun1 tcIsMacro                    u)
+             , (f `isDef` primAgdaTCMFreshName,                  tcFun1 tcFreshName                  u)
+             ]
              failEval
     I.Def f [u, v] ->
       choice [ (f `isDef` primAgdaTCMUnify,      tcFun2 tcUnify      u v)
@@ -532,7 +528,7 @@ evalTCM v = do
       if norm then normalise v else instantiateFull v
 
     mkT l a = El s a
-      where s = Type $ Max [Plus 0 $ UnreducedLevel l]
+      where s = Type $ Max 0 [Plus 0 $ UnreducedLevel l]
 
     -- Don't catch Unquote errors!
     tcCatchError :: Term -> Term -> UnquoteM Term
@@ -592,7 +588,7 @@ evalTCM v = do
     tcCommit = do
       dirty <- gets fst
       when (dirty == Dirty) $
-        typeError $ GenericError "Cannot use commitTC after declaring new definitions."
+        liftTCM $ typeError $ GenericError "Cannot use commitTC after declaring new definitions."
       s <- getTC
       modify (second $ const s)
       liftTCM primUnitUnit
@@ -626,8 +622,7 @@ evalTCM v = do
     tcUnquoteTerm :: Type -> R.Term -> TCM Term
     tcUnquoteTerm a v = do
       e <- toAbstract_ v
-      v <- checkExpr e a
-      return v
+      checkExpr e a
 
     tcNormalise :: R.Term -> TCM Term
     tcNormalise v = do
@@ -653,12 +648,12 @@ evalTCM v = do
     tcExtendContext :: Term -> Term -> UnquoteM Term
     tcExtendContext a m = do
       a <- unquote a
-      extendCxt a (evalTCM m)
+      strengthen __UNREACHABLE__ <$> extendCxt a (evalTCM $ raise 1 m)
 
     tcInContext :: Term -> Term -> UnquoteM Term
     tcInContext c m = do
       c <- unquote c
-      liftU1 inTopContext $ go c (evalTCM m)
+      liftU1 unsafeInTopContext $ go c (evalTCM m)
       where
         go :: [Arg R.Type] -> UnquoteM Term -> UnquoteM Term
         go []       m = m
@@ -695,7 +690,7 @@ evalTCM v = do
       liftTCM $ do
         reportSDoc "tc.unquote.decl" 10 $ sep
           [ "declare" <+> prettyTCM x <+> ":"
-          , nest 2 $ prettyTCM a
+          , nest 2 $ prettyR a
           ]
         a <- isType_ =<< toAbstract_ a
         alreadyDefined <- isRight <$> getConstInfo' x
@@ -708,7 +703,7 @@ evalTCM v = do
     tcDeclarePostulate (Arg i x) a = inOriginalContext $ do
       clo <- commandLineOptions
       when (Lens.getSafeMode clo) $ liftTCM $ typeError . GenericDocError =<<
-        "Cannot postulate '" <+> prettyTCM x <+> ":" <+> prettyTCM a <+> "' with safe flag"
+        "Cannot postulate '" <+> prettyTCM x <+> ":" <+> prettyR a <+> "' with safe flag"
       setDirty
       let r = getRelevance i
       when (hidden i) $ liftTCM $ typeError . GenericDocError =<<
@@ -717,7 +712,7 @@ evalTCM v = do
       liftTCM $ do
         reportSDoc "tc.unquote.decl" 10 $ sep
           [ "declare Postulate" <+> prettyTCM x <+> ":"
-          , nest 2 $ prettyTCM a
+          , nest 2 $ prettyR a
           ]
         a <- isType_ =<< toAbstract_ a
         alreadyDefined <- isRight <$> getConstInfo' x
@@ -732,7 +727,9 @@ evalTCM v = do
         genericError $ "Missing declaration for " ++ prettyShow x
       cs <- mapM (toAbstract_ . QNamed x) cs
       reportSDoc "tc.unquote.def" 10 $ vcat $ map prettyA cs
-      let i = mkDefInfo (nameConcrete $ qnameName x) noFixity' PublicAccess ConcreteDef noRange
+      let accessDontCare = __IMPOSSIBLE__  -- or ConcreteDef, value not looked at
+      ac <- asksTC (^. lensIsAbstract)     -- Issue #4012, respect AbstractMode
+      let i = mkDefInfo (nameConcrete $ qnameName x) noFixity' accessDontCare ac noRange
       checkFunDef NotDelayed i x cs
       primUnitUnit
 
