@@ -96,7 +96,7 @@ iterateUntilM r f = loop where
 -- The applications are calculated strictly.
 
 iterate' :: Integral i => i -> (a -> a) -> a -> a
-iterate' 0 f x             = x
+iterate' 0 _ x             = x
 iterate' n f x | n > 0     = iterate' (n - 1) f $! f x
                | otherwise = error "iterate': Negative input."
 
