@@ -22,9 +22,7 @@ hashByteString :: ByteString -> Hash
 hashByteString = H.asWord64 . B.foldl' (\h b -> H.combine h (H.hashWord8 b)) (H.hashWord8 0)
 
 hashTextFile :: AbsolutePath -> IO Hash
-hashTextFile file = do
-  s <- readTextFile (filePath file)
-  return $ hashText s
+hashTextFile file = hashText <$> readTextFile (filePath file)
 
 -- | Hashes a piece of 'Text'.
 
