@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP #-}
 
 module Agda.Utils.Monad
     ( module Agda.Utils.Monad
@@ -8,16 +7,14 @@ module Agda.Utils.Monad
     )
     where
 
-import Prelude       hiding (concat)
+import Control.Applicative  (liftA2)
 import Control.Monad hiding (mapM, forM)
 
-#if __GLASGOW_HASKELL__ >= 800
 import qualified Control.Monad.Fail as Fail
-#endif
 
 import Control.Monad.Identity ( Identity )
 import Control.Monad.State
-import Control.Monad.Writer
+
 import Data.Traversable as Trav hiding (for, sequence)
 import Data.Foldable as Fold
 import Data.Maybe
@@ -28,21 +25,12 @@ import Agda.Utils.Except
   , MonadError(catchError, throwError)
   )
 
-import Agda.Utils.List
 import Agda.Utils.Null (ifNotNullM)
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 ---------------------------------------------------------------------------
 
-#if __GLASGOW_HASKELL__ >= 800
-instance Fail.MonadFail Identity where
-  fail = error
-#endif
-
----------------------------------------------------------------------------
-
 -- | Binary bind.
 (==<<) :: Monad m => (a -> b -> m c) -> (m a, m b) -> m c
 k ==<< (ma, mb) = ma >>= \ a -> k a =<< mb
@@ -61,9 +49,7 @@ guardM c = guard =<< c
 
 -- | Monadic if-then-else.
 ifM :: Monad m => m Bool -> m a -> m a -> m a
-ifM c m m' =
-    do  b <- c
-        if b then m else m'
+ifM c m m' = c >>= \b -> if b then m else m'
 
 -- | @ifNotM mc = ifM (not <$> mc)@
 ifNotM :: Monad m => m Bool -> m a -> m a -> m a
@@ -81,7 +67,7 @@ allM xs f = andM $ fmap f xs
 
 -- | Lazy monadic disjunction.
 or2M :: Monad m => m Bool -> m Bool -> m Bool
-or2M ma mb = ifM ma (return True) mb
+or2M ma = ifM ma (return True)
 
 orM :: (Foldable f, Monad m) => f (m Bool) -> m Bool
 orM = Fold.foldl or2M (return False)
@@ -105,7 +91,7 @@ orEitherM (m : ms) = caseEitherM m (\e -> mapLeft (e `mappend`) <$> orEitherM ms
 
 -- Loops gathering results in a Monoid ------------------------------------
 
--- | Generalized version of @mapM_ :: Monad m => (a -> m ()) -> [a] -> m ()@
+-- | Generalized version of @traverse_ :: Applicative m => (a -> m ()) -> [a] -> m ()@
 --   Executes effects and collects results in left-to-right order.
 --   Works best with left-associative monoids.
 --
@@ -116,11 +102,11 @@ orEitherM (m : ms) = caseEitherM m (\e -> mapLeft (e `mappend`) <$> orEitherM ms
 --   that collects results in right-to-left order
 --   (effects still left-to-right).
 --   It might be preferable for right associative monoids.
-mapM' :: (Foldable t, Monad m, Monoid b) => (a -> m b) -> t a -> m b
-mapM' f = Fold.foldl (\ mb a -> liftM2 mappend mb (f a)) (return mempty)
+mapM' :: (Foldable t, Applicative m, Monoid b) => (a -> m b) -> t a -> m b
+mapM' f = Fold.foldl (\ mb a -> liftA2 mappend mb (f a)) (pure mempty)
 
--- | Generalized version of @forM_ :: Monad m => [a] -> (a -> m ()) -> m ()@
-forM' :: (Foldable t, Monad m, Monoid b) => t a -> (a -> m b) -> m b
+-- | Generalized version of @for_ :: Applicative m => [a] -> (a -> m ()) -> m ()@
+forM' :: (Foldable t, Applicative m, Monoid b) => t a -> (a -> m b) -> m b
 forM' = flip mapM'
 
 -- Variations of Traversable
@@ -131,6 +117,14 @@ mapMM f mxs = Trav.mapM f =<< mxs
 forMM :: (Traversable t, Monad m) => m (t a) -> (a -> m b) -> m (t b)
 forMM = flip mapMM
 
+-- Variations of Foldable
+
+mapMM_ :: (Foldable t, Monad m) => (a -> m ()) -> m (t a) -> m ()
+mapMM_ f mxs = Fold.mapM_ f =<< mxs
+
+forMM_ :: (Foldable t, Monad m) => m (t a) -> (a -> m ()) -> m ()
+forMM_ = flip mapMM_
+
 -- Continuation monad -----------------------------------------------------
 
 -- Andreas, 2017-04-11, issue #2543
@@ -150,10 +144,18 @@ forMM = flip mapMM
 mapMaybeM :: Monad m => (a -> m (Maybe b)) -> [a] -> m [b]
 mapMaybeM f xs = catMaybes <$> Trav.mapM f xs
 
+-- | A version of @'mapMaybeM'@ with a computation for the input list.
+mapMaybeMM :: Monad m => (a -> m (Maybe b)) -> m [a] -> m [b]
+mapMaybeMM f m = mapMaybeM f =<< m
+
 -- | The @for@ version of 'mapMaybeM'.
 forMaybeM :: Monad m => [a] -> (a -> m (Maybe b)) -> m [b]
 forMaybeM = flip mapMaybeM
 
+-- | The @for@ version of 'mapMaybeMM'.
+forMaybeMM :: Monad m => m [a] -> (a -> m (Maybe b)) -> m [b]
+forMaybeMM = flip mapMaybeMM
+
 -- | A monadic version of @'dropWhile' :: (a -> Bool) -> [a] -> [a]@.
 dropWhileM :: Monad m => (a -> m Bool) -> [a] -> m [a]
 dropWhileM p []       = return []
@@ -191,7 +193,7 @@ catMaybesMP = (>>= fromMaybeMP)
 
 finally :: MonadError e m => m a -> m () -> m a
 first `finally` after = do
-  r <- catchError (liftM Right first) (return . Left)
+  r <- catchError (fmap Right first) (return . Left)
   after
   case r of
     Left e  -> throwError e
@@ -202,6 +204,11 @@ first `finally` after = do
 tryMaybe :: (MonadError e m, Functor m) => m a -> m (Maybe a)
 tryMaybe m = (Just <$> m) `catchError` \ _ -> return Nothing
 
+-- | Run a command, catch the exception and return it.
+
+tryCatch :: (MonadError e m, Functor m) => m () -> m (Maybe e)
+tryCatch m = (Nothing <$ m) `catchError` \ err -> return $ Just err
+
 -- State monad ------------------------------------------------------------
 
 -- | Bracket without failure.  Typically used to preserve state.
@@ -227,33 +234,3 @@ readM s = case reads s of
             [(x,"")]    -> return x
             _           ->
               throwError $ strMsg $ "readM: parse error string " ++ s
-
-
--- RETIRED STUFF ----------------------------------------------------------
-
-{- RETIRED, ASR, 09 September 2014. Not used.
--- | Bracket for the 'Error' class.
-
--- bracket :: (Error e, MonadError e m)
---         => m a         -- ^ Acquires resource. Run first.
---         -> (a -> m c)  -- ^ Releases resource. Run last.
---         -> (a -> m b)  -- ^ Computes result. Run in-between.
---         -> m b
--- bracket acquire release compute = do
---   resource <- acquire
---   compute resource `finally` release resource
--}
-
-{- RETIRED, Andreas, 2012-04-30. Not used.
-concatMapM :: Applicative m => (a -> m [b]) -> [a] -> m [b]
-concatMapM f xs = concat <$> traverse f xs
-
--- | Depending on the monad you have to look at the result for
---   the force to be effective. For the 'IO' monad you do.
-forceM :: Monad m => [a] -> m ()
-forceM xs = do () <- length xs `seq` return ()
-               return ()
-
-commuteM :: (Traversable f, Applicative m) => f (m a) -> m (f a)
-commuteM = traverse id
--}
