@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP                   #-}
 {-# LANGUAGE DeriveDataTypeable    #-}
 
 module Agda.Utils.Permutation where
@@ -22,7 +21,6 @@ import Agda.Utils.List ((!!!))
 import Agda.Utils.Null
 import Agda.Utils.Size
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 -- | Partial permutations. Examples:
@@ -101,7 +99,7 @@ instance InversePermute [Maybe a] [Maybe a] where
     where tabulate m = for [0..n-1] $ \ i -> IntMap.lookup i m
 
 instance InversePermute (Int -> a) [Maybe a] where
-  inversePermute (Perm n xs) f = for [0..n-1] $ \ x -> f <$> List.findIndex (x ==) xs
+  inversePermute (Perm n xs) f = for [0..n-1] $ \ x -> f <$> List.elemIndex x xs
 
 -- | Identity permutation.
 idP :: Int -> Permutation
@@ -196,7 +194,7 @@ expandP i n (Perm m xs) = Perm (m + n - 1) $ concatMap expand xs
 -- | Stable topologic sort. The first argument decides whether its first
 --   argument is an immediate parent to its second argument.
 topoSort :: (a -> a -> Bool) -> [a] -> Maybe Permutation
-topoSort parent xs = fmap (Perm (size xs)) $ topo g
+topoSort parent xs = Perm (size xs) <$> topo g
   where
     nodes     = zip [0..] xs
     g         = [ (n, parents x) | (n, x) <- nodes ]
