@@ -1,7 +1,5 @@
-{-# LANGUAGE CPP                #-}
 {-# LANGUAGE BangPatterns       #-}
 {-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE StandaloneDeriving #-}
 
 module Agda.Utils.Pointer
   ( Ptr, newPtr, derefPtr, setPtr
@@ -10,17 +8,16 @@ module Agda.Utils.Pointer
 
 import Control.DeepSeq
 import Control.Concurrent.MVar
-import Data.Foldable
+
 import Data.Function
 import Data.Hashable
 import Data.IORef
-import Data.Traversable
+
 import System.IO.Unsafe
 
 import Data.Data (Data (..))
 import Data.Typeable (Typeable)
 
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 data Ptr a = Ptr { ptrTag :: !Integer
