@@ -1,12 +1,8 @@
-{-# LANGUAGE CPP #-}
 
 module Agda.Utils.Suffix where
 
 import Data.Char
 
-import Agda.Utils.Function
-
-#include "undefined.h"
 import Agda.Utils.Impossible
 
 import Data.IORef
@@ -23,20 +19,15 @@ subscriptAllowed = UNSAFE.unsafePerformIO (readIORef unicodeOrAscii)
 -- | Is the character one of the subscripts @'₀'@-@'₉'@?
 
 isSubscriptDigit :: Char -> Bool
-isSubscriptDigit c = case subscriptAllowed of
-  UnicodeOk -> '₀' <= c && c <= '₉'
-  AsciiOnly -> '0' <= c && c <= '9'
+isSubscriptDigit c = '₀' <= c && c <= '₉'
 
--- | Converts @'0'@-@'9'@ to @'₀'@-@'₉'@ unless the user doesn't want us
--- to use unicode characters
+-- | Converts @'0'@-@'9'@ to @'₀'@-@'₉'@
 --
 -- Precondition: The digit needs to be in range.
 
 toSubscriptDigit :: Char -> Char
 toSubscriptDigit d
-  | isDigit d = case subscriptAllowed of
-      UnicodeOk -> toEnum (fromEnum '₀' + (fromEnum d - fromEnum '0'))
-      AsciiOnly -> d
+  | isDigit d = toEnum (fromEnum '₀' + (fromEnum d - fromEnum '0'))
   | otherwise = __IMPOSSIBLE__
 
 -- | Converts @'₀'@-@'₉'@ to @'0'@-@'9'@.
@@ -45,9 +36,7 @@ toSubscriptDigit d
 
 fromSubscriptDigit :: Char -> Char
 fromSubscriptDigit d
-  | isSubscriptDigit d = case subscriptAllowed of
-      UnicodeOk -> toEnum (fromEnum '0' + (fromEnum d - fromEnum '₀'))
-      AsciiOnly -> d
+  | isSubscriptDigit d = toEnum (fromEnum '0' + (fromEnum d - fromEnum '₀'))
   | otherwise          = __IMPOSSIBLE__
 
 ------------------------------------------------------------------------
@@ -61,10 +50,13 @@ data Suffix
   | Index     Int  -- ^ Identifier ends in number @Int@ (ordinary digits).
   | Subscript Int  -- ^ Identifier ends in number @Int@ (subscript digits).
 
--- | Increase the suffix by one.  If no suffix yet, put a subscript @1@.
+-- | Increase the suffix by one.  If no suffix yet, put a subscript @1@
+--   unless users do not want us to use any unicode.
 
 nextSuffix :: Suffix -> Suffix
-nextSuffix NoSuffix      = Subscript 1
+nextSuffix NoSuffix      = case subscriptAllowed of
+  UnicodeOk -> Subscript 1
+  AsciiOnly -> Index 1
 nextSuffix (Prime i)     = Prime $ i + 1
 nextSuffix (Index i)     = Index $ i + 1
 nextSuffix (Subscript i) = Subscript $ i + 1
@@ -75,8 +67,7 @@ suffixView :: String -> (String, Suffix)
 suffixView s
     | (ps@(_:_), s') <- span (=='\'') rs         = (reverse s', Prime $ length ps)
     | (ns@(_:_), s') <- span isDigit rs          = (reverse s', Index $ read $ reverse ns)
-    | (ns@(_:_), s') <- span isSubscriptDigit rs = (reverse s',
-                                                    Subscript $ read $
+    | (ns@(_:_), s') <- span isSubscriptDigit rs = (reverse s', Subscript $ read $
                                                       map fromSubscriptDigit $ reverse ns)
     | otherwise                                  = (s, NoSuffix)
     where rs = reverse s
