@@ -1,4 +1,3 @@
-{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 -- To avoid warning on derived Integral instance for CPUTime.
@@ -18,7 +17,7 @@ module Agda.Utils.Time
 import Control.Monad.Trans
 import qualified System.CPUTime as CPU
 
-import Data.Functor
+
 import qualified Data.Time
 
 import Agda.Utils.Pretty
