@@ -1,6 +1,4 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE BangPatterns #-}
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 -- | Strict tries (based on "Data.Map.Strict" and "Agda.Utils.Maybe.Strict").
 
@@ -15,16 +13,15 @@ module Agda.Utils.Trie
   ) where
 
 import Prelude hiding (null, lookup, filter)
-import qualified Prelude
 
-import Control.Monad
+
+
 import Data.Function
-import Data.Functor
 import Data.Foldable (Foldable)
 import qualified Data.Maybe as Lazy
 import Data.Map.Strict (Map)
 import qualified Data.Map.Strict as Map
-import Data.Monoid
+
 import qualified Data.List as List
 
 import qualified Agda.Utils.Maybe.Strict as Strict
